/******************************************************************************************************************
* CMBase
* This class provides basic support for objects that appear in the table of contents (TOC).
*
* - Selecting and unselecting
* - Settings
* - Containing children and being contained by a parent object
* - Listeners
*
* @module CMBase
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/

//*****************************************************************************************************************
// Global Definitions
//*****************************************************************************************************************
/**
* Types of objects that are used through out CanvasMap.
*
* @public, @enum
*/
CMBase.DATA_TYPE_COORDINATES=1; // coordinates as lat/lon or easting/northing
CMBase.DATA_TYPE_COLOR=2; // 
CMBase.DATA_TYPE_INTEGER=3;
CMBase.DATA_TYPE_BOOLEAN=4;
CMBase.DATA_TYPE_FLOAT=5;
CMBase.DATA_TYPE_CSS_STYLE=6; // 
CMBase.DATA_TYPE_ENUMERATED=7; // 
CMBase.DATA_TYPE_URL=8; // 
CMBase.DATA_TYPE_STRING=9; // 
CMBase.DATA_TYPE_IMAGE=10; // 
CMBase.DATA_TYPE_FONT=11; // 
CMBase.DATA_TYPE_BOUNDS=12; // bounds objects with XMin,YMin,XMax,YMax 

CMBase.DATA_TYPE_BOOLEAN_ARRAY=13; // 
CMBase.DATA_TYPE_VECTOR=14; // x,y and optional z value

/** This is the unique number that is used to create message values
* @private
*/
CMBase.UniqueNumber=1;

//*****************************************************************************************************************
// Constructors
//*****************************************************************************************************************
/**
* Constructor for a base object.   The base object is inherited by most objects and provides function definitions
* for:
* - Child (contained) objects
* - Parent (container) objects
* - Messaging to listeners
* - Basic settings 
*
* CMBase also contains static functions to generate unique numbers for definitions that are shared between classes
*
* @public, @constructs
*/
function CMBase() 
{
}
//******************************************************************
// Protected Functions
//******************************************************************
/**
* Error hangling function for when an unknown Group or Key are specified
* @protected
*/
CMBase.prototype.SettingsError=function(Group,Key)
{
	//var ClassName=this.constructor.name; // jjg does not work, we have to wait for IE to get real class support
				
	if (Key==undefined)
	{
		CMMainContainer.Error("Sorry, we could not find the settings Group '"+Group+"'");
	}
	else
	{
		CMMainContainer.Error("Sorry, we could not find the Key '"+Key+"' in the Group '"+Group+"'");
	}
}

//******************************************************************
// Functions overriden by subclasses
//******************************************************************
/*
* Gets the current name of the object
* @public
* @returns Name - the name of this object
*/
CMBase.prototype.GetName=function()  { return("Untitled"); }

/*
* Adds an entry to the child array
* @public
* @param TheObject -
*/
CMBase.prototype.AddChild=function(TheObject)  
{ 
	if (this.TheChildren==undefined) this.TheChildren=[];
	
	var Result=this.TheChildren.length; // return the index to the new entry
	
	this.TheChildren.push(TheObject);
	
	TheObject.SetParent(this);
	
	return(Result);
}
/*
* Adds an entry to the child array
* @public
* @param TheObject -
*/
CMBase.prototype.RemoveChild=function(TheObject)  
{ 
	var Index=this.TheChildren.indexOf(TheObject);
	
	if (Index!=-1)
	{
		this.TheChildren.splice(Index,1);
	
		TheObject.SetParent(null);
	}
	return(Index);
}
/*
* Gets the number of children the object contains.
* @public
* @param TheClass - optional class to filter the children
* @returns NumChildren - the number of children contained in this object.
*/
CMBase.prototype.GetNumChildren=function(TheClass)  
{ 
	var Result=0;
	
	if (this.TheChildren!=undefined)
	{
		if (TheClass==undefined) 
		{
			Result=this.TheChildren.length;
		}
		else
		{
			for (var i=0;i<this.TheChildren.length;i++)
			{
				if (this.TheChildren[i] instanceof TheClass) Result++;
			}
		}
	}
	return(Result); 
}

/*
* Gets a child from the object
* @public
* @param Index - zero-based index to the desired child.  If not provided, the first entry is returned.
* @param TheClass - optional class to filter the children
* @returns TheChild - the child object contained within this object.
*/
CMBase.prototype.GetChild=function(Index,TheClass)  
{ 
	var Result=null;
	
	if (this.TheChildren!=undefined)
	{
		if (Index==undefined) Index=0;
		
		if (TheClass==undefined) 
		{
			Result=this.TheChildren[Index];
		}
		else
		{
			var TempIndex=0;
			for (var i=0;(i<this.TheChildren.length)&&(Result==null);i++)
			{
				if (this.TheChildren[i] instanceof TheClass)
				{
					if (TempIndex==Index) Result=this.TheChildren[i];
					TempIndex++;
				}
			}
		}
	}
	return(Result); 
}
CMBase.prototype.DeleteAllChildren=function() 
{
	this.TheChildren=undefined;
}

/*
* Gets the total number of descendants that are in this object.  This includes
* direct children, their children, and so on.
* @public
* @param Index - zero-based index to the desired child
* @param TheClass - optional class to filter the children
* @returns TheChild - the child object contained within this object.
*/
CMBase.prototype.GetNumDescendants=function(TheClass)
{
	var NumDescendants=0; // 
	
	var NumChildren=this.GetNumChildren(TheClass);
	
	for (var i=0;i<NumChildren;i++)
	{
		NumDescendants+=1; // one for this child
		
		 var TheChild=this.GetChild(i);
		
		NumDescendants+=TheChild.GetNumDescendants(TheClass); // add the childs descendents
	}
	return(NumDescendants);
}
/**
* Returns a descendent of this object.
* @public
* @param Index - 0 this object's first child, 2 for this objects first grandchild or second child, etc.
* @param CurrentRow - optional parameter for the row we are current looking for a descendent on.
*/
CMBase.prototype.GetDescendant=function(Index,CurrentRow)
{
	var TheObject=null;
	
	if (CurrentRow==undefined) CurrentRow=0; // just getting started
	
	{
		var NumChildren=this.GetNumChildren();
		
		for (var i=0;(i<NumChildren)&&(TheObject==null);i++)
		{
			if (CurrentRow==Index)// found the right child
			{ 
				TheObject=this.GetChild(i);
			}
			else // see if this one of the childs descendants
			{
				var TheChild=this.GetChild(i);
				
				var NumDescendants=TheChild.GetNumDescendants();
				
				if (NumDescendants+CurrentRow>=Index) // the desired descenent is in this child
				{
					TheObject=TheChild.GetDescendant(Index,CurrentRow+1);
				}
				else // go to the next child
				{
					CurrentRow+=NumDescendants;
				}
			}
			// move to the next child
			CurrentRow++; 
		}
	}
	return(TheObject);
}
/*
* Sets the parent that this object is contained in.
* @public
* @param NewParent - the new parent that this object is contained within
*/
CMBase.prototype.SetParent=function(NewParent)  
{ 
	this.TheParent=NewParent;
}
/*
* Gets the parent that this object is contained in.
* @public
* @returns TheParent - the parent this object is within
*/
CMBase.prototype.GetParent=function(Class)  
{ 
	var Result=null;
	
	if (this.TheParent!=undefined) 
	{
		if (Class!=undefined)
		{
			if (this.TheParent instanceof Class) Result=this.TheParent;
			else 
			{
				try{
					
					Result=this.TheParent.GetParent(Class);
				}
				catch(Error) {
					throw(Error);
				}
			}
		}
		else
		{
			if (this.TheParent!=undefined) Result=this.TheParent;
		}
	}
	return(Result); 
}
/**
* General function to send a message to all the descendents of an object.
* Used to send event messages to children.
* @public
*/
CMBase.prototype.SendMessageToDescendants=function(Message,AdditionalInfo)  
{ 
	var Used=false;
	
	var NumChildren=this.GetNumChildren();
	
	for (var i=0;(i<NumChildren)&&(Used==false);i++)
	{
		var TheChild=this.GetChild(i);
		
		Used=TheChild.SendMessageToDescendants(Message,AdditionalInfo);
	}
	return(Used);
}

//******************************************************************
// Listener functions
//******************************************************************
/**
* Add a listener to listen for messages from this object.  Messages are
* delivered with the function call:
*
* TheFunction(TheSender,TheListener,AdditionalInfo)
*
* TheFunction will receive the messages when sent by TheSender.  TheListener
* is the object that added itself as the listener so the function can access
* its properties.  AdditionaInfo is provided by the sender based on the type of message.
*
* @public
* @param TheFunction - The function that will listen to messages
* @param TheMessage - The message number the object wants to listen to
* @param TheListener - The object that will receive messages from this object.
* @returns UniqueNumber - returns a unique number for the function that has been added so it can be removed easily.
* 	This avoids the problem with addEventListener() that you have to have a uniquely named function outside the class.
*/
CMBase.prototype.AddListener=function(TheMessage,TheListener,TheFunction)
{
	if (TheMessage==undefined) 
	{
		throw("Sorry, the message is undefined");
	}
	if (this.Listeners==undefined) this.Listeners=[];
	
	if (this.Listeners[TheMessage]==undefined) this.Listeners[TheMessage]=[];
	
	var UniqueNumber=CMBase.GetUniqueNumber();
	
	this.Listeners[TheMessage].push(
	{
		TheFunction:TheFunction,
		TheListener:TheListener,
		UniqueNumber:UniqueNumber
	});
	
	return(UniqueNumber);
}
/**
* Removes a listener from the list of listeners
* @public
* @param TheMessage - The message that the listener was listening for
* @param UniqueNumber - the unique number that was generated when the listener was added.
*/
CMBase.prototype.RemoveListener=function(TheMessage,UniqueNumber)
{
	if (this.Listeners!=undefined)
	{		
		var TheListeners=this.Listeners[TheMessage];
		
		for (var i=0;i<TheListeners.length;i++)
		{
			if (TheListeners[i].UniqueNumber==UniqueNumber)
			{
				TheListeners.splice(i,1);
			}
		}
	}
}
/**
* Removes a listener from the list of listeners
* @public
* @param TheMessage - The message to send
* @param AdditionalInfo - additional information based on the message type
*/
CMBase.prototype.SendMessageToListeners=function(TheMessage,AdditionalInfo)
{
	if (TheMessage==undefined)
	{
		throw("Sorry, the message is undefined");
	}
	// make sure we have listeners and we are not sending to ourselves (stops infiinite message loops)
	
	if (this.Listeners!=undefined) 
	{	
		var TheListeners=this.Listeners[TheMessage];
		
		if (TheListeners!=undefined)
		{
			for (var i=0;i<TheListeners.length;i++)
			{
				var TheListener=TheListeners[i].TheListener;
				
				TheListeners[i].TheFunction(this,TheListener,AdditionalInfo);
			}
		}
	}
}

//******************************************************************
// Functions for managing settings and timeslices
//******************************************************************
/**
* Returns the specified Group settings.  If the group has not allocated will return an empty
* dictionary.  If the group is not defined, will general an error
* @protected
*/
CMBase.prototype.GetGroupFromSettings=function(GroupKey,Settings)
{
	var GroupSettings=Settings[GroupKey];
	
	if (GroupSettings==undefined) // if the group is not in the current settings, see if the group is defined
	{
		var SettingsDefinitions=this.GetSettingsDefinitions();
		
		var GroupDefinitions=SettingsDefinitions[GroupKey];
		
		if (GroupDefinitions!=undefined) // if the group is defined, try to set the values with the keys
		{
			GroupSettings={};
			Settings[GroupKey]=GroupSettings;
		}
		else this.SettingsError(GroupKey);
	}
	return(GroupSettings);
}

/**
* Override the GetSettingsDefinition() function to add our settings definitions
* Combined with our superclass's settings definitions.  Each sub class
* should add their settings definitions to the JSON object with the class's
* name as the key.
* @protected
*/
CMBase.prototype.GetSettingsDefinitions=function() 
{
	var Result={ }; 
	
	return(Result); // eventually, this will return a JSON object with objects for each superclass's settings definitions
}
/**
* Called when any settings are changed.
* @protected
*/
CMBase.prototype.SettingsChanged=function() 
{
}
/**
* Get a set of settings.  
* @public
* @returns Settings - current settings object (may be undefined)
*/
CMBase.prototype.GetSettings=function()
{
	return(this.Settings);
}
/**
* Sets all the settings for this object.  
* @public
* @param NewSettings - The settings to replace the existing settings.
*/
CMBase.prototype.SetSettings=function(NewSettings)
{
	this.Settings=NewSettings;
	
	this.SettingsChanged();
	
	this.SendMessageToListeners(CMBase.MESSAGE_SETTINGS_CHANGED,null);
}

/**
* Returns the specified Group settings.  If the group has not allocated will return an empty
* dictionary.  If the group is not defined, will general an error
* @protected
*/
CMBase.prototype.SetSettingGroup=function(GroupKey,GroupSettings)
{
	if (this.Settings==undefined) this.Settings={};
	
	this.Settings[GroupKey]=GroupSettings;
	
	this.SettingsChanged();
	
	this.SendMessageToListeners(CMBase.MESSAGE_SETTINGS_CHANGED,null);
}
/**
* Get a group of settings
* @public
* @param Group - Group for the setting
* @param Default - default value to use in none has been specified as of yet (optional)
* @returns Value - curnet property value.
*/
CMBase.prototype.GetSettingGroup=function(Group)
{
	var Result=this.Settings[Group];

	return(Result);
}

/**
* Sets a set of settings for this object.  The TimeSlice must already exist within the object.
* @public
* @param NewSettings - The settings to replace the existing settings.
*/
CMBase.prototype.GetSetting=function(Group,Key,Default)
{
	var Result=null; //this.SettingDefintions[Group][Key].Default;
	
	if (Default!=undefined) Result=Default;

	//if (this.Settings!=undefined)
	{
		var GroupSettings;
		
		// try to get the settings from the current object settings
		if (this.Settings!=undefined) GroupSettings=this.Settings[Group];
		
		if (GroupSettings!=undefined) // group has not been defined yet
		{
			if (Key in GroupSettings) Result=GroupSettings[Key];
		}
		// if we could not find a setting value and the default was not provided, get the default from the definitions
		if (Result==null) 
		{
			var SettingsDefinitions=this.GetSettingsDefinitions();
			
			if (Group in SettingsDefinitions)
			{
				var GroupDefinitions=SettingsDefinitions[Group];
				
				if (Key in GroupDefinitions) 
				{
					var Temp=GroupDefinitions[Key];
					var Result=Temp.Default;
				}
				else this.SettingsError(Group,Key);
			}
			else this.SettingsError(Group);
		}
	}
	return(Result);
}

/**
* Sets a set of settings for this object. ct.
* @public
* @param Group - The settings to replace the existing settings.
*/
CMBase.prototype.SetSetting=function(Group,Key,Value)
{
	// put the settings into a dictionary entry
	if (this.Settings==undefined) this.Settings={};
	
	if (this.Settings[Group]!=undefined) // group is in current settings
	{
		if (Key in this.Settings[Group]) // key is in current settings
		{
			this.Settings[Group][Key]=Value; // replace the current value
		}
		else // make sure the key is in the definitions and then add it
		{
			var SettingsDefinitions=this.GetSettingsDefinitions();
			
			var GroupDefinitions=SettingsDefinitions[Group];
			
			if (GroupDefinitions!=undefined) // group is in the group definitions
			{
				if (Key in GroupDefinitions)  // check if the key is in the definitions
				{
					this.Settings[Group][Key]=Value; // add the new value
				}
				else this.SettingsError(Group,Key);
			}
			else this.SettingsError(Group);
		}
	}
	else // group is not in current settings
	{
		var SettingsDefinitions=this.GetSettingsDefinitions();
		
		var GroupDefinitions=SettingsDefinitions[Group];
			
		if (GroupDefinitions!=undefined) // group is in the group definitions
		{
			if (Key in GroupDefinitions)   // check if the key is in the definitions
			{
				this.Settings[Group]={}; // add the new group
				this.Settings[Group][Key]=Value; // add the new value
			}
			else this.SettingsError(Group,Key);
		}
		else this.SettingsError(Group);
	}
	// notify decendents of the settings change
	this.SettingsChanged();
	
	// notify colleagues of the settings change
	this.SendMessageToListeners(CMBase.MESSAGE_SETTINGS_CHANGED,null);
}
/**
* High performance funtion to set settings into the existing Settings
* dictionary.  This function should be used when speed and not error
* checking are needed.  For ENUMEATED arrays, Values are integers. 
*
* This is the function that should be overriden by subclasses
* @protected
* @param Dictionary - Contains groups of settings based on SettingsDefinitions for this class.
*/
CMBase.prototype.Set=function(Dictionary)
{
	if (this.Settings==undefined) this.Settings={};
	
	for (var Group in Dictionary) 
	{
		if (this.Settings[Group]==undefined) 
		{
			this.Settings[Group]=Dictionary[Group];
		}
		else
		{
			for (var Key in Dictionary)
			{
				this.Settings[Group][Key]=Value;
			}
		}
	}
	
	this.SettingsChanged();
	
	this.SendMessageToListeners(CMBase.MESSAGE_SETTINGS_CHANGED,null);
}

//******************************************************************
// static functions
//******************************************************************
/**
* Gets an integer that is unique for this instance of CanvasMap.  This is
* used to ensure that all messages have a unique value.
* @public
* @returns UniqueNumber unique integer
*/
CMBase.GetUniqueNumber=function() 
{
	CMBase.UniqueNumber++;
	return(CMBase.UniqueNumber-1);
}
/**
* Types of messages
* @protected
*/
CMBase.MESSAGE_SETTINGS_CHANGED=CMBase.GetUniqueNumber();

