/******************************************************************************************************************
* CMDatasetPyramidOpenFormat
*
* @module CMDatasetPyramidOpenFormat
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/
//******************************************************************
// Constructor
//******************************************************************
/*
* Constructor
* @protected, @constructs
*/
function CMDatasetPyramidOpenFormat() 
{
	CMDataset.call(this);

	// properties
	this.TopTile=null; // first tile at ZoomLevel=-18, one tile=whole world
	this.Visible=true; // keeps track of visibility so we know when to make the tiles invisible (jjg - setting?)
	
	this.TileWidthInPixels=256;
	
	this.PaintDebugTile=false; // (jjg - setting?)
	this.DebugZoomLevel=-1;
	this.DebugGlobalColumn=0;
	this.DebugGlobalRow=0;
	
	this.MeshIsInGeo=false;
	this.TheMesh=null;
	
	this.FileExtension=null;
	
	// OpenPyramids do not have a bounds so we have to assume the entire world
	this.TheBounds={
		XMin:-180,
		XMax:180,
		YMin:-90,
		YMax:90
	};

}
CMDatasetPyramidOpenFormat.prototype=Object.create(CMDataset.prototype); // inherit prototype functions from PanelBase()

CMDatasetPyramidOpenFormat.prototype.contructor=CMDatasetPyramidOpenFormat; // override the constructor to go to ours

//******************************************************************
// Functions used by subclasses and not overriden
//*****************************************************************
CMDatasetPyramidOpenFormat.prototype.SetDebugTile=function(ZoomLevel,GlobalColumn,GlobalRow) 
{
	this.PaintDebugTile=true;
	this.DebugZoomLevel=ZoomLevel;
	this.DebugGlobalColumn=GlobalColumn;
	this.DebugGlobalRow=GlobalRow;
}

/*
* Called to obtain the data for the layer from a URL.
* Currently,only GeoJSON is supported as the transfer type.
*/
CMDatasetPyramidOpenFormat.prototype.SetURL=function(URL) 
{
	if (this.URL!=URL)
	{
		// get the file extension, if any
		
		URL=URL.trim();
		
		var LastChar=URL[URL.length-1];
		
		if (LastChar!="}") // have an extension
		{
			var Index=URL.lastIndexOf(".");
		
			this.FileExtension=URL.substr(Index+1);
		}
		
		// strip off the z,x,y 
		
		var XIndex=URL.indexOf("{x}");
		var YIndex=URL.indexOf("{y}");
		var ZIndex=URL.indexOf("{z}");
		
		var StartIndex=ZIndex;
		
		this.CoordinateValueOrder=[];
		
		if (XIndex<YIndex) // is before y (x is first or second)
		{
			if (XIndex<ZIndex) // x is before z (x is first)
			{
				StartIndex=XIndex;
				
				if (YIndex<ZIndex) this.CoordinateValueOrder=["x","y","z"];
				else this.CoordinateValueOrder=["x","z","y"];
			}
			else // (x is before y and after z)
			{
				this.CoordinateValueOrder=["z","x","y"];
			}
		}
		else // x is after y
		{
			if (XIndex<ZIndex) // x is after y and before z
			{
				this.CoordinateValueOrder=["y","x","z"];
			}
			else // x is after z (x is at the end)
			{
				if (YIndex<ZIndex) 
				{
					StartIndex=YIndex;
				
					this.CoordinateValueOrder=["y","z","x"];
				}
				else this.CoordinateValueOrder=["z","y","x"];
			}
		}
		
		// strip off everything after the first brace
		
		URL=URL.substr(0,StartIndex);
	
		// save the information
		
		this.URL=URL;
	
		//
	
		this.TopTile=new CMTileOpenFormat(this,-18,0,0);
	
		this.GetParent(CMScene).Repaint();
	}
}
CMDatasetPyramidOpenFormat.prototype.GetCoordinateValueOrder=function() 
{
	return(this.CoordinateValueOrder);
}
//******************************************************************
// CMDatasetPyramid Functions
//******************************************************************
CMDatasetPyramidOpenFormat.prototype.GetRasterFileExtension=function() 
{
	return(this.FileExtension);
}

CMDatasetPyramidOpenFormat.prototype.GetBounds=function() 
{
	return(this.TheBounds);
}

/**
* Making things visible is handled in paint but child tiles must be 
* make invisible through SetVisible()
* @protected
*/
CMDatasetPyramidOpenFormat.prototype.SetVisible=function(TheLayer,Flag) 
{ 
	// we only have to make the files hidden when they are visible
	// they will be made visible on a repaint if hidden and then switched to visible
	if ((this.Visible)&&(Flag==false)&&(this.TopTile!=null))
	{
		this.TopTile.HideTiles(TheLayer,true);
		
		if ((this.MeshIsInGeo))
		{
			var TheGeo=TheLayer.GetParent(CMGeo);
			TheGeo.RemoveOGLObjectFromGeo(this.TheMesh);
			TheGeo.RemoveOGLObjectFromGeo(this.TopMesh);
			TheGeo.RemoveOGLObjectFromGeo(this.BottomMesh);
			this.MeshIsInGeo=false;
		}
	}
	this.Visible=Flag;
}

//******************************************************************
// Painting
//******************************************************************

/*
* Paints a layer into the canvas
*/
CMDatasetPyramidOpenFormat.prototype.Paint=function(TheLayer,TheView,SelectedOnly,MouseOverOnly) 
{
	if ((SelectedOnly==false)&&((MouseOverOnly==undefined)||(MouseOverOnly==false)))
	{
		if ((typeof(CM3View)!="undefined")&&(TheView instanceof CM3View)) // have a 3d view
		{
			if ((TheLayer.GetVisible())&&(this.MeshIsInGeo==false))
			{
				var TheGeo=TheLayer.GetParent(CMGeo);
				var TheProjector=TheGeo.GetProjector();
				
				if ((TheProjector!=null)&&(TheProjector instanceof CM3ProjectorSpherical))
				{
					var NorthPoleColor=TheLayer.GetSetting("Dataset","NorthPoleColor");
					var SouthPoleColor=TheLayer.GetSetting("Dataset","SouthPoleColor");
					
					var points = [];
					
					if ((NorthPoleColor!=undefined)||(SouthPoleColor!=undefined))
					{
						for ( var i = 83; i <=90; i ++ ) 
						{
							var Coordinate=TheGeo.ProjectFromGeographic(0, i, 0);
							
							var Vector2=new THREE.Vector2(Coordinate[0],Coordinate[2]);
							
							points.push( Vector2 );
						}
					}
					if (NorthPoleColor!=undefined)
					{
						var geometry = new THREE.LatheBufferGeometry( points );
						var material = new THREE.MeshBasicMaterial( { color: NorthPoleColor } );
						this.TopMesh = new THREE.Mesh( geometry, material );
						
						//var Coordinate=TheGeo.ProjectFromGeographic(0, 90, 0);
						this.TopMesh.rotateX(Math.PI/2);
						
						TheGeo.AddOGLObjectToGeo(this.TopMesh);
					}
					if (SouthPoleColor!=undefined)
					{
						var material = new THREE.MeshBasicMaterial( { color: SouthPoleColor } );
						this.BottomMesh = new THREE.Mesh( geometry, material );
						this.BottomMesh.rotateX(-Math.PI/2);
						
						TheGeo.AddOGLObjectToGeo(this.BottomMesh);
					}
					this.MeshIsInGeo=true;
				}
			}
		}
		this.TopTile.PaintTile(TheLayer,TheView,SelectedOnly,MouseOverOnly);
	}
}


