/******************************************************************************************************************
* CMDatasetVector
* This is effecitvely an abstract class that provides an API between
* the layer class and the data that may be shared between muliptle layers.
* The data typically contains individual spatial features with attributes
* But can also contain raster data organized as tiles.
*
* @module CMDatasetVector
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/
//******************************************************************
// Global Definitions
//******************************************************************

CMDatasetVector.TYPE_POINTS=1;
CMDatasetVector.TYPE_POLYLINES=2;
CMDatasetVector.TYPE_POLYGONS=3;

//******************************************************************
// CMDatasetVector Constructor
//******************************************************************

//******************************************************************
// CMDatasetVector Constructor
//******************************************************************
/*
* Constructor
* @protected, @constructs
*/
function CMDatasetVector() 
{
	CMDataset.call(this);

	// properties
	this.TheRegions=null; // standard regions
	this.Type=null; // type of primitives in the regions
	
	// attributes
	this.ColumnHeadings=[];
	this.Columns=[];
}
CMDatasetVector.prototype=Object.create(CMDataset.prototype); // inherit prototype functions from PanelBase()

CMDatasetVector.prototype.contructor=CMDatasetVector; // override the constructor to go to ours

//**************************************************************
// Prviate functions
//**************************************************************

/**
* Get the boounds of the specified data in GeoJSON Format
* @private
* @param TheData - An object containing GeoJSON data
* @returns - An object containining: xMin,xMax,yMin,yMax
*/
CMDatasetVector.GetBoundingBoxFromRegion=function(TheRegion) 
{
	var TheFeatureBounds=null;

	for (var AreaIndex=0; AreaIndex < TheRegion.length; AreaIndex++) 
	{
		var TheArea=TheRegion[AreaIndex];
		
		var TheExterior=TheArea[0];
		
		var Xs=TheExterior[0];
		var Ys=TheExterior[1];
		
		for (var i=0;i<Xs.length;i++)
		{
			if (TheFeatureBounds==null)
			{
				TheFeatureBounds={
					XMin:Xs[i],
					XMax:Xs[i],
					YMin:Ys[i],
					YMax:Ys[i]
				}
			}
			else
			{
				// Update the TheBounds recursively by comparing the current
				// xMin/xMax and yMin/yMax with the coordinate 
				// we're currently checking
				if (TheFeatureBounds.XMin > Xs[i]) TheFeatureBounds.XMin =Xs[i];
				if (TheFeatureBounds.XMax < Xs[i]) TheFeatureBounds.XMax=Xs[i];
				if (TheFeatureBounds.YMin >Ys[i]) TheFeatureBounds.YMin =Ys[i];
				if (TheFeatureBounds.YMax < Ys[i]) TheFeatureBounds.YMax =Ys[i];
			}
		}
	}
	return(TheFeatureBounds);
}
CMDatasetVector.prototype.SetupBoundingBoxesFromRegions=function(TheRegions) 
{
	var TheBounds=null;
	this.FeatureBounds=[];
	
	if (TheRegions!=null)
	{
		// Loop through each region
		for (var RegionIndex=0; RegionIndex < TheRegions.length; RegionIndex++) 
		{
			var TheRegion=TheRegions[RegionIndex];
			
			var TheFeatureBounds=CMDatasetVector.GetBoundingBoxFromRegion(TheRegion);
			
			if (TheBounds==null) 
			{
				TheBounds=CMUtilities.CloneBounds(TheFeatureBounds);
			}
			else
			{
				CMUtilities.AddToBounds(TheBounds,TheFeatureBounds);
			}
			this.FeatureBounds[RegionIndex]=TheFeatureBounds;
		}
		this.TheBounds=TheBounds;
	}
}
//**************************************************************
// Protected CMDatasetVector Functions for attributes
//**************************************************************
/**
* @protected
*/
CMDatasetVector.prototype.SetAttributes=function(NewColumnHeadings,NewColumns) 
{ 
	this.ColumnHeadings=NewColumnHeadings;
	this.Columns=NewColumns;
	this.SendMessageToListeners(CMDataset.MESSAGE_ATTRIBUTES_CHANGED);
}
//**************************************************************
// public Functions for Bounds
//**************************************************************

/**
* Check if the feature is visible in the view.
* This should be called by subclasses but can also be called to limit a layer's bounds after loading.
* @public
* @param NewBounds - Bounds with format {XMin,XMax,YMin,YMax}
*/
CMDatasetVector.prototype.SetBounds=function(NewBounds) 
{
	this.TheBounds=NewBounds;
}
/**
* Returns the bounds of the data within the layer.  Computed after loading the data.
* @public
* @returns Bounds - with format {XMin,XMax,YMin,YMax}
*/
CMDatasetVector.prototype.GetBounds=function() 
{
	return(this.TheBounds);
}
CMDatasetVector.prototype.GetFeatureBounds=function(FeatgureIndex) 
{
	return(this.FeatureBounds[FeatgureIndex]);
}
//**************************************************************
// CMData Functions for attributes
//**************************************************************

CMDatasetVector.prototype.GetNumAttributeRows=function() 
{ 
	var Result=0;
	
	if (this.Columns.length>0) Result=this.Columns[0].length;
	
	return(Result); 
}

CMDatasetVector.prototype.GetNumAttributeColumns=function() 
{ 
	return(this.ColumnHeadings.length); 
}

CMDatasetVector.prototype.GetAttributeHeading=function(ColumnIndex) 
{ 
	return(this.ColumnHeadings[ColumnIndex]); 
}

CMDatasetVector.prototype.GetAttributeHeadings=function(ColumnIndex) 
{ 
	return(this.ColumnHeadings); 
}

CMDatasetVector.prototype.GetAttributeCell=function(ColumnIndex,RowIndex) 
{ 
	return(this.Columns[ColumnIndex][RowIndex]); 
}

CMDatasetVector.prototype.AddAttributeHeading=function(NewHeading,DefaultValue) 
{ 
	this.ColumnHeadings.push(NewHeading);
	
	var TheColumn=[];
	
	var NumRows=this.GetNumAttributeRows();
	
	for (var i=0;i<NumRows;i++) TheColumn[i]=DefaultValue;
	
	this.Columns.push(TheColumn);
}


CMDatasetVector.prototype.GetAttributeCellByHeading=function(Heading,RowIndex)  //jjg
{ 
	var Result=null;
	
	var ColumnIndex=this.ColumnHeadings.indexOf(Heading);
	
	if (ColumnIndex==-1) 
	{
		throw("Sorry, the heading "+Heading+" was not found");
	}
	else
	{
		Result=this.GetAttributeCell(ColumnIndex,RowIndex);
	}
	return(Result);
}
CMDatasetVector.prototype.GetAttributeIndexFromHeading=function(Heading) 
{ 
	var ColumnIndex=this.ColumnHeadings.indexOf(Heading);
	
	return(ColumnIndex); 
}

CMDatasetVector.prototype.SetAttributeCell=function(ColumnIndex,RowIndex,Value) 
{ "use strict";
	this.Columns[ColumnIndex][RowIndex]=Value;
	
	this.SendMessageToListeners(CMDataset.MESSAGE_ATTRIBUTES_CHANGED);
};
/*
CMDatasetVector.prototype.SetAttributeCellByHeading=function(Heading,RowIndex,Value)  // jjg dgoing away
{ 
	var ColumnIndex=this.ColumnHeadings.indexOf(Heading);
	
	var Result=this.SetAttributeCell(ColumnIndex,RowIndex,Value);
};*/
//******************************************************************
// CMData Mouse event handling
//******************************************************************
CMDatasetVector.prototype.GetNumFeatures=function() 
{
	var Result=0;
	if (this.TheRegions!=null) Result=this.TheRegions.length;
	return(Result);
}

CMDatasetVector.prototype.In=function(TheView,RefX,RefY,RefTolerance) 
{
	var FeatureIndex=-1;
	
	if ((this.TheRegions!=null)) 
	{
		// Loop over the features
		for (var i=0;( i < this.TheRegions.length)&&(FeatureIndex==-1); i++) 
		{
			var Result=this.InFeature(TheView,RefX,RefY,i,RefTolerance);
			
			if (Result) FeatureIndex=i;
		}
	}
	return(FeatureIndex);
}
/**
* Checks if the specified coordinate is in the specified feature
*/
CMDatasetVector.prototype.InFeature=function(TheView,RefX,RefY,FeatureIndex,RefTolerance) 
{
	var Result=false;
	
	if (this.TheRegions!=null) 
	{
		//var RefTolerance=TheView.GetRefWidthFromPixelWidth(this.ClickTolerance);
		
		switch (this.Type)
		{
		case CMDatasetVector.TYPE_POINTS:
			{
				var TheRegion=this.TheRegions[FeatureIndex];
				
				for (var i=0;(i<TheRegion.length)&&(Result==false);i++)
				{
					var TheArea=TheRegion[i];
					
					var ThePoly=TheArea[0];
					
					var X=ThePoly[0][0];
					var Y=ThePoly[1][0];
					
					if ((Math.abs(X-RefX)<=RefTolerance)&&(Math.abs(Y-RefY)<=RefTolerance))
					{
						Result=true;
					}
				}
			}
			break;
		case CMDatasetVector.TYPE_POLYLINES:
			{
				var TheRegion=this.TheRegions[FeatureIndex];
				
				for (var i=0;(i<TheRegion.length)&&(Result==false);i++)
				{
					var TheArea=TheRegion[i];
					
					var ThePoly=TheArea[0];
					
					var Xs=ThePoly[0];
					var Ys=ThePoly[1];
					
					Result=CMUtilities.InPolyline(RefX,RefY,Xs,Ys,RefTolerance);
				}
			}
			break;
			
		case CMDatasetVector.TYPE_POLYGONS:
			{
				var TheRegion=this.TheRegions[FeatureIndex];
				
				for (var i=0;(i<TheRegion.length)&&(Result==false);i++)
				{
					var TheArea=TheRegion[i];
					
					var ThePoly=TheArea[0];
					
					var Xs=ThePoly[0];
					var Ys=ThePoly[1];
					
					Result=CMUtilities.InsideAPolygon(RefX,RefY,Xs,Ys,Xs.length);
				}
			}
			break;
		}
	}
	return(Result);
};

//******************************************************************
// CMData Painting Functions
//******************************************************************
/*
* Paints the specified data into the view.
* 
*/
CMDatasetVector.prototype.Paint=function(TheLayer,TheView,SelectedOnly,MouseOverOnly) 
{
	if ((TheView instanceof CMView2D)&&(this.TheRegions!=null)) //(this.GetVisible())&&
	{
		if (MouseOverOnly==undefined) MouseOverOnly=false;
		
		var ViewBounds=TheView.GetBounds();
			
		// draw each feature
		for (var FeatureIndex=0; FeatureIndex < this.TheRegions.length; FeatureIndex++) 
		{
			// jjg - not the most efficient approach
			var Draw=false; 
			if ((SelectedOnly==false)&&(MouseOverOnly==false))
			{
				Draw=true;
			}
			else if ((SelectedOnly)&&(this.SelectedFeature==FeatureIndex)) 
			{
				Draw=true;
			}
			else if ((MouseOverOnly)&&(this.MouseOverFeatureIndex==FeatureIndex)) 
			{
				Draw=true;
			}
			if (Draw)
			{
				var TheFeatureBounds=this.GetFeatureBounds(FeatureIndex);
				
				var TheGeo=TheLayer.GetParent(CMGeo);
				
				if ((CMUtilities.BoundsOverlap(ViewBounds,TheFeatureBounds))||
					(TheGeo.GetProjectorType()==CMGeo.PROJECTOR_DYNAMIC)) // jjg - needs to support dynamic projection
				{
					var Result=false;
					if (this.Type==CMDatasetVector.TYPE_POINTS) // draw a point as an icon or a mark
					{
						var TheRegion=this.TheRegions[FeatureIndex];
						
						var TheArea=TheRegion[0];
						
						var ThePoly=TheArea[0];
						
						TheLayer.PaintPoint(TheView,FeatureIndex,ThePoly[0][0],ThePoly[1][0],SelectedOnly,MouseOverOnly);
					}
					else // draw a polyline or polygon
					{
						if (this.TheRegions!=null)
						{
							var TheRegion=this.TheRegions[FeatureIndex];
							
							for (var i=0;i<TheRegion.length;i++)
							{
								var TheArea=TheRegion[i];
								
								TheLayer.PaintRefArea(TheView,FeatureIndex,TheArea,this.Type,SelectedOnly,MouseOverOnly);
							}
						}
					}
				}
			}
		}
	}
}

//******************************************************************
// CMLayer Searching Functions
//******************************************************************
/*
* Requests search results from a layer.  The scene calls this function.
*/
CMDatasetVector.prototype.SearchColumn=function(ColumnIndex,SearchPhrase) 
{
	var Result=[];
	
	if ((this.Columns!=null)) //(this.GetVisible())&&
	{
		SearchPhrase=SearchPhrase.toLowerCase();
		
		var TheColumn=this.Columns[ColumnIndex];
		
		for (var j=0;j<TheColumn.length;j++)
		{
			var TheProperty=TheColumn[j];
			
			if (typeof(TheProperty)!="string") TheProperty=""+TheProperty;
			
			TheProperty=TheProperty.toLowerCase();
			
			if (TheProperty.indexOf(SearchPhrase)!=-1)
			{
				Result.push(j);
			}
		}
	}
	return(Result);
}

/*
* Requests search results from a layer.  The scene calls this function.
*/
CMDatasetVector.prototype.GetSearchResults=function(SearchPhrase,ResultsPanel) 
{
	if ((this.Columns!=null)) //(this.GetVisible())&&
	{
		// draw each feature
		for (var i=0; i < this.Columns.length; i++) 
		{
			var TheColumn=this.Columns[i];
			
			for (var j=0;j<TheColumn.length;j++)
			{
				var TheProperty=TheColumn[j];
				
				if (typeof(TheProperty)=="string")
				{
					TheProperty=TheProperty.toLowerCase();
					
					if (TheProperty.indexOf(SearchPhrase)!=-1)
					{
						var ThisResult=document.createElement("DIV");
						ThisResult.innerHTML=TheProperty;
						ThisResult.className="CM_SearchResult";
						
						ThisResult.TheDataset=this;
						ThisResult.FeatureIndex=j;
						
						// setup function called when user clicks on search result
						ThisResult.onclick=function()
						{
							var TheScene=this.TheDataset.GetParent(CMScene);
							TheScene.UnselectAll();
							this.TheDataset.SetSelectedFeature(this.FeatureIndex);
							
							var TheBounds=this.TheDataset.FeatureBounds[this.FeatureIndex];
							var TheView=TheScene.GetView(0);
							
							//this.className="CM_SearchResultSelected";
							
							TheView.ZoomToBounds(TheBounds);
						}
						ResultsPanel.appendChild(ThisResult);
						
						Found=true;
					}
				}
			}
		}
	}
}

CMDatasetVector.prototype.CMDataset_GetIcon=CMDataset.prototype.GetIcon;

CMDatasetVector.prototype.GetIcon=function(TheLayer,Default)
{
	var TheIcon=this.CMDataset_GetIcon(TheLayer,Default);
	
	if ((this.TheRegions!=null)) //
	{
		switch (this.Type)
		{
		case CMDatasetVector.TYPE_POLYLINES:
			{
				TheIcon=document.createElement('CANVAS');
				TheIcon.className="CM_LayerListIconClass";
				TheIcon.TheLayer=this;
				TheIcon.style.borderColor="rgba(0,0,0,0)";
				
				TheIcon.width=16;
				TheIcon.height=16;
				
				var TheView=new CMView2D();
				TheView.Setup(TheIcon);
					
				var StrokeStyle=TheLayer.GetSetting("Style","strokeStyle");
				var LineWidth=TheLayer.GetSetting("Style","lineWidth");
	
				var TheContext = TheIcon.getContext("2d");
				TheContext.strokeStyle=StrokeStyle;
				TheContext.lineWidth=LineWidth;
				
				TheContext.moveTo(0,0);
				TheContext.lineTo(16,16);
				TheContext.moveTo(0,16);
				TheContext.lineTo(7,7);
				TheContext.stroke();
			}
			break;
		case CMDatasetVector.TYPE_POINTS:
			{
				var TheType=TheLayer.GetSetting("Mark","Type",undefined);
		
				if (TheType==undefined) TheType=CMLayer.MARK_CIRCLE;
				
				var StrokeStyle=TheLayer.GetSetting("Style","strokeStyle");
				var FillStyle=TheLayer.GetSetting("Style","fillStyle");
				TheIcon=TheLayer.GetMarkIcon(TheType,FillStyle,StrokeStyle);
			}
			break;
		}
	}
	return(TheIcon);
}

//******************************************************************
// Public CMDatasetVector functions
//******************************************************************

/**
* Add a new point to the data
* @public
*/
CMDatasetVector.prototype.AddPoint=function(X,Y,Z)
{
	var Projector=this.GetProjector();
	
	if (Projector!=null)
	{
		var Result=Projector.ProjectFromGeographic(X,Y);
		X=Result[0];
		Y=Result[1];
	}

	if (this.TheRegions==null) 
	{
		this.TheRegions=[];
		this.Type=CMDatasetVector.TYPE_POINTS;
	}
	var ThePoly=[[X],[Y]];
	if (Z!=undefined) ThePoly.push([Z]);
	
	// create the region
	var TheArea=[ThePoly];
	var TheRegion=[TheArea];
	this.TheRegions.push(TheRegion);
	
	this.SetupBoundingBoxesFromRegions(this.TheRegions);

}

CMDatasetVector.prototype.AddPolyline=function(Xs,Ys,Zs)
{
	this.AddPoly(CMDatasetVector.TYPE_POLYLINES,Xs,Ys,Zs);
}

CMDatasetVector.prototype.AddPolygon=function(Xs,Ys,Zs)
{
	this.AddPoly(CMDatasetVector.TYPE_POLYGONS,Xs,Ys,Zs);
}

/**
* Add a new polyline to the data
* @public
*/
CMDatasetVector.prototype.AddPoly=function(Type,Xs,Ys,Zs)
{
	var Projector=this.GetProjector();
	
	if (Projector!=null)
	{
		var Result=Projector.ProjectFromGeographic(Xs,Ys);
		Xs=Result[0];
		Ys=Result[1];
	}

	if (this.TheRegions==null) 
	{
		this.TheRegions=[];
		this.Type=Type;
	}
	var ThePoly=[Xs,Ys];
	if (Zs!=undefined) ThePoly.push(Zs);
	
	var TheArea=[ThePoly];
	var TheRegion=[TheArea];
	this.TheRegions.push(TheRegion);
	
	this.SetupBoundingBoxesFromRegions(this.TheRegions);
}
/**
* Add a new point to the data
* @public
*/
CMDatasetVector.prototype.SetPosition=function(RegionIndex,AreaIndex,PolyIndex,CoordinateIndex,Position)
{
	var ThePoly=this.TheRegions[RegionIndex][AreaIndex][PolyIndex];
	
	ThePoly[0][CoordinateIndex]=Position.x;
	ThePoly[1][CoordinateIndex]=Position.y;
	if (ThePoly.length>2) ThePoly[2][CoordinateIndex]=Position.z;
	
	this.SetupBoundingBoxesFromRegions(this.TheRegions);
}

//**************************************************************
// Public CMDatasetVector Functions for attributes
//**************************************************************
/*
* Set the vector data from a GeoJSON object
*/
CMDatasetVector.prototype.GetRegions=function() 
{
	return(this.TheRegions);
}
CMDatasetVector.prototype.GetType=function() 
{
	return(this.Type);
}
