/******************************************************************************************************************
* CMGeo
* Contains a list of layers and a list of background layers.
* Also has a background style that will be drawn if set.
*
* @module CMGeo
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/
//******************************************************************
// Global definitions
//******************************************************************

CMGeo.NumGeos=0;

//CMGeo.PROJECTOR_NONE=0; // no projection
//CMGeo.PROJECTOR_COORDINATES=1; // data is already projected, just use projector to find geographic coordinates
//CMGeo.PROJECTOR_LOAD=1; // project data from geographic to projected system on load
CMGeo.PROJECTOR_DYNAMIC=2; // projection can change at any time and projection chnaged messages should be sent

//******************************************************************
// CMScene Class
//******************************************************************
/*
* Constructor
* @public, @constructs
*/
function CMGeo(TheScene) 
{
	CMItem.call(this);

	this.SetName("Geo"+CMGeo.NumGeos);
	CMGeo.NumGeos++;
	
	this.SetParent(TheScene);
	
	// contained objects
 	this.Layers=[]; // this is the Geo's children (jjg - move to children)
	
	this.Backgrounds=[]; // (jjg - move to children)
	this.SelectedBackgroundIndex=-1;
	
	this.TheProjector=null;
	this.ProjectorType; // undefined for default
	
	this.MinTime=0; 
	
	this.SetSetting("Style","fillStyle","rgb(210,220,255)"); // only works if there is a projection
	
	// private properties
	this.Times=[0];
	
	this.TheBounds=null;

	this.NumRepaintBlocks=0;
	this.NeedRepaint=false;
}

CMGeo.prototype=Object.create(CMItem.prototype); // inherit prototype functions from PanelBase()

CMGeo.prototype.contructor=CMGeo; // override the constructor to go to ours

//******************************************************************
// CMBase Functions
//******************************************************************


CMGeo.prototype.GetTimes=function(TheTimeSlices) 
{
	return(this.Times);
}
//

CMGeo.prototype.GetNumChildren=function(TheClass)  
{ 
	var Result=0;

	if (this.Layers!=undefined) Result=this.Layers.length;
	
	return(Result);
}

CMGeo.prototype.GetChild=function(Index,TheClass)  
{ 
	return(this.Layers[Index]);
}
/*
* Geos send event messages to their layers in the opposite of the order they are added.
*/
CMGeo.prototype.SendMessageToDescendants=function(Message,AdditionalInfo)  
{ 
	var Used=false;
	
	var NumChildren=this.GetNumChildren();
	
	for (var i=NumChildren-1;(i>=0)&&(Used==false);i--)
	{
		var TheChild=this.GetChild(i);
		
		Used=TheChild.SendMessageToDescendants(Message,AdditionalInfo);
	}
	return(Used);
}

//******************************************************************
// CMGeo Functions
//******************************************************************
//******************************************************************
// Background Management
//******************************************************************
/*
* Adds a background layer to the scene
* @public
* @param TheLayer
* @returns LayerIndex 
*/
CMGeo.prototype.AddBackground=function(TheLayer) 
{
	TheLayer.SetParent(this);
	 
	var LayerIndex=this.Backgrounds.length;
	
	this.Backgrounds[LayerIndex]=TheLayer;
	
	if (this.SelectedBackgroundIndex==-1) this.SetSelectedBackgroundIndex(0);
	
	this.GetParent(CMScene).BackgroundListChanged();
	
	return(LayerIndex);
}
/*
* Sets which background will be selected for painting
* @public
* @param BackgroundIndex - 
*/
CMGeo.prototype.SetSelectedBackgroundIndex=function(New)
{
	if (New!=this.SelectedBackgroundIndex) // something has changed
	{
		if (this.SelectedBackgroundIndex!=-1) // make the current background layer invisible
		{
			this.Backgrounds[this.SelectedBackgroundIndex].SetVisible(false);
		}
		// set the new selected background layer (could be -1 for none)
		this.SelectedBackgroundIndex=New;
		
		if (New!=-1) // make the new layer visible
		{
			this.Backgrounds[New].SetVisible(true);
		}
		// notify the scene that the backgrounds have changed
		this.GetParent(CMScene).BackgroundListChanged();
		this.Repaint();
	}
}
CMGeo.prototype.GetNumBackgrounds=function()
{
	return(this.Backgrounds.length);
}
CMGeo.prototype.GetBackground=function(Index)
{
	return(this.Backgrounds[Index]);
}


//******************************************************************
// public CMScene functions for layer Management
//******************************************************************
/*
* Adds a layer to the scene
* @public
* @param TheLayer
* @returns LayerIndex 
*/
CMGeo.prototype.AddLayer=function(TheLayer) 
{
	TheLayer.SetParent(this);
	
	var LayerIndex=this.Layers.length;
	
	this.Layers[LayerIndex]=TheLayer;
	
	// make sure the layer's time slices are reprented
	
	var NewTimes=TheLayer.GetTimes(this.Times);
	
	for (Time in NewTimes)
	{
		var Index=this.Times.indexOf(Time);
		
		if (Index==-1) // time slice is not in the time slice array
		{
			this.InsertTime(Time);
		}
	}
	this.GetParent(CMScene).LayerListChanged();
	
	return(LayerIndex);
}
/*
* Return the index for the specified layer.  If the 
* layer does not appear in the CMScene, return -1
* @public
* @param TheLayer - 
* @returns LayerIndex
*/
CMGeo.prototype.GetLayerIndex=function(TheLayer) 
{
	var Result=-1;
	
	for (var i=0;i<this.Layers.length;i++)
	{
		if (this.Layers[i]==TheLayer) Result=i;
	}
	return(Result);
}
/*
* Returns the layer at the specified index.
* @public
* @param Index - 
* @returns TheLayer
*/
CMGeo.prototype.GetLayer=function(Index) 
{
	var Result=this.Layers[Index];
	
	return(Result);
}
/*
* Swap the layer with the one above it.
* @public
* @param Index - 
*/
CMGeo.prototype.MoveLayerUp=function(Index) 
{
	var TheLayer=this.Layers[Index];
	if (Index>=0)
	{
		this.Layers[Index]=this.Layers[Index-1];
		this.Layers[Index-1]=TheLayer;
	}
	this.GetParent(CMScene).LayerListChanged();
	this.Repaint();
}
/*
* Swap the specified layer with the one below it.
* @public
* @param Index - 
*/
CMGeo.prototype.MoveLayerDown=function(Index) 
{
	var TheLayer=this.Layers[Index];
	if (Index<this.Layers.length)
	{
		this.Layers[Index]=this.Layers[Index+1];
		this.Layers[Index+1]=TheLayer;
	}
	this.GetParent(CMScene).LayerListChanged();
	this.Repaint();
}

/**
* Remove the layer at the specified index from the list of layers
* @public
* @param Index - Index to the layer to delete
*/
CMGeo.prototype.DeleteLayer=function(Index) 
{
	var Result=this.Layers.splice(Index,1);

	this.GetParent(CMScene).LayerListChanged();
	this.Repaint();
	
	return(Result);
}
/**
* Moves a layer to a new position in the layer list 
* @public
* @param TheLayer - The layer to move
* @param NewIndex - Index to the new position for the layer
*/
CMGeo.prototype.MoveLayer=function(TheLayer,NewIndex) 
{
	var CurrentIndex=this.GetLayerIndex(TheLayer);
	
	// make sure the new index is still within the list of layers
	
	if (NewIndex<0) NewIndex=0;
	if (NewIndex>=this.Layers.length) NewIndex=this.Layers.length-1;
	
	if (NewIndex<CurrentIndex) // move the layer up in the list (shift the layers from the new index to the old down one layer
	{
		for (var i=CurrentIndex;i>NewIndex;i--)
		{
			this.Layers[i]=this.Layers[i-1];
		}
		this.Layers[NewIndex]=TheLayer;
	}
	else if (NewIndex>CurrentIndex) // move the layer down in the list
	{
		for (var i=CurrentIndex;i<NewIndex;i++)
		{
			this.Layers[i]=this.Layers[i+1];
		}
		this.Layers[NewIndex]=TheLayer;
	}
	this.GetParent(CMScene).LayerListChanged();
	
	this.Repaint();
}
//******************************************************************
// Search
//******************************************************************
/**
* Get the search results from each layer.  The layers will insert 
* elements into the SearchResults element
*
* Called by CanvasMap
* @private
*/
CMGeo.prototype.GetSearchResults=function(TheText,SearchResults) 
{
	var NumLayers=this.GetNumChildren(CMLayer);
	
	for (var i=0;i<NumLayers;i++)
	{
		var TheLayer=this.GetChild(i,CMLayer);
		
		TheLayer.GetSearchResults(TheText,SearchResults);
	}
}

//******************************************************************
// Search
//******************************************************************
/**
* Get the search results from each layer.  The layers will insert 
* elements into the SearchResults element
*
* Called by CanvasMap
* @protected
*/
CMGeo.prototype.ProjectionChanged=function() 
{
	var NumLayers=this.GetNumChildren(CMLayer);
	
	for (var i=0;i<NumLayers;i++)
	{
		var TheLayer=this.GetChild(i,CMLayer);
		
		TheLayer.ProjectionChanged();
	}
}

//******************************************************************
// Event handlers
//******************************************************************
/*
* Not used
*/
CMGeo.prototype.In=function(TheView,RefX,RefY,TheEvent) 
{
	var Result=null;
	
	// in operates the opposite direction as painting
	
	var NumLayers=this.GetNumChildren(CMLayer);
	
	for (var i=NumLayers-1;(i>=0)&&(Result==null);i--)
	{
		var TheLayer=this.GetChild(i);
		
		var FeatureIndex=TheLayer.In(TheView,RefX,RefY,TheEvent);
		
		if (FeatureIndex!=-1)
		{
			Result=
			{
				LayerIndex: i,
				FeatureIndex: FeatureIndex
			}
		}
	}
	return(Result);
};
/**
* Called when there is a mouse down event received by CanvasMap
* @public
* @override
* @param TheView - the view that recieved the event
* @param RefX - reference horiziontal coordinate value
* @param RefY - reference vertical coordinate value
*//*
CMGeo.prototype.MouseDown=function(TheView,RefX,RefY,TheEvent) 
{
	var Used=false;
	
	var NumLayers=this.GetNumChildren(CMLayer);
	
	for (var i=NumLayers-1;(i>=0)&&(Used==false);i--)
	{
		var TheLayer=this.GetChild(i);
		
		Used=TheLayer.MouseDown(TheView,RefX,RefY,TheEvent);
	}
	return(Used);
};*/
/**
* Called when there is a mouse move event received by CanvasMap
* @public
* @override
* @param TheView - the view that recieved the event
* @param RefX - reference horiziontal coordinate value
* @param RefY - reference vertical coordinate value
*//*
CMGeo.prototype.MouseMove=function(TheView,RefX,RefY,TheEvent) 
{
	var Used=false;
	
	var NumLayers=this.GetNumChildren(CMLayer);
	
	for (var i=NumLayers-1;(i>=0)&&(Used==false);i--)
	{
		var TheLayer=this.GetChild(i);
		
		Used=TheLayer.MouseMove(TheView,RefX,RefY,TheEvent);
	}
	return(Used);
};*/
/**
* Called when there is a mouse up event received by CanvasMap
* @public
* @override
* @param TheView - the view that recieved the event
* @param RefX - reference horiziontal coordinate value
* @param RefY - reference vertical coordinate value
*//*
CMGeo.prototype.MouseUp=function(TheView,RefX,RefY,TheEvent) 
{
	var Used=false;
	
	var NumLayers=this.GetNumChildren(CMLayer);
	
	for (var i=NumLayers-1;(i>=0)&&(Used==false);i--)
	{
		var TheLayer=this.GetChild(i);
		
		Used=TheLayer.MouseUp(TheView,RefX,RefY,TheEvent);
	}
	return(Used);
};*/

//******************************************************************
// Painting
//******************************************************************
/**
* Called when the window is resized
* @public
* @override
* @param TheView - the view that recieved the event
*/
CMGeo.prototype.Resize=function(TheView) 
{
	var NumLayers=this.GetNumChildren(CMLayer);
	
	for (var i=0;i<NumLayers;i++)
	{
		var TheLayer=this.GetChild(i);
		
		TheLayer.Resize(TheView);
	}
	this.Repaint();
}
/**
* Paint the layers into the view
* @public
* @override
* @param TheView - the view that recieved the event
*/
CMGeo.prototype.Paint=function(TheView) 
{
//	if (this.NumRepaintBlocks==0)
	{
		// clear the background
		if ((typeof(CM3View)=="undefined")||((TheView instanceof CM3View)==false))
		{
			var Style=this.GetStyle(TheView,0,"Style");
			
			if (Style!=undefined)
			{
				TheView.SetStyle(Style);
				
				if (this.TheProjector!=null)
				{
					var RegionToFill=this.TheProjector.GetRegionToFill();
					var RegionToDraw=this.TheProjector.GetRegionToDraw();
					
					TheView.PaintRefRegion(RegionToFill,false,true,false); // area, closed, fill, no stroke
					TheView.PaintRefRegion(RegionToDraw,false,false,true); // area, closed, no fill, stroke
				}
				TheView.RestoreStyle();
			}
		}
		// draw the backgrounds first
		
		if (this.SelectedBackgroundIndex!=-1)
		{
			this.Backgrounds[this.SelectedBackgroundIndex].Paint(TheView);
		}
	
		// then the other layers
		
		var NumLayers=this.GetNumChildren(CMLayer);
	
		for (var i=0;i<NumLayers;i++)
		{
			var TheLayer=this.GetChild(i);
		
			TheLayer.Paint(TheView);
		}
		
		// then the selected features
		
		for (var i=0;i<NumLayers;i++)
		{
			var TheLayer=this.GetChild(i);
		
			TheLayer.PaintSelected(TheView);
		}
	}
}

		
//******************************************************************
// Additional public functions
//******************************************************************
/**
* Set the projection for coordinate conversion
* @public
* @param TheProjector - the projector to use to project coordinates
*/
CMGeo.prototype.SetProjector=function(TheProjector) 
{ 
	this.TheProjector=TheProjector; 
	
	TheProjector.AddListener(CMBase.MESSAGE_SETTINGS_CHANGED,this,function(TheDataset,This,AdditionalInfo)
	{
		var NumLayers=This.GetNumChildren(CMLayer);
		
		for (var i=0;i<NumLayers;i++)
		{
			var TheLayer=This.GetChild(i);
		
			TheLayer.ProjectorChanged();
		}
	});
	// jjg - should be a message
	var NumLayers=this.GetNumChildren(CMLayer);
	
	for (var i=0;i<NumLayers;i++)
	{
		var TheLayer=this.GetChild(i);
	
		TheLayer.ProjectorChanged();
	}
};
	
CMGeo.prototype.GetProjector=function() 
{ 
	return(this.TheProjector); 
};

/**
* Set the projector type to be NONE, COORDINATES, 
* @public
* @param TheProjector - the projector to use to project coordinates
*/
CMGeo.prototype.SetProjectorType=function(ProjectorType) 
{ 
	this.ProjectorType=ProjectorType; 
};
CMGeo.prototype.GetProjectorType=function() 
{ 
	return(this.ProjectorType); 
};

/**
* Called by layers to force the bounds to be recomputed.
* @protected
*/
CMGeo.prototype.SetBoundsDirty=function() 
{ 
	this.TheBounds=null;
}
/**
* Gets the bounds of all the layer bounds combined.
* @public
*/
CMGeo.prototype.GetBounds=function() 
{ 
	var NumLayers=this.GetNumChildren(CMLayer);
	
	if ((NumLayers>0)&&(this.TheBounds==null)) // bounds is dirty and we have layer data
	{
		for (var i=0;i<NumLayers;i++)
		{
			var TheLayer=this.GetChild(i,CMLayer);
			
			var TheBounds=TheLayer.GetBounds();
			
			if ((this.TheBounds==null)&&(TheBounds!=null))
			{
				this.TheBounds=CMUtilities.CloneBounds(TheBounds);
			}
			else if (TheLayer.TheBounds!=null)
			{
				CMUtilities.AddToBounds( this.TheBounds,TheBounds);
			}
		}
	}
	return(this.TheBounds);
}
/**
* Insert a Time in the appropraite location in the array
* @public 
* @param Time - the time slice to insert
*/
CMGeo.prototype.InsertTime=function(Time) 
{
	Time=parseFloat(Time); // make sure we have a number

	CMUtilities.InsertIntoSortedArray(this.Times,Time);
	
	this.SendMessageToListeners(CMScene.MESSAGE_TIME_SLICES_CHANGED);

}
/**
* Remove the entry for the specified Time
* @public 
* @param Time - the time slice to delete
*/
CMGeo.prototype.DeleteTime=function(Time) 
{ 
	Time=parseFloat(Time); // make sure we have a number

	var Index=this.Times.indexOf(Time);
	if (Index!=-1)
	{
		this.Times.splice(Index,1);
	}
	this.SendMessageToListeners(CMScene.MESSAGE_TIME_SLICES_CHANGED);
}


/**
* Unselect all information in the layers
* @public
*/
CMGeo.prototype.CMItem_UnselectAll=CMItem.prototype.UnselectAll;

CMGeo.prototype.UnselectAll=function(SendMessageFlag) 
{
	this.CMItem_UnselectAll(SendMessageFlag);
	
	var NumLayers=this.GetNumChildren(CMLayer);
	
	for (var i=0;i<NumLayers;i++)
	{
		var TheLayer=this.GetChild(i);
		
		TheLayer.UnselectAll(SendMessageFlag);
	}
}

