/******************************************************************************************************************
* CMItemPoly Class
*
* Represents:
* - Straight lines
* - Polylines
* - Polygons
* - Bezier curves
* - Bezier polygons
* Also a superclass for the CMItemPolyArrow class
*
* @module CMDatasetPyramidOpenFormat
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/
//******************************************************************
// Definitions
//******************************************************************
/**
* Below are the settings definitions.
* @public, @settings
*/
CMItemPoly.SettingDefintions=
{
	Curve:
	{
		Coordinates: { Name:"Coordinates",Type:CMBase.DATA_TYPE_COORDINATES, Default:null },
		Smoothing: { Name:"Smoothing",Type:CMBase.DATA_TYPE_BOOLEAN_ARRAY, Default:null }
	}
};

//******************************************************************
// CMItemPoly Constructor
//******************************************************************
/*
* Constructor
* @public, @constructs
*/
function CMItemPoly() 
{
	CMItem.call(this);
	
	// settings
	this.TimeSlices[0].Settings.Curve=	
	{
	};
	// other properties
	this.Anchor=null; // JSON with X,Y
	this.Dragging=false;
	this.Creating=false;
	this.SelectedPart=-1;
	this.Closed=false;
	
	this.Xs=null; // coordinates tweened from settings
	this.Ys=null;
}

CMItemPoly.prototype=Object.create(CMItem.prototype); // inherit prototype functions from PanelBase()

CMItemPoly.prototype.contructor=CMItemPoly; // override the constructor to go to ours

//******************************************************************
// Private Functions
//******************************************************************
/**
* @private
*/
CMItemPoly.prototype.GetAnchor=function(TheView,RefX,RefY,ThePart)
{
	var TheScene=this.GetParent(CMScene);
	
	var MinTime=TheScene.GetTimeRange();

	//var Tolerance=TheView.GetRefWidthFromPixelWidth(3);
	
	var Result=this.GetControlPoints(MinTime);
	
	var AnchorY=RefX-Result.Xs[ThePart];
	var AnchorX=RefY-Result.Ys[ThePart];
		
	return({X:AnchorX,Y:AnchorY});
}
/**
* @private
*/
CMItemPoly.prototype.Reset=function()
{
	this.Xs=null;
	this.Ys=null;
}
/**
* Sets up the Bezier curves
* @private
*/
CMItemPoly.prototype.Initialize=function(MinTime) 
{
	if ((this.Xs==null)||(MinTime!=this.TimeOfLastInitialize)) // need to allocate the curves
	{
		// setup the return values
		var X1s=null;
		var Y1s=null;
		
		// get the control points
		var Result=this.GetTweenedControlPoints();
		var Xs=Result.Xs;
		var Ys=Result.Ys;
	
		var Smoothing=this.GetSetting("Curve","Smoothing");
		
		// get the segments
		if (Xs.length<=1) // one point,do nothing
		{
		}
		else if (Xs.length==2) // straight line
		{
			X1s=[Xs[0],Xs[1]];
			Y1s=[Ys[0],Ys[1]];
		}
		else // Bezier curves
		{
			var NumPoints=Xs.length;
			
			var NumCurves=Xs.length;
			if (this.Closed==false) NumCurves=Xs.length-1;
			
			X1s=[Xs[0]];
			Y1s=[Ys[0]];
			
			// start at the start point
			
			for (var Curve=0;Curve<NumCurves;Curve++)
			{
				// for each curve, add the additional points to the end of the curve (which is the start of the next curve)
				
				//*************************************************
				// find the start and end points
				var StartPoint=Curve;
				var EndPoint=Curve+1;
				
				if (EndPoint==Xs.length) EndPoint=0; // wraps around
				
				//*************************************************
				// find the smoothing factors
				var StartSmooth=false;
				
				if (Smoothing!=null)
				{
					if ((Curve!=0)||(this.Closed)) StartSmooth=Smoothing[StartPoint]; // allow smoothing once past first point or if closed
					
					var EndSmooth=false;
					if ((Curve+1<NumPoints-1)||(this.Closed)) EndSmooth=Smoothing[EndPoint]; // not at the end
				}
				//*****************************************************
				var PreviousPoint=StartPoint-1;
				if (PreviousPoint<0) PreviousPoint=Xs.length-1; // wrap back to last point
				
				var NextPoint=EndPoint+1;
				if (NextPoint==Xs.length) NextPoint=0;
				
				//*****************************************************
				if (StartSmooth==false) // start sharp, draw a 3rd order Bezier or straight line
				{
					if (EndSmooth==false) // start sharp and end sharp, draw straight line
					{
						X1s.push(Xs[EndPoint]);
						Y1s.push(Ys[EndPoint]);
					}
					else // start sharp and end smooth, 3rd order Bezier 
					{
						var Points=CMUtilityBezier.GetThreePoint3D(10,Xs[StartPoint],Ys[StartPoint],0,
							Xs[EndPoint],Ys[EndPoint],0,Xs[NextPoint],Ys[NextPoint],0);
						
						for (var i=0;i<Points[0].length;i++)
						{
							X1s.push(Points[0][i]);
							Y1s.push(Points[1][i]);
						}
						X1s.push(Xs[EndPoint]);
						Y1s.push(Ys[EndPoint]);
						//this.DrawCurve(TheView,Xs,Ys,Points,0,1,X1s,Y1s);
					}
				}
				else // start smooth, draw a 3rd order or 4th order Bezier 
				{
					if (EndSmooth) // both smooth, draw a 4th order Bezier
					{
						var Points=CMUtilityBezier.GetFourPoint2D(10,Xs[PreviousPoint],Ys[PreviousPoint],
								Xs[StartPoint],Ys[StartPoint],Xs[EndPoint],Ys[EndPoint],
								Xs[NextPoint],Ys[NextPoint]);
						
						for (var i=0;i<Points[0].length;i++)
						{
							X1s.push(Points[0][i]);
							Y1s.push(Points[1][i]);
						}
						X1s.push(Xs[EndPoint]);
						Y1s.push(Ys[EndPoint]);
					}
					else // start smooth, end sharp
					{
						var LastPoint=Xs.length-1;
						
						var Points=CMUtilityBezier.GetThreePoint3D(10,Xs[EndPoint],Ys[EndPoint],0,
									Xs[StartPoint],Ys[StartPoint],0,Xs[PreviousPoint],Ys[PreviousPoint],0);
						
						for (var i=Points[0].length-1;i>=0;i--)
						{
							X1s.push(Points[0][i]);
							Y1s.push(Points[1][i]);
						}
						X1s.push(Xs[EndPoint]);
						Y1s.push(Ys[EndPoint]);
					}
				}
			}
		}
		this.Xs=X1s;
		this.Ys=Y1s;
		this.TimeOfLastInitialize=MinTime;
	}
}

//******************************************************************
// CMBase Functions
//******************************************************************

CMItemPoly.prototype.GetName=function()  
{ 
	return("Curve"); 
}

CMItemPoly.prototype.CMItem_GetSettingsDefinitions=CMItem.prototype.GetSettingsDefinitions; // save the superclass function

CMItemPoly.prototype.GetSettingsDefinitions=function() 
{
	var Result=this.CMItem_GetSettingsDefinitions(); // get the settings definitions from the superclass
	
	for (Key in CMItemPoly.SettingDefintions)
	{
		Result[Key]=CMItemPoly.SettingDefintions[Key];
	}

	return(Result); 
}

//******************************************************************
// CMItem Functions
//******************************************************************
CMItemPoly.prototype.MouseDown=function(TheView,RefX,RefY,TheEvent) 
{
	var Used=false;

	if (this.Creating) // 
	{
		var TheScene=this.GetParent(CMScene);
		
		var MinTime=TheScene.GetTimeRange();
		
		var ControlPoints=this.GetControlPoints(MinTime);
		
		var NumPoints=ControlPoints.Xs.length;
		
		if (TheEvent.detail==2) // double click, remove the last point and 
		{
			this.Creating=false;
			this.Dragging=false;
			
			ControlPoints.Xs.pop(); // remove the last point which was added by the second button click
			ControlPoints.Ys.pop(); // remove the last point which was added by the second button click
			ControlPoints.Smoothing.pop(); // remove the last point which was added by the second button click
		}
		else // add a point
		{
			var RefTolerance=this.GetRefTolerance(TheView);
			
			var ThePart=this.InPart(TheView,RefX,RefY,RefTolerance);
			
			if (ThePart==0) // last point, close the polygon
			{
				// remove the last point which is the one being dragged around
				ControlPoints.Xs.pop(); 
				ControlPoints.Ys.pop(); 
				ControlPoints.Smoothing.pop(); 
				
				// close the polygon
				this.Closed=true;
				
				this.Creating=false;
				this.Dragging=false;
			}
			else // add the point
			{
				ControlPoints.Xs.splice(NumPoints-1,0,RefX);
				ControlPoints.Ys.splice(NumPoints-1,0,RefY);
				ControlPoints.Smoothing.splice(NumPoints-1,0,ControlPoints.Smoothing[NumPoints-1]); // duplicate the last points smoothing setting
			}
		}
		this.SetControlPoints(MinTime,ControlPoints.Xs,ControlPoints.Ys,ControlPoints.Smoothing);
		
		Used=true;
		this.Repaint();
	}
	else if (Used==false)
	{
		if (TheEvent.button==0) // left button clicked
		{
			if ((TheView.GetTool()==CMView.TOOL_INFO)||(TheView.GetTool()==CMView.TOOL_SELECT)) // select a part to drag
			{
				var RefTolerance=this.GetRefTolerance(TheView);
			
				var SelectedPart=this.InPart(TheView,RefX,RefY,RefTolerance);
		
				if (SelectedPart!=-1)
				{
					var TheScene=this.GetParent(CMScene);
		
					TheScene.UnselectAll(true);
					
					this.Dragging=true;
					this.SelectedPart=SelectedPart;
					this.SetSelected(true);
					
					if (SelectedPart==-2) // inside
					{
						this.Anchor={X:RefX,Y:RefY};
					}
					else
					{
						this.Anchor=this.GetAnchor(TheView,RefX,RefY,SelectedPart);
					}
					Used=true;
				}
			}
		}
		else // right buttonclicked
		{
			var RefTolerance=this.GetRefTolerance(TheView);
			
			var SelectedPart=this.InPart(TheView,RefX,RefY,RefTolerance);
	
			if (SelectedPart>=0) // a point was clicked in
			{
				var ThePopupMenu=CMUtilities.GetPopupMenu("CM_LayerPopupMenu",event.clientX,event.clientY);
				
				//*******************************************************************
				// Add the delete item
				
				var DeleteElement=document.createElement('div');
				DeleteElement.setAttribute("id","CM_DeleteElementMenuItem");
				DeleteElement.className="CM_LayerListPopupMenuItem";
				
				DeleteElement.innerHTML="Delete Vertex";
				
				DeleteElement.TheItem=this;
				DeleteElement.SelectedPart=SelectedPart;
				DeleteElement.ThePopupMenu=ThePopupMenu;
				
				DeleteElement.addEventListener('click', function(event)
				{
					this.TheItem.DeleteControlPoint(this.SelectedPart);
					this.ThePopupMenu.style.visibility = "hidden";
					
					// stop event from propogating
					event.stopPropagation();
					
					this.TheItem.Repaint();
				});
				
				ThePopupMenu.appendChild(DeleteElement);
				
				//*******************************************************************
				// Add the duplicate item
				
/*				var DeleteElement=document.createElement('div');
				DeleteElement.setAttribute("id","CM_DuplicateElementMenuItem");
				DeleteElement.className="CM_LayerListPopupMenuItem";
				
				DeleteElement.innerHTML="Duplicate";
				
				DeleteElement.TheItem=this;
				DeleteElement.SelectedPart=SelectedPart;
				DeleteElement.ThePopupMenu=ThePopupMenu;
				
				DeleteElement.addEventListener('click', function(event)
				{
					this.TheItem.DuplicateControlPoint(this.SelectedPart);
					
					// stop event from propogating
					event.stopPropagation();
					
					this.TheItem.Repaint();
				});
				
				ThePopupMenu.appendChild(DeleteElement);
				
				//*******************************************************************
				// Add the smooth/sharp item
				
				var DeleteElement=document.createElement('div');
				DeleteElement.setAttribute("id","CM_SmoothElementMenuItem");
				DeleteElement.className="CM_LayerListPopupMenuItem";
				
				var Smoothing=this.GetSetting("Curve","Smoothing",undefined,MinTime);
					
				if (Smoothing[SelectedPart]) DeleteElement.innerHTML="Make sharp";
				else  DeleteElement.innerHTML="Make smooth";
				
				DeleteElement.TheItem=this;
				DeleteElement.SelectedPart=SelectedPart;
				DeleteElement.ThePopupMenu=ThePopupMenu;
				
				DeleteElement.addEventListener('click', function(event)
				{
					var Smoothing=this.TheItem.GetSetting("Curve","Smoothing",undefined,MinTime);
					
					if (Smoothing[this.SelectedPart]) Smoothing[this.SelectedPart]=false;
					else Smoothing[this.SelectedPart]=true;
					
					this.TheItem.SetSetting("Curve","Smoothing",Smoothing,MinTime);
					
					// stop event from propogating
					event.stopPropagation();
					
					this.TheItem.Repaint();
				});
				
				ThePopupMenu.appendChild(DeleteElement);
*/				
						
				TheEvent.stopPropagation(); // stop the document from hidding a popup window
				TheEvent.preventDefault(); // keeps regular menu from appearing
		
				Used=true;
			}
			if (SelectedPart==-2) // a point was clicked in
			{
				var ThePopupMenu=CMUtilities.GetPopupMenu("CM_LayerPopupMenu",event.clientX,event.clientY);
				
				//*******************************************************************
				// Add the delete item
				
				var DeleteElement=document.createElement('div');
				DeleteElement.setAttribute("id","CM_DeleteElementMenuItem");
				DeleteElement.className="CM_LayerListPopupMenuItem";
				
				DeleteElement.innerHTML="Delete Poly";
				
				DeleteElement.TheItem=this;
				DeleteElement.ThePopupMenu=ThePopupMenu;
				
				DeleteElement.addEventListener('click', function(event)
				{
					var TheLayer=this.TheItem.GetParent();
					TheLayer.RemoveChild(this.TheItem);
					this.ThePopupMenu.style.visibility = "hidden";
					
					// stop event from propogating
					event.stopPropagation();
					
					TheLayer.Repaint();
				});
				
				ThePopupMenu.appendChild(DeleteElement);
						
				TheEvent.stopPropagation(); // stop the document from hidding a popup window
				TheEvent.preventDefault(); // keeps regular menu from appearing
		
				Used=true;
			}
		}
	}
	return(Used);
}

CMItemPoly.prototype.MouseMove=function(TheView,RefX,RefY,TheEvent) 
{
	var Used=false;
	
//	if (this.Clickable)
	{
		var TheScene=this.GetParent(CMScene);
		
		var MinTime=TheScene.GetTimeRange();
		
		if (this.Creating)
		{
			var ControlPoints=this.GetControlPoints(MinTime);
			
			var NumPoints=ControlPoints.Xs.length;
			
			ControlPoints.Xs[NumPoints-1]=RefX;
			ControlPoints.Ys[NumPoints-1]=RefY;
			
			this.SetControlPoints(MinTime,ControlPoints.Xs,ControlPoints.Ys,ControlPoints.Smoothing);
			this.Repaint();
		}
		else if (this.Dragging) // dragging an existing item (create or update)
		{
			var ControlPoints=this.GetControlPoints(MinTime);
			
			if (this.SelectedPart==-2) // move entire poly
			{
				var DeltaX=RefX-this.Anchor.X;
				var DeltaY=RefY-this.Anchor.Y;
				
				for (var i=0;i<ControlPoints.Xs.length;i++)
				{
					ControlPoints.Xs[i]=ControlPoints.Xs[i]+DeltaX;
					ControlPoints.Ys[i]=ControlPoints.Ys[i]+DeltaY;
					
					this.Anchor.X=RefX;
					this.Anchor.Y=RefY;
				}
			}
			else
			{
				ControlPoints.Xs[this.SelectedPart]=RefX+this.Anchor.X;
				ControlPoints.Ys[this.SelectedPart]=RefY+this.Anchor.Y;
			}
			
			this.SetControlPoints(MinTime,ControlPoints.Xs,ControlPoints.Ys,ControlPoints.Smoothing);
			this.Repaint();
		}
		else // update the mouse cursor
		{
			var RefTolerance=this.GetRefTolerance(TheView);
			
			var ThePart=this.InPart(TheView,RefX,RefY,RefTolerance);
			
			if (ThePart!==-1) 
			{
				var TheCanvasMap=this.GetParent(CMMainContainer);
				var TheCanvasElement=TheCanvasMap.GetElement(CMMainContainer.CANVAS);
				
				TheCanvasElement.style.cursor = "move";
			}
		}
	}
	return(Used);
};
CMItemPoly.prototype.MouseUp=function(TheView,RefX,RefY,TheEvent)
{
	var Used=false;
	
//	if (this.Clickable)
	{
		if ((this.Dragging)&&(this.Creating==false)) 
		{
			this.Dragging=false;
			Used=true;
		}
	}
	return(Used);
};

CMItemPoly.prototype.StartCreating=function(RefX,RefY) 
{
	this.Anchor={X:0,Y:0};
	this.SelectedPart=CMItem.PART_LOWER_RIGHT;
	this.Dragging=true;
	this.Creating=true;
}

CMItemPoly.prototype.Paint=function(TheView) 
{
	var TheScene=this.GetParent(CMScene);
	
	var MinTime=TheScene.GetTimeRange();
	
	this.Initialize(MinTime);
	
	if (this.Xs!=null) // have more than one point
	{
		var X1s=this.Xs;
		var Y1s=this.Ys;
		
		// setup the style for the lines
		var TheStyle=this.GetStyle(TheView,MinTime);
		
		if (TheStyle!=undefined) TheView.SetStyle(TheStyle,true);
	
		TheView.PaintRefPoly(X1s,Y1s,this.Closed);		// draw the left side of the arrow
		
		if (TheStyle!=undefined) TheView.RestoreStyle();
		
		// write out the text along the curve if any
		var Text=this.GetSetting("Text","Text");
		
		if (Text!=null)
		{
			var TheStyle=this.GetStyle(TheView,0,"Text");
			
			if (TheStyle!=undefined)
			{
				TheView.SetStyle(TheStyle);
				
				// find the text distance
				
				var TextWidths=[];
				var TotalTextWidth=0;
				
				for (var TextIndex=0;TextIndex<Text.length;TextIndex++)
				{
					var TextWidth=TheView.GetTextWidthInPixels(Text[TextIndex]);
	
					var RefTextWidth=TheView.GetRefWidthFromPixelWidth(TextWidth);
					
					TextWidths.push(RefTextWidth);
					
					TotalTextWidth+=RefTextWidth;
				}
	
				// find the spacing along the curve
				
				var TotalDistance=0;
				
				for (var i=0;i<X1s.length-1;i++)
				{
					var X1=X1s[i]; 
					var Y1=Y1s[i];
					var X2=X1s[i+1];
					var Y2=Y1s[i+1];
					
					TotalDistance+=Math.sqrt(Math.pow(X2-X1,2)+Math.pow(Y2-Y1,2));
				}
				var EndPad=0;
				
				var Spacing=(TotalDistance-(EndPad*2)-TotalTextWidth)/(Text.length);
				
				// draw each character along the curve
				
				var SegmentDistance=0;
				var SegmentLength=0;
				var LastSegmentDistance=0;
				var CurrentSegment=0;
				var DesiredDistance=EndPad;
				var X;
				var Y;
				var Angle;
				
				for (var TextIndex=0;TextIndex<Text.length;TextIndex++)
				{
					// move forward along the segments until we go past the one we want or hit the end of the segments
					while ((SegmentDistance<DesiredDistance)&&(CurrentSegment<X1s.length))
					{
						LastSegmentDistance=SegmentDistance;
						
						// move forward one segment
						CurrentSegment+=1;
						
						var X1=X1s[CurrentSegment-1]; 
						var Y1=Y1s[CurrentSegment-1];
						var X2=X1s[CurrentSegment];
						var Y2=Y1s[CurrentSegment];
						
						SegmentLength=Math.sqrt(Math.pow(X2-X1,2)+Math.pow(Y2-Y1,2));
						
						SegmentDistance+=SegmentLength;
					}
					
					if (CurrentSegment>0)
					{
						// get the angle of the segment
						var X1=X1s[CurrentSegment-1]; 
						var Y1=Y1s[CurrentSegment-1];
						var X2=X1s[CurrentSegment];
						var Y2=Y1s[CurrentSegment];
						
						var DX=X2-X1;
						var DY=Y2-Y1;
					
						Angle=Math.atan2(DY,DX)
						
						// get the position on the segment
						
						var RemamingDistance=DesiredDistance-LastSegmentDistance;
						
						var Factor=RemamingDistance/SegmentLength;
						
						X=X1+(DX*Factor);
						Y=Y1+(DY*Factor);
					}
					else
					{
						// get the angle of the segment
						var X1=X1s[CurrentSegment]; 
						var Y1=Y1s[CurrentSegment];
						var X2=X1s[CurrentSegment+1];
						var Y2=Y1s[CurrentSegment+1];
						
						var DX=X2-X1;
						var DY=Y2-Y1;
					
						Angle=Math.atan2(DY,DX)
						
						X=X1;
						Y=Y1;
					}
					// paint the text
					TheView.PaintRefText(Text[TextIndex],X,Y,12,"center",-Angle);
					
					// move the desired spacing forward
					DesiredDistance+=(TextWidths[TextIndex])+(Spacing);
					
					// if we are not at the last character, add 1/2 the width of the text character
					//if ((TextIndex+1)<Text.length) DesiredDistance+=TextWidths[TextIndex+1]/2;
				}
				TheView.RestoreStyle();
			}
		}
	}
}

CMItemPoly.prototype.PaintSelected=function(TheView) 
{
	if (this.Selected)
	{
		TheView.SaveStyle();
		
		TheView.SetStyle({fillStyle:"rgba(0,0,0,1)",strokeStyle:"white"});
		
		var TheScene=this.GetParent(CMScene);
		
		var MinTime=TheScene.GetTimeRange();
		
		var Result=this.GetControlPoints(MinTime);
		
		if (Result!=null)
		{
			var X1s=Result.Xs;
			var Y1s=Result.Ys;
				
			for (var i=0;i<X1s.length;i++)
			{
				TheView.PaintRefCircle(X1s[i],Y1s[i],5);
			}
		}
		TheView.RestoreStyle();
	}
}

//******************************************************************
// CMItemPoly Functions
//******************************************************************
/**
* Returns the desired control points based on the TimeSlice setting.
* This function ignores tweening as the control points are at fixed
* locations
*/
CMItemPoly.prototype.GetControlPoints=function(TimeSlice)
{
	var Result=null;
	
	var Settings=this.GetBoundingTimeSlices(TimeSlice);
	
	Settings=Settings[0].Settings;
	
	if (Settings!=undefined)
	{
		if (Settings.Curve.Coordinates!=undefined)
		{
			var Xs=Settings.Curve.Coordinates.Xs;
			var Ys=Settings.Curve.Coordinates.Ys;
			var Smoothing=Settings.Curve.Smoothing;
			
			Result={Xs:Xs,Ys:Ys,Smoothing:Smoothing};
		}
	}
	
	return(Result);
}
CMItemPoly.prototype.SetControlPoints=function(TimeSlice,Xs,Ys,Smoothing)
{
	var Result=this.GetBoundingTimeSlices(TimeSlice);
	
	Settings=Result[0].Settings;
	
	if (Settings!=undefined)
	{
		if (Settings.Curve.Coordinates==undefined) Settings.Curve["Coordinates"]={};
		if (Settings.Curve.Smoothing==undefined) Settings.Curve["Smoothing"]=[];
		
		Settings.Curve.Coordinates.Xs=Xs;
		Settings.Curve.Coordinates.Ys=Ys;
		Settings.Curve.Smoothing=Smoothing;
		
		this.SendMessageToListeners(CMItem.MESSAGE_CONTENT_CHANGED);
	}
	else
	{
		alert("Sorry, time slice "+TimeSlice+" is not defined");
	}
	this.Reset();
}
/**
* Find the final points that define the outside of the arrow
* @protected
*/
CMItemPoly.prototype.GetTweenedControlPoints=function() 
{
	// interpolate between the coordinates to find the set of control points for this time slice
	var Xs=[];
	var Ys=[];
	
	var TheCanvsMap=this.GetParent(CMMainContainer);
	
	var TheScene=this.GetParent(CMScene);
	
	var TheTime=TheScene.GetTimeRange();
	
	var Result=this.GetBoundingTimeSlices(TheTime);
	var TheTimeSlice1=Result[0];
	
	var TheTime1=TheTimeSlice1.Time;
		
	if (Result[1]==null) // no tweening
	{
		var Result=this.GetControlPoints(TheTime1);
		Xs=Result.Xs;
		Ys=Result.Ys;
	}
	else
	{
		var Factor=CMItem.GetTimeFactor(Result);
		
		var Result=this.GetControlPoints(TheTime1);
		var X1s=Result.Xs;
		var Y1s=Result.Ys;
		var Result=this.GetControlPoints(TheTime2);
		var X2s=Result.Xs;
		var Y2s=Result.Ys;
		
		for (var i=0;i<X1s.length;i++)
		{
			Xs[i]=X1s[i]*Factor1+X2s[i]*Factor2;
			Ys[i]=Y1s[i]*Factor1+Y2s[i]*Factor2;
		}
	}
	return({Xs:Xs,Ys:Ys});
}
CMItemPoly.prototype.DeleteControlPoint=function(SelectedPart)
{
	var TheScene=this.GetParent(CMScene);
	
	var MinTime=TheScene.GetTimeRange();
	
	// delete the control point
	var ControlPoints=this.GetControlPoints(MinTime);

	ControlPoints.Xs.splice(SelectedPart,1); // delete the point clicked on
	ControlPoints.Ys.splice(SelectedPart,1);
	ControlPoints.Smoothing.splice(SelectedPart,1);
				
	this.SetControlPoints(MinTime,ControlPoints.Xs,ControlPoints.Ys,ControlPoints.Smoothing);

}

CMItemPoly.prototype.DuplicateControlPoint=function(SelectedPart)
{
	var TheScene=this.GetParent(CMScene);
	
	var MinTime=TheScene.GetTimeRange();
	
	// duplicate
	var ControlPoints=this.GetControlPoints(MinTime);

	ControlPoints.Xs.splice(SelectedPart,0,ControlPoints.Xs[SelectedPart]); // duplicate the point clicked on
	ControlPoints.Ys.splice(SelectedPart,0,ControlPoints.Ys[SelectedPart]);
	ControlPoints.Smoothing.splice(SelectedPart,0,ControlPoints.Smoothing[SelectedPart]);
				
	this.SetControlPoints(MinTime,ControlPoints.Xs,ControlPoints.Ys,ControlPoints.Smoothing);
}
/**
* Returns an index if the coordinate is in the polygon
* @returns ThePart - -1 for none, -2 for inside, >=0 for one of the control points
*/
CMItemPoly.prototype.InPart=function(TheView,RefX,RefY,RefTolerance)
{
	var ThePart=-1;

	var TheScene=this.GetParent(CMScene);
	
	var MinTime=TheScene.GetTimeRange();
	
	var Result=this.GetControlPoints(MinTime);
	if (Result!=null)
	{
		var X1s=Result.Xs;
		var Y1s=Result.Ys;
		
		// first, try the control points
		for (var i=0;(i<X1s.length)&&(ThePart==-1);i++)
		{
			var X=X1s[i];
			var Y=Y1s[i];
			
			if ((Math.abs(X-RefX)<RefTolerance)&&(Math.abs(Y-RefY)<RefTolerance))
			{
				ThePart=i;
			}
		}
		// if not part selected and the poly is a polygon, check inside
		if (ThePart==-1)
		{
			if (this.Closed) // try for the interior
			{
				Result=CMUtilities.InsideAPolygon(RefX,RefY,X1s,Y1s,X1s.length);
				
				if (Result) ThePart=-2;
			}
			else // check the edges of the line
			{
				this.Initialize(MinTime);
				
				var Result=CMUtilities.InPolyline(RefX,RefY,this.Xs,this.Ys,RefTolerance);
				
				if (Result) ThePart=-2;
			}
		}
	}
	
	return(ThePart);
}
