/******************************************************************************************************************
* CMLayer
* This class is the base class for all other layers.  This class contains functions and properties for feature and raster based
* classes.  This is because the CMLayerPyramid supports both raster and vector data.  This also may be a trend with data formats like KML.
*
* @module CMLayer
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/

//*****************************************************************************************************************
// Definitions
//*****************************************************************************************************************

/**
* Definitions for types of marks
* @public, @enum
*/
CMLayer.MARK_CIRCLE=0;
CMLayer.MARK_TRIANGLE=1;
CMLayer.MARK_SQUARE=2;
CMLayer.MARK_STAR=3;

/**
* Array for the types of directions
* @private (may be made public later)
*/
CMLayer.LABEL_DIRECTIONS=["TL","T","TR","R","BR","B","BL","L"]; // definitions for the direction a label will be from its point

/**
* Below are the additional settings definitions for STLayers.  Setings for the basic drawing style and text
* style are inherited from STItem
* @public, @settings
*/
CMLayer.SettingDefintions=
{
	Layer:
	{
		InfoText: { Name:"Info Text",Type:CMBase.DATA_TYPE_STRING, Default:undefined }, // HTML content to appear in the Info box
		Metadata: { Name:"Metadata",Type:CMBase.DATA_TYPE_STRING, Default:undefined }, // source information for the data in the layer
		MinZoom: { Name:"Min Zoom",Type:CMBase.DATA_TYPE_FLOAT, Default:-Infinity }, // layer is hidden below this zoom level
		MaxZoom: { Name:"Max Zoom",Type:CMBase.DATA_TYPE_FLOAT, Default:Infinity }, // layer is hidden above this zoom level
		ZoomToBoundsOnLoad: { Name:"Zoom To Bounds On Load",Type:CMBase.DATA_TYPE_BOOLEAN, Default:false }, // If true, the scene will zoom to the layer bounds after the layer's data is loaded
	}, 
	Mark: 
	{ 
		Type: { Name:"Type",Type:CMBase.DATA_TYPE_ENUMERATED, Options:[CMLayer.MARK_CIRCLE,CMLayer.MARK_TRIANGLE,CMLayer.MARK_SQUARE,CMLayer.MARK_STAR],Default:CMLayer.MARK_CIRCLE },
		Size: { Name:"Size",Type:CMBase.DATA_TYPE_FLOAT, Default:3 },
	},
	IconImage:
	{
		URL: { Name:"URL",Type:CMBase.DATA_TYPE_URL, Default:"" },
		TheImage: { Name:"Image",Type:CMBase.DATA_TYPE_IMAGE, Default:null },
		OffsetX: { Name:"Offset X",Type:CMBase.DATA_TYPE_FLOAT, Default:0 },
		OffsetY: { Name:"Offset Y",Type:CMBase.DATA_TYPE_FLOAT, Default:0 },
	},
	LabelBox: 
	{
		Position: { Name:"Position",Type:CMBase.DATA_TYPE_ENUMERATED, Options:CMLayer.LABEL_DIRECTIONS,Default:"TR" },
		OffsetX: { Name:"Offset X",Type:CMBase.DATA_TYPE_FLOAT, Default:0 },
		OffsetY: { Name:"Offset Y",Type:CMBase.DATA_TYPE_FLOAT, Default:0 },
		Padding: { Name:"Padding",Type:CMBase.DATA_TYPE_FLOAT, Default:2 },
		
		strokeStyle: { Name:"Line Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(0,0,0)" },
		lineWidth: { Name:"Width",Type:CMBase.DATA_TYPE_INTEGER, Default:1 },
		lineCap: { Name:"Line Cap",Type:CMBase.DATA_TYPE_ENUMERATED, Options:['butt','round','square'],Default:'round' },
		lineJoin: { Name:"Line Join",Type:CMBase.DATA_TYPE_ENUMERATED, Options:['bevel','round','miter'],Default:'round' },
		
		fillStyle: { Name:"Fill Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(255,255,255)" },
		
		shadowColor: { Name:"Shadow Color",Type:CMBase.DATA_TYPE_COLOR, Default:"rgb(0,0,0)" },
		shadowBlur: { Name:"Shadow Blur",Type:CMBase.DATA_TYPE_FLOAT, Default:1 },
		shadowOffsetX: { Name:"Shadow X Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:1 },
		shadowOffsetY: { Name:"Shadow Y Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:1 }
	},
	MouseOverStyle:
	{
		strokeStyle: { Name:"Line Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(255,0,0)" },
		lineWidth: { Name:"Width",Type:CMBase.DATA_TYPE_INTEGER, Default:1 },
		lineCap: { Name:"Line Cap",Type:CMBase.DATA_TYPE_ENUMERATED, Options:['butt','round','square'],Default:'round' },
		lineJoin: { Name:"Line Join",Type:CMBase.DATA_TYPE_ENUMERATED, Options:['bevel','round','miter'],Default:'round' },
		
		fillStyle: { Name:"Fill Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(255,255,255)" },
		
		shadowColor: { Name:"Shadow Color",Type:CMBase.DATA_TYPE_COLOR, Default:"rgb(0,0,0)" },
		shadowBlur: { Name:"Shadow Blur",Type:CMBase.DATA_TYPE_FLOAT, Default:1 },
		shadowOffsetX: { Name:"Shadow X Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:1 },
		shadowOffsetY: { Name:"Shadow Y Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:1 }
	},
	SelectedStyle:
	{
		strokeStyle: { Name:"Line Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(0,0,255)" },
		lineWidth: { Name:"Width",Type:CMBase.DATA_TYPE_INTEGER, Default:1 },
		lineCap: { Name:"Line Cap",Type:CMBase.DATA_TYPE_ENUMERATED, Options:['butt','round','square'],Default:'round' },
		lineJoin: { Name:"Line Join",Type:CMBase.DATA_TYPE_ENUMERATED, Options:['bevel','round','miter'],Default:'round' },
		
		fillStyle: { Name:"Fill Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(255,255,255)" },
		
		shadowColor: { Name:"Shadow Color",Type:CMBase.DATA_TYPE_COLOR, Default:"rgb(0,0,0)" },
		shadowBlur: { Name:"Shadow Blur",Type:CMBase.DATA_TYPE_FLOAT, Default:1 },
		shadowOffsetX: { Name:"Shadow X Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:1 },
		shadowOffsetY: { Name:"Shadow Y Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:1 }
	},
};

//*****************************************************************************************************************
// Constructors
//*****************************************************************************************************************

/**
* Constructor for a layer object.  Layers contain spatial data that appears in a scene
* @public, @constructs
*/
function CMLayer() 
{
	CMItem.call(this);
	
	// settings
	
	this.InfoWindowWidth=300; // jjg setting
	
	this.TimeSlices[0].Settings.Layer=	
	{
	};
	this.TimeSlices[0].Settings.Dataset=	
	{
	};
	this.TimeSlices[0].Settings.Mark=	
	{
	};
	this.TimeSlices[0].Settings.IconImage=
	{
	};
	this.TimeSlices[0].Settings.LabelBox=
	{
	};
	this.TimeSlices[0].Settings.MouseOverStyle=
	{
	};
	this.TimeSlices[0].Settings.SelectedStyle=
	{
		strokeStyle:"rgb(120,100,255)",
		lineWidth:"3",
		fillStyle:"rgba(0,0,0,0)"
	};
	
	/*
	* settings specific to individual features, must be set outside constructor to be unique to each layer. 
	* This is an array.  Each element has the same form as other settings:
	* { Group:
	*    { Key:Value }
	* }
	*/
	this.FeatureSettings=null;  // 
	/*
	*/
	this.SettingsAttributes=null; // objects with {Group:{Key: AttributeName} } with entries for each setting that is coming from an attribute

	//************************************************
	// general properties
	
	this.TheBounds=null; // the overall bounds (only as good as the information from the layers)
	
	// internally set properties
	
	this.FeatureBounds=null; // bounds of each feature, cached for speed (jjg move to CMDataset?)
	
	var SettingsDirty=null; // array for attriubtes, if true, settings need to be updated from attributes
	
	this.TheProjector=null;
	
	// layers should be selectable on initialization
	this.SetSetting("Item","Status",CMItem.STATUS_SELECTABLE);
}

CMLayer.prototype=Object.create(CMItem.prototype); // inherit prototype functions from CMBase()

CMLayer.prototype.contructor=CMLayer; // override the constructor to go to ours

//******************************************************************
// CMBase Functions
//******************************************************************

CMLayer.prototype.CMItem_GetSettingsDefinitions=CMItem.prototype.GetSettingsDefinitions;

CMLayer.prototype.GetSettingsDefinitions=function() 
{
	var Result=this.CMItem_GetSettingsDefinitions();
	
	for (Key in CMLayer.SettingDefintions)
	{
		Result[Key]=CMLayer.SettingDefintions[Key];
	}

	return(Result); 
}
/**
* Utility function to call old event handlers
* @public
*/
CMLayer.prototype.SendMessageToDescendants=function(Message,AdditionalInfo)
{
	Used=false;
	
	switch (Message)
	{
	case CMView.MESSAGE_MOUSE_DOWN:
		Used=this.MouseDown(AdditionalInfo.TheView,AdditionalInfo.RefX,AdditionalInfo.RefY,AdditionalInfo.TheEvent);
		break;
	case CMView.MESSAGE_MOUSE_MOVE:
		Used=this.MouseMove(AdditionalInfo.TheView,AdditionalInfo.RefX,AdditionalInfo.RefY,AdditionalInfo.TheEvent);
		break;
	case CMView.MESSAGE_MOUSE_UP:
		Used=this.MouseUp(AdditionalInfo.TheView,AdditionalInfo.RefX,AdditionalInfo.RefY,AdditionalInfo.TheEvent);
		break;
	}
	return(Used);
}

//******************************************************************
// Functions used by subclasses and not overriden
//******************************************************************

/**
* Checks if the layer is visible.  This is different form GetVisible() because it
* also checks for an optional zoom range
* @public
*/
CMLayer.prototype.IsVisible=function(TheView)
{
	var Result=this.GetVisible();
	
	if ((Result)&&(TheView instanceof CMView2D)) // may have to check zoom range
	{
		var MinZoom=this.GetSetting("Layer","MinZoom"); // typically this will not be set so this will be fast
		var MaxZoom=this.GetSetting("Layer","MaxZoom"); // typically this will not be set so this will be fast
		
		if (CMUtilities.IsDefined(MinZoom)) { Result=this.InZoomRange([MinZoom,MaxZoom]); }
	}
	return(Result);
};
/**
* Returns true if the view is within the specified Zoom Range or the zoome range is null
* This is a utility function for the IsVisible() and IsFeatureVisible() functions
* @param ZoomRange - array with [MinZoom,MaxZoom]
* @returns Flag - true if the view is visible in the zoom range, false otherwise.
*/
CMLayer.prototype.InZoomRange=function(ZoomRange)
{
	var Result=true;
	
	var TheScene=this.GetScene();
	
	if ((ZoomRange!=null)&&(TheScene!=null)&&(TheScene.GetView(0)!=null))
	{
		var TheView=TheScene.GetView(0);
		var TheZoom=TheView.GetZoomLevel();
		
		if ((TheZoom<ZoomRange[0])||(TheZoom>ZoomRange[1])) // current zoom is less than min or greater than max
		{
			Result=false;
		}
	}
	return(Result);
}
/**
* Check if the feature is visible in the view.
* @private
* @param FeatureIndex - index for the feature to check
* @returns Flag - true if the feature is visible in the view, false otherwise.
*/
CMLayer.prototype.IsFeatureVisible=function(TheView,FeatureIndex)
{
	var Result=this.IsVisible(TheView);
	
	if (Result) // may have to check zoom range
	{
		var MinZoom=this.GetFeatureSetting("Layer","MinZoom",FeatureIndex); // typically this will not be set so this will be fast
		var MaxZoom=this.GetFeatureSetting("Layer","MaxZoom",FeatureIndex); // typically this will not be set so this will be fast
		
		if (MinZoom!=null) { Result=this.InZoomRange([MinZoom,MaxZoom]); }
	}
	return(Result);
};

/**
* Check if the feature is visible in the view.
* This should be called by subclasses but can also be called to limit a layer's bounds after loading.
* @public
* @param NewBounds - Object for Bounds with format {XMin,XMax,YMin,YMax}
*/
CMLayer.prototype.SetBounds=function(NewBounds) 
{
	this.TheBounds=NewBounds;
}
/**
* Returns the bounds of the data within the layer.  Computed after loading the data.
* @public
* @returns Bounds - with format {XMin,XMax,YMin,YMax}
*/
CMLayer.prototype.GetBounds=function() 
{
	return(this.TheBounds);
}

/**
* Returns the Scene the layer is in.
* @public
* @returns TheScene - the scene the layer is contained in or null for none.
*/
CMLayer.prototype.GetScene=function() 
{ 
	var Result=this.GetParent(CMScene);
	return(Result); 
}

/**
* Sets the width of the information window
* @public
* @param NewWidth - Width of the info popup window in pixels
*/
CMLayer.prototype.SetInfoWindowWidth=function(NewWidth) 
{
	this.InfoWindowWidth=NewWidth;
}
/**
* Gets the width of the information window
* @public
* @returns NewWidth - Width of the info popup window in pixels
*/
CMLayer.prototype.GetInfoWindowWidth=function() 
{
	return(this.InfoWindowWidth);
}

//******************************************************************
// Feature-based functions
//******************************************************************

/**
* Sets the feature that is selected
* @public
* @param NewFeatureIndex - >=0 indicates a feature, -1 is for none.
*/
CMLayer.prototype.SetSelectedFeature=function(New) 
{
}
/**
* Sets the current feature that the mouse is over
* @public
* @param NewFeatureIndex - index to the feature the mouse is over (typcially returned by In())
*/
CMLayer.prototype.SetMouseOverFeature=function(New) 
{
}
/**
* Returns the current feature that the mouse is over
* @public
* @returns FeatureIndex - index to the feature the mouse is over (typcially returned by In()), or -1 for none
*/
CMLayer.prototype.ResetMouseOverFeature=function() 
{
}

//******************************************************************
// Property Gets and Sets (jjg - move to CMLayerDataset?)
//******************************************************************
/**
* Set the attribute column to obtain property values for features from
* @public
* @param Key - on of the CMLayer.INFO enumerated types
* @param Value - Heading of the column to extract values from.
*/
CMLayer.prototype.SetSettingAttribute=function(Group,Key,AttributeName)
{
	// allocate the settings object if needed
	if (this.SettingsAttributes==null) this.SettingsAttributes={};

	// add the group if needed
	if ((Group in this.SettingsAttributes)==false) this.SettingsAttributes[Group]={}
	
	// save the key and name of the attribute column for this setting
	this.SettingsAttributes[Group][Key]=AttributeName;
}
/**
* Get the attribute column to obtain property values for features from
* @public
* @param Group - 
* @param Key - 
* @returns Value - Heading of the column to extract values from.
*/
CMLayer.prototype.GetSettingsAttribute=function(Group,Key)
{
	var Result=null;
	
	if (this.SettingsAttributes!==null)
	{
		if (Group in this.SettingsAttributes) { Result=this.SettingsAttributes[Group][Key]; }
	}
	return(Result);
}
/**
* Set the feature settings into an array indexed by FeatureIndexes
* @public
* @param FeatureIndex - Index to the feature
* @param FeatureProperties - array of properties indexed by the feature indexes
*/
CMLayer.prototype.SetFeatureSettings=function(FeatureIndex,OneFeaturesSettings) // jjg - probably not needed (just use Group)
{
	if (this.FeatureSettings==null)
	{
		this.FeatureSettings=new Array();
	}
	this.FeatureSettings[FeatureIndex]=OneFeaturesSettings;
	
	this.SendMessageToListeners(CMBase.MESSAGE_SETTINGS_CHANGED,null);
}

/**
* Sets an entire group of settings for one feature
* @public
* @param Group - The name of the group to set
* @param FeatureIndex - Index to the feature
* @param OneFeaturesSettingGroup - the settings
*/
CMLayer.prototype.SetFeatureSettingGroup=function(Group,FeatureIndex,OneFeaturesSettingGroup)
{
	if (this.FeatureSettings==null)
	{
		this.FeatureSettings=new Array();
	}
	if (this.FeatureSettings[FeatureIndex]==undefined) this.FeatureSettings[FeatureIndex]={};
	
	if (this.FeatureSettings[FeatureIndex][Group]==undefined) this.FeatureSettings[FeatureIndex][Group]={};
	
	this.FeatureSettings[FeatureIndex][Group]=OneFeaturesSettingGroup;
	
	this.SendMessageToListeners(CMBase.MESSAGE_SETTINGS_CHANGED,null);
}
/**
* Gets an entire group of settings for one feature
* @public
* @param Group - The name of the group to set
* @param FeatureIndex - Index to the feature
* @returns - the desired group of settings or TheDefault if none
*/
CMLayer.prototype.GetFeatureSettingGroup=function(Group,FeatureIndex,Default)
{
//	this.UpdateSettingsFromAttributes(FeatureIndex);
		
	var Result=Default;
	
	if (this.FeatureSettings!=null)
	{
		if (this.FeatureSettings[FeatureIndex]!=undefined)
		{
			if (this.FeatureSettings[FeatureIndex][Group]!=undefined) Result=this.FeatureSettings[FeatureIndex][Group];
		}
	}
	return(Result);
}

/**
* Set an individual feature property based on it's FeatureIndex
* @public
* @param Key - on of the CMLayer.INFO enumerated types
* @param FeatureIndex - Feature to apply the property to
* @param Value - value for the property for the feature
*/
CMLayer.prototype.SetFeatureSetting=function(Group,Key,FeatureIndex,Value)
{
	if (this.FeatureSettings==null) this.FeatureSettings=new Array();

	if (this.FeatureSettings[FeatureIndex]==undefined) this.FeatureSettings[FeatureIndex]={};
	
	if ((Group in this.FeatureSettings[FeatureIndex])==false) this.FeatureSettings[FeatureIndex][Group]={};
	
	this.FeatureSettings[FeatureIndex][Group][Key]=Value;
	
	this.SendMessageToListeners(CMBase.MESSAGE_SETTINGS_CHANGED,null);
}
/**
* Returns a property for a specific feature.  Returns null if the property
* is not specified for the feature.
* @public
* @param Key - on of the CMLayer.INFO enumerated types
* @param FeatureIndex - Feature to apply the property to
* @param Default - The default value to return if the property has not been specified.
* @returns Value - value for the property for the feature
*/
CMLayer.prototype.GetFeatureSetting=function(Group,Key,FeatureIndex,Default)
{
//	this.UpdateSettingsFromAttributes(FeatureIndex);
		
	var Result=this.GetSetting(Group,Key,Default);
	
	if (this.FeatureSettings!==null)
	{
		var TheFeaturesSettings=this.FeatureSettings[FeatureIndex];
		
		if (TheFeaturesSettings!=undefined)
		{
			if (Group in TheFeaturesSettings)
			{
//				var Result=this.FeatureSettings[FeatureIndex][Group][Key];
				var GroupSettings=TheFeaturesSettings[Group];
				
				if (Key in GroupSettings)
				{
					Result=GroupSettings[Key];
				}
			}
		}
	}
	return(Result);
}
//**************************************************************
// CMLayer functions
//**************************************************************
CMLayer.prototype.SetProjector=function(NewProjector)
{
	this.TheProjector=NewProjector;
}

CMLayer.prototype.GetProjector=function()
{
	return(this.TheProjector);
}

/*
* Called to obtain the data for the layer from a URL.
*
* @override, @public
*/
CMLayer.prototype.RequestData=function() 
{
	var URL=this.GetSetting("Dataset","URL",null);
	this.SetURL(URL);
}
/*
* Called to obtain the data for the layer from a URL.
* Currently,only GeoJSON is supported as the transfer type.
*/
CMLayer.prototype.SetURL=function(URL) 
{
	CMMainContainer("Sorry, this function must be overriden");
}
//******************************************************************
// Painting Style functions
//******************************************************************

/**
* Icon for point features
* @public
* @param TheURL - the URL of the image to use for all features
* @param OffsetX - horizontal offset in pixels to go from the upper left of the image to the hot spot
* @param OffsetY - vertical offset in pixels to go from the upper left of the image to the hot spot
*/
CMLayer.prototype.SetIconImage=function(TheURL,OffsetX,OffsetY) 
{
	var TheImage=new Image(); 
	TheImage.Loaded=false;
	TheImage.TheURL=TheURL;
	TheImage.TheLayer=this;
	TheImage.OffsetX=OffsetX;
	TheImage.OffsetY=OffsetY;
	
	TheImage.onload=function() 
	{ 
		this.Loaded=true;
		
		this.TheLayer.SetSetting("IconImage","URL",this.TheURL);
		this.TheLayer.SetSetting("IconImage","TheImage",this);
		this.TheLayer.SetSetting("IconImage","OffsetX",this.OffsetX);
		this.TheLayer.SetSetting("IconImage","OffsetY",this.OffsetY);
		
		this.TheLayer.Repaint(); 
	};
	
	TheImage.src=TheURL; // triggers the load
};
/**
* Icon for point features
* @public
* @param TheURL - the URL of the image to use for all features
* @param OffsetX - horizontal offset in pixels to go from the upper left of the image to the hot spot
* @param OffsetY - vertical offset in pixels to go from the upper left of the image to the hot spot
*/
CMLayer.prototype.SetFeatureIconImage=function(FeatureIndex,TheURL,OffsetX,OffsetY) 
{
	var TheImage=new Image(); 
	TheImage.FeatureIndex=FeatureIndex;
	TheImage.Loaded=false;
	TheImage.TheLayer=this;
	TheImage.OffsetX=OffsetX;
	TheImage.OffsetY=OffsetY;
	
	TheImage.onload=function() 
	{ 
		this.Loaded=true;
		
		this.TheLayer.SetFeatureSetting("IconImage","URL",this.FeatureIndex,this.TheURL);
		this.TheLayer.SetFeatureSetting("IconImage","TheImage",this.FeatureIndex,this);
		this.TheLayer.SetFeatureSetting("IconImage","OffsetX",this.FeatureIndex,this.OffsetX);
		this.TheLayer.SetFeatureSetting("IconImage","OffsetY",this.FeatureIndex,this.OffsetY);
		
		this.TheLayer.Repaint(); 
	};
	
	TheImage.src=TheURL; // triggers the load
};
//******************************************************************
// Functions for subclasses to call (protected)
// These functions are called by subclass layers to draw various types
// of data into the view of the scene.
//******************************************************************

/**
* Utility function to set the label font into the context and to return
* the current font size for bounds calculations.
* @protected
* @param TheView - the view to paint into
* @param FeatureIndex - the feature to paint
* @param RefX - X reference coordinate value
* @param RefY - Y reference coordinate value
* @param TheIconImage
*/
CMLayer.prototype.SetupLabelFont=function(TheView,FeatureIndex)
{
	//************************
	// get font size to determine the width of the text to place the label
	
	var LabelFont=this.GetFeatureSetting("Text","font",FeatureIndex,"Arial 12px");
//	var FontFace="Arial";
	var FontSize=20;
	
	if ((LabelFont!=null)&&(LabelFont!=undefined)) // make sure we strip "px" off the size
	{
		var Index=LabelFont.indexOf("px");
		if (Index!=-1)
		{
			var Temp=LabelFont.substring(0,Index);
			var Index2=Temp.lastIndexOf(" ");
			if (Index2!=-1) Temp=Temp.substring(Index2+1);
			FontSize=parseInt(Temp);
		}
	}
	// setup the font
	var TheContext=TheView.GetContext();
	
	TheContext.font = LabelFont;
	
	return(FontSize); // this is returned so calling code does not have to deal with it.
}
/**
* Utility function to paint a point at a reference coordinate
* @protected
* @param TheView - the view to paint into
* @param FeatureIndex - the feature to paint
* @param RefX - X reference coordinate value
* @param RefY - Y reference coordinate value
* @param SelectedOnly - false unless we are drawing the selected point on top of everything else
*/
CMLayer.prototype.PaintPoint=function(TheView,FeatureIndex,RefX,RefY,SelectedOnly,MouseOverOnly)
{
	if (this.IsFeatureVisible(TheView,FeatureIndex))
	{
//		this.UpdateSettingsFromAttributes(FeatureIndex);
		
		// convert the ref coordinate to pixels
		
		var Result=TheView.GetPixelFromRef(RefX,RefY);
		var PixelX=Result.PixelX;
		var PixelY=Result.PixelY;
	
		// find the correct image for this point, if any
		
		var TheFeatureImage=this.GetSettingGroup("IconImage");
		
		var TheFeatureImage=this.GetFeatureSettingGroup("IconImage",FeatureIndex,TheFeatureImage);
		
		//
		
		if (TheFeatureImage.TheImage!==undefined) // there is an icon image
		{
			var Bounds={
				XMin:PixelX+TheFeatureImage.OffsetX,
				XMax:PixelX+TheFeatureImage.OffsetX+TheFeatureImage.TheImage.width,
				YMin:PixelY+TheFeatureImage.OffsetY,
				YMax:PixelY+TheFeatureImage.OffsetY+TheFeatureImage.TheImage.height
			};
			if (TheView.CheckCollision(Bounds)==false)
			{
				TheView.AddToCollisions(Bounds);
				
				TheView.PaintImage(TheFeatureImage.TheImage,PixelX+TheFeatureImage.OffsetX,PixelY+TheFeatureImage.OffsetY);
			}
		}
		else // must be a mark
		{
			var TheStyle=null;
			
			if (SelectedOnly)
			{
				TheStyle=this.GetStyle(TheView,0,"SelectedStyle");
			}
			else if (MouseOverOnly)
			{
				TheStyle=this.GetStyle(TheView,0,"MouseOverStyle");
			}
			else
			{
				TheStyle=this.GetStyle(TheView);
				TheStyle=this.GetFeatureSettingGroup("Style",FeatureIndex,TheStyle);
			}
			if (TheStyle!==null) { TheView.SetStyle(TheStyle,true); }
//			if (TheStyle!==null) { TheView.SetStyle(TheStyle,true); }
			
//			var TheFeatureStyle=this.GetFeatureSettingGroup("Style",i,TheStyle);
				
//			if (TheFeatureStyle!==null) { TheView.SetStyle(TheFeatureStyle,false); }
			
			// draw the mark
			
			var TheSize=this.GetFeatureSetting("Mark","Size",FeatureIndex,5);
			var TheType=this.GetFeatureSetting("Mark","Type",FeatureIndex,CMLayer.MARK_CIRCLE);
			
			var HalfSize=TheSize/2;
			
			var Bounds={
				XMin:PixelX-HalfSize,
				XMax:PixelX+HalfSize,
				YMin:PixelY-HalfSize,
				YMax:PixelY+HalfSize
			};
			if (TheView.CheckCollision(Bounds)==false)
			{
				TheView.AddToCollisions(Bounds);
			
				switch (TheType)
				{
				case CMLayer.MARK_CIRCLE:
					TheView.PaintCircle(PixelX,PixelY,HalfSize);
					break;
				case CMLayer.MARK_SQUARE:
					TheView.PaintRect(PixelX-HalfSize,PixelX+HalfSize,PixelY-HalfSize,PixelY+HalfSize);
					break;
				case CMLayer.MARK_TRIANGLE:
					{
						var TheRefSize=TheView.GetRefWidthFromPixelWidth(TheSize);
						
						var Triangle=CMUtilities.GetRegularPolygon(3,TheRefSize/2,RefX,RefY, 180);
						TheView.PaintRefPoly(Triangle[0],Triangle[1],true);
					}
					break;
				case CMLayer.MARK_STAR:
					{
						var TheRefSize=TheView.GetRefWidthFromPixelWidth(TheSize);
						
						var Star=CMUtilities.GetStar(5,TheRefSize,RefX,RefY, 0);
						TheView.PaintRefPoly(Star[0],Star[1],true);
					}
					break;
				}
			}
			if (TheStyle!=undefined) TheView.RestoreStyle();
		}
		//************************************************************************
		// see if there is a label
		var TheLabel=this.GetFeatureSetting("Text","Text",FeatureIndex,null);
		
		if (TheLabel!=null) // draw the label if there is one
		{
			TheLabel=""+TheLabel; // make sure the label is a string
			
			var LabelStyle=this.GetStyle(TheView,0,"Text") ;
			
			if (LabelStyle!=undefined) TheView.SetStyle(LabelStyle);
			
			var FontSize=this.SetupLabelFont(TheView,FeatureIndex);
			
			var TheContext=TheView.GetContext();
			
			// setup the parameters for drawing
			var OffsetX=10;
			var OffsetY=10;
			
			var Direction=this.GetFeatureSetting("LabelBox","Position",FeatureIndex,"T");
			OffsetX=this.GetFeatureSetting("LabelBox","OffsetX",FeatureIndex,OffsetX);
			OffsetY=this.GetFeatureSetting("LabelBox","OffsetY",FeatureIndex,OffsetY);
			
			// find the array of lines
			
			var Lines=TheLabel.split("<br>");
			
			// find the dimensions
			
			var Height=Lines.length*FontSize;
			var LineWidths=[];
			
			var MaxWidth=0;
			for (var i=0;i<Lines.length;i++)
			{
				Lines[i]=Lines[i].trim(); // remove any white space
				
				LineWidths.push(TheContext.measureText(Lines[i]).width);
				
				if (LineWidths[i]>MaxWidth) MaxWidth=LineWidths[i];
			}
			
			// find the x start position
			
			var PixelLeft=0;
			
			switch (Direction)
			{
			case "TL":
			case "L":
			case "BL":
				PixelLeft=PixelX-OffsetX-MaxWidth;
				break;
			case "T":
			case "B":
			default: // centered
				PixelLeft=PixelX-MaxWidth/2;
				break;
			case "BR":
			case "TR":
			case "R":
				PixelLeft=PixelX+OffsetY;
				break;
			}
			
			// find the y start position
			
			var PixelTop=0;

			switch (Direction)
			{
			case "TL":
			case "T":
			case "TR":
				PixelTop=PixelY-OffsetY-Height;
				break;
			case "R":
			case "L":
			default: // centered
				PixelTop=PixelY-Height/2;
				break;
			case "BR":
			case "B":
			case "BL":
				PixelTop=PixelY+OffsetY;
				break;
			}
			
			// find the x start position
			
			var X=PixelLeft;
			var Y=PixelTop+FontSize; // move to the first baseline
			
			var Bounds={
				XMin:X,
				XMax:X+MaxWidth,
				YMin:PixelTop,
				YMax:PixelTop+Height
			};

			var ViewRefExtent=TheView.GetBounds();
			var ViewPixelExtent={};
			
			ViewPixelExtent.XMin=TheView.GetPixelXFromRefX(ViewRefExtent.XMin);
			ViewPixelExtent.XMax=TheView.GetPixelXFromRefX(ViewRefExtent.XMax);
			ViewPixelExtent.YMin=TheView.GetPixelYFromRefY(ViewRefExtent.YMax);
			ViewPixelExtent.YMax=TheView.GetPixelYFromRefY(ViewRefExtent.YMin);
			
			if (CMUtilities.BoundsOverlap(ViewPixelExtent,Bounds))
			{
				if (TheView.CheckCollision(Bounds)==false)
				{
					TheView.AddToCollisions(Bounds);
				
					var TheLabelBoxStyle=this.GetStyle(TheView,0,"LabelBox");
					
					if ((TheLabelBoxStyle!=null)&&((TheLabelBoxStyle.fillStyle!=undefined)||(TheLabelBoxStyle.strokeStyle!=undefined)))
					{
						var Padding=this.GetSetting("LabelBox","Padding");
						
						TheView.SetStyle(TheLabelBoxStyle);
						TheView.PaintRect(Bounds.XMin-Padding,Bounds.XMax+Padding,Bounds.YMin-Padding,Bounds.YMax+Padding);
						TheView.RestoreStyle();
					}
	
					TheContext.textBaseline="bottom"; 
					
					for (var i=0;i<Lines.length;i++)
					{
						switch (Direction)
						{
						case "TL":
						case "L":
						case "BL":
							X=PixelLeft+MaxWidth-LineWidths[i];
							break;
						case "T":
						case "B":
						default: // centered
							X=PixelLeft+((MaxWidth-LineWidths[i])/2);
							break;
						case "BR":
						case "TR":
						case "R":
							X=PixelLeft;
							break;
						}
						TheContext.strokeText(Lines[i],X,Y);
						TheContext.fillText(Lines[i],X,Y);
							
						Y+=FontSize;
					}
				} // end if collision
				
			}
			if (LabelStyle!=undefined) TheView.RestoreStyle();
		}
	}
};
/*
* Paints a single ploy (polygon or polyline) into the view
* @protected
* @param TheView - the view to paint into
* @param FeatureIndex - the feature to paint
* @param TheCoordinates - Array of coordinate values of the form TheCoordinates[NumCoordinates][0=x,1=y]
* @param Closed - true to paint a polygon (closed), false for polyline (open)
* @param SelectedOnly - false unless we are drawing the selected point on top of everything else
* @param MouseOverOnly - false unless we are drawing the mouse over on top of everything else
*/
CMLayer.prototype.PaintRefArea=function(TheView,FeatureIndex,AreaCoordinates,TheType,SelectedOnly,MouseOverOnly)
{
	if (this.IsFeatureVisible(TheView,FeatureIndex))
	{
		var TheStyle=null;
		
		if (SelectedOnly)
		{
			TheStyle=this.GetStyle(TheView,0,"SelectedStyle");
		}
		else if (MouseOverOnly)
		{
			TheStyle=this.GetStyle(TheView,0,"MouseOverStyle");
		}
		else
		{
			TheStyle=this.GetFeatureSettingGroup("Style",FeatureIndex,null);
		}
		
		if (TheStyle!==null) { TheView.SetStyle(TheStyle,true); }
		
		var TheGeo=this.GetParent(CMGeo);
		
		//
		if ((TheGeo.GetProjectorType()==CMGeo.PROJECTOR_DYNAMIC)) // project and paint the area
		{
			var TheProjector=TheGeo.GetProjector();
			
			if (TheType==CMDatasetVector.TYPE_POLYGONS) // fill the area
			{
				var ProjectedAreaCoordinates=TheProjector.ProjectAreaFromGeographic(AreaCoordinates,TheType);
			
				TheView.PaintRefArea(ProjectedAreaCoordinates,true);
			}
			else
			{
				// stoke the area
				var ProjectedAreaCoordinates=TheProjector.ProjectAreaFromGeographic(AreaCoordinates,TheType);
			
				TheView.PaintRefArea(ProjectedAreaCoordinates,false,false,true);
			}
		}
		else // paint the area without project on the fly
		{
			if (TheType==CMDatasetVector.TYPE_POLYGONS) // fill the area
			{
				TheView.PaintRefArea(AreaCoordinates,true);
			}
			else
			{
				// stoke the area
				TheView.PaintRefArea(AreaCoordinates,false);
			}
		}
		if (TheStyle!==null) { TheView.RestoreStyle(TheStyle); }
	}
}

/**
* Should be called by default after a data set is loaded so the layer can have it's properties setup
* that are based on attributes.  Can be overriden by a caller or subclass.
* @override, @public
*/
// jjg - should be replaced by listener
CMLayer.prototype.OnLoad=function() 
{
	//this.UpdateSettingsFromAttributes();
	
	if (this.GetScene()==null)
	{
		alert("Sorry, you'll need to add the layer to the CanvasMap before you can load data.");
	}
	else 
	{
		this.GetScene().SetBoundsDirty();
		
		if (this.GetSetting("Layer","ZoomToBoundsOnLoad",false)==true)
		{
			var TheMainContainer=this.GetParent(CMMainContainer);
			
			TheMainContainer.ZoomToBounds(this.GetBounds());
		}
		
		this.GetScene().Repaint();
	}
}

//******************************************************************
// Mouse event handling
//******************************************************************
/*
* returns the feature index for the coordinate in projected space
* @override
* @param TheView - the view with the size of the canvas.
* @param RefX - the horiziontal porition of the reference coordinate for the mouse cursor 
* @param RefY - the vertical porition of the reference coordinate for the mouse cursor 
* @returns FeatureIndex - -1 if the coordinate is not in a feature
*/
CMLayer.prototype.In=function(TheView,RefX,RefY) 
{
	var FeatureIndex=-1;
	
	return(FeatureIndex);
};
/**
* Called when the mouse button is pressed down in a canvas.
* @override
* @param TheView - the view with the size of the canvas.
* @param RefX - the horiziontal porition of the reference coordinate for the mouse cursor 
* @param RefY - the vertical porition of the reference coordinate for the mouse cursor 
* @returns Used - true if the event was used, false otherwise.
*/
CMLayer.prototype.MouseDown=function(TheView,RefX,RefY) 
{
	var Used=false;
	
	if ((this.IsVisible(TheView))&&(this.GetSetting("Item","Status")>=CMItem.STATUS_SELECTABLE)&&
		((TheView.GetTool()==CMView.TOOL_INFO)||(TheView.GetTool()==CMView.TOOL_SELECT))) // check if we where clicked in
	{
		var FeatureIndex=this.In(TheView,RefX,RefY);
		
		if (FeatureIndex!=-1)
		{
			this.SetSelectedFeature(FeatureIndex);
			
			this.ShowInfoWindow(FeatureIndex,TheView,RefX,RefY);
			
			Used=true;
		}
	}
	return(Used);
};

/**
* Called when the mouse is moved over the canvas.  The layer should check if it
* can use the mouse over event and then return "true" if it used the event and
* "false" if it wants other layers to examine the event.
* @override
* @param TheView - the view with the size of the canvas.
* @param RefX - the horiziontal porition of the reference coordinate for the mouse cursor 
* @param RefY - the vertical porition of the reference coordinate for the mouse cursor 
* @param FeatureIndex - the feature the mouse has moved over.
*/
CMLayer.prototype.MouseMove=function(TheView,RefX,RefY) 
{
	var Used=false;
	
	if (this.GetSetting("Item","Status")>=CMItem.STATUS_SELECTABLE)
	{
		var MouseOverSetting=this.GetSettingGroup("MouseOverStyle");
		
		// only check for "In" if the user has set something for the mouse over style
		if ((MouseOverSetting!=undefined)&&(Object.keys(MouseOverSetting).length>0)) // have something in the mouse over settings
		{
			var FeatureIndex=this.In(TheView,RefX,RefY);
			
			if (FeatureIndex!==-1) 
			{
				this.MouseOver(TheView,RefX,RefY,FeatureIndex);
			}
			else 
			{
				this.ResetMouseOverFeature();
			}
		}
	}
	return(Used);
};
/**
* Called when the mouse button is released.
* @override
* @param TheView - the view with the size of the canvas.
* @param RefX - the horiziontal porition of the reference coordinate for the mouse cursor 
* @param RefY - the vertical porition of the reference coordinate for the mouse cursor 
* @returns Used - true if the event was used, false otherwise.
*/
CMLayer.prototype.MouseUp=function(TheView,RefX,RefY) 
{
	var Used=false;
	
	return(Used);
};
/**
* Called when the mouse is moved over a feature.  Returns "true" if it used the event and
* "false" if it wants other layers to examine the event.
* @override
* @param TheView - the view with the size of the canvas.
* @param RefX - the horiziontal porition of the reference coordinate for the mouse cursor 
* @param RefY - the vertical porition of the reference coordinate for the mouse cursor 
* @param FeatureIndex - the feature the mouse has moved over.
* @returns Used - true if the event was used, false otherwise.
*/
CMLayer.prototype.MouseOver=function(TheView,RefX,RefY,FeatureIndex) 
{
	var Used=false;
	
	this.SetMouseOverFeature(FeatureIndex);
	
	return(Used);
};

//******************************************************************
// CMLayer Painting Functions
//******************************************************************
/**
* Called when the layer is resized
* @override
* @param TheView - the view with the size of the canvas.
*/
CMLayer.prototype.Resize=function(TheView) 
{
}

//******************************************************************
// CMLayer Projection On Fly Functions
//******************************************************************

CMLayer.prototype.ProjectBounds=function(Bounds) 
{
	var Result=Bounds; // sssume no change
	
	// see if the coordinates 
	var TheGeo=this.GetParent(CMGeo);
	
	if (TheGeo.GetProjectorType()==CMGeo.PROJECTOR_DYNAMIC)
	{
		var TheProjector=TheGeo.GetProjector();
		
		var Temp=TheProjector.ProjectToGeographic(Bounds.XMin,Bounds.YMin);
		
		Result.XMin=Temp[0];
		Result.YMin=Temp[1];
		
		var Temp=TheProjector.ProjectToGeographic(Bounds.XMax,Bounds.YMax);
		
		Result.XMax=Temp[0];
		Result.YMax=Temp[1];
	}
	return(Result);
}

CMLayer.prototype.ProjectCoordinate=function(RefX,RefY,Elevation) 
{
	var Result=[RefX,RefY,Elevation];
	
	// see if the coordinates 
	var TheGeo=this.GetParent(CMGeo);
	
	if (TheGeo.GetProjectorType()==CMGeo.PROJECTOR_DYNAMIC)
	{
		var TheProjector=TheGeo.GetProjector();
		
		Result=TheProjector.ProjectToGeographic(RefX,RefY,Elevation);
	}
	return(Result);
}
CMLayer.prototype.ProjectorChanged=function() 
{
	this.Repaint();
}
//******************************************************************
// CMLayer Search Functions
//******************************************************************

/*
* Requests search results from a layer.  The scene calls this function
* @override
* @param SearchPhrase - the phrase to search for in the layer's features
* @param ResultsPanel - the DOM element to put the results of the search into (i.e. set innerHTML)
*/
CMLayer.prototype.GetSearchResults=function(SearchPhrase,ResultsPanel) 
{
}
/*
* Requests search results from a layer.  The scene calls this function
* @override
* @param FeatureIndex - 
* @param TheView - 
* @param RefX - 
* @param RefY - 
*/
CMLayer.prototype.ShowInfoWindow=function(FeatureIndex,TheView,RefX,RefY) 
{
	var TheHTML=this.GetFeatureSetting("Layer","InfoText",FeatureIndex,null);
	
	if (TheHTML!=null)
	{
		var InfoWindow=TheView.CreateInfoWindow("CMLayer.InfoWindow",RefX,RefY,this.GetInfoWindowWidth(),30,TheHTML);
		
		CMMainContainer.SetPopupWindow(InfoWindow);
	}
};
/*
* Gets an icon for a specified mark.  This is used by subclasses.
* @override
* @protected
* @returns - an element that can be placed next to the layer in the layer list (should be 16x16)
*/
CMLayer.prototype.GetMarkIcon=function(MarkType,FillStyle,StrokeStyle)
{
	TheIcon=document.createElement('CANVAS');
	TheIcon.className="CM_LayerListIconClass";
	TheIcon.TheLayer=this;
	TheIcon.style.borderColor="rgba(0,0,0,0)";
	
	TheIcon.width=16; // jjg, these should be in settings eventually
	TheIcon.height=16;
	
	var TheView=new CMView2D();
	TheView.Setup(TheIcon);
	
	TheView.SetStyle({fillStyle:FillStyle,strokeStyle:StrokeStyle});
	
	// draw the mark
	
	var PixelX=0;
	var PixelY=0;
	var HalfSize=7;
	var HalfSize=7;
	var TheSize=14;
	
	switch (MarkType)
	{
	case CMLayer.MARK_CIRCLE:
		TheView.PaintCircle(PixelX+HalfSize+1,PixelY+HalfSize+1,HalfSize);
		break;
	case CMLayer.MARK_SQUARE:
		TheView.PaintRect(PixelX+1,PixelX+TheSize,PixelY+1,PixelY+TheSize);
		break;
	case CMLayer.MARK_TRIANGLE:
		{
			var Triangle=CMUtilities.GetRegularPolygon(3,HalfSize+2,HalfSize+1,-HalfSize-2, 180);
			TheView.PaintRefPoly(Triangle[0],Triangle[1],true);
		}
		break;
	case CMLayer.MARK_STAR:
		{
			var Star=CMUtilities.GetStar(5,TheSize,HalfSize+1,-HalfSize-1, 36);
			TheView.PaintRefPoly(Star[0],Star[1],true);
		}
		break;
	}
	return(TheIcon);
}
/*
* Gets an icon for a specified mark.  
* @override
* @protected
* @returns - an element that can be placed next to the layer in the layer list (should be 16x16)
*/
CMLayer.prototype.GetIcon=function()
{
	var TheIcon=this.GetSetting("IconImage","TheImage",1);
	
	if (TheIcon==1) // create an icon
	{
		var TheType=this.GetSetting("Mark","Type",CMLayer.MARK_SQUARE);
		
		var FillStyle=this.GetSetting("Style","fillStyle");
		var StrokeStyle=this.GetSetting("Style","strokeStyle");
		
		if (CMUtilities.IsDefined(TheType)==false) // does not have a mark, just draw a box that is colored the same 
		{
			TheIcon=document.createElement('div');
			TheIcon.className="CM_LayerListIconClass";
			TheIcon.TheLayer=this;
			
			if (CMUtilities.IsDefined(FillStyle))  TheIcon.style.backgroundColor=FillStyle;
			else TheIcon.style.backgroundColor=undefined;
			
			if (CMUtilities.IsDefined(StrokeStyle))  TheIcon.style.borderColor=StrokeStyle;
			else TheIcon.style.borderColor=undefined;
		}
		else // has a mark (circle, triangle, etc.)
		{
			TheIcon=this.GetMarkIcon(TheType,FillStyle,StrokeStyle);
		}
	}
	return(TheIcon);
}


CMLayer.prototype.AddMetadataMenuItem=function(ThePopupMenu)
{
	//********************************************
	// Add the metadata option if there is metadata
	
	var Metadata=this.GetSetting("Layer","Metadata",undefined);
	
	if (Metadata!=undefined)
	{
		var MetadataElement=document.createElement('div');
		MetadataElement.className="CM_LayerListPopupMenuItem";
		MetadataElement.setAttribute("id","CM_ZoomToExtentMenuItem");
		MetadataElement.innerHTML="Metadata";
		
		MetadataElement.TheLayer=this;
		
		MetadataElement.ThePopupMenu=ThePopupMenu;
		
		MetadataElement.addEventListener('click', function(event)
		{
			this.ThePopupMenu.style.visibility="hidden";
			
			var TheDialog=new CMDialog("CMMetadataDialog",400,200,false,"Metadata");
			
			// get the panel to add the label to
			var ThePanel=TheDialog.GetBodyElement();
			
			var Label=document.createElement("div");
			Label.style.margin="10px";
			Label.innerHTML=this.TheLayer.GetSetting("Layer","Metadata",undefined);
			ThePanel.appendChild(Label); // add the dialog element to the document

			event.stopPropagation();
		});
		ThePopupMenu.appendChild(MetadataElement);
	}
}
/*
* Allows the layer to customize the popup menu that apperas when the user right-clicks on the layer in the layer list
* @override
* @protected
* @param ThePopupMenu - 
*/
CMLayer.prototype.FillPopupMenu=function(ThePopupMenu)
{
	//********************************************
	// create the delete menu item
	var DeleteElement=document.createElement('div');
	DeleteElement.setAttribute("id","CM_DeleteElementMenuItem");
	DeleteElement.className="CM_LayerListPopupMenuItem";
	
	DeleteElement.innerHTML="Delete";
	
	DeleteElement.TheLayer=this;
	DeleteElement.ThePopupMenu=ThePopupMenu;
	
	DeleteElement.addEventListener('click', function(event)
	{
		var TheScene=this.TheLayer.GetParent(CMScene);
		
		this.ThePopupMenu.style.visibility= "hidden";
		var LayerIndex=TheScene.GetLayerIndex(this.TheLayer);
		TheScene.DeleteLayer(LayerIndex);
			
		event.stopPropagation();
	});
	
	ThePopupMenu.appendChild(DeleteElement);
	
	//********************************************
	// See if we should add the "Zoom to" option
	
	if (this.GetBounds()!=null)
	{
		var ZoomToElement=document.createElement('div');
		ZoomToElement.className="CM_LayerListPopupMenuItem";
		ZoomToElement.setAttribute("id","CM_ZoomToExtentMenuItem");
		ZoomToElement.innerHTML="Zoom To This Layer";
		
		ZoomToElement.TheLayer=this;
		
		ZoomToElement.ThePopupMenu=ThePopupMenu;
		
		ZoomToElement.addEventListener('click', function(event)
		{
			this.ThePopupMenu.style.visibility="hidden";
			
			var TheBounds=this.TheLayer.GetBounds();
			
			var TheScene=this.TheLayer.GetParent(CMScene);
												 
			var TheView=TheScene.GetView(0);
			
			TheView.ZoomToBounds(TheBounds);
			
			event.stopPropagation();
		});
		ThePopupMenu.appendChild(ZoomToElement);
	}
	
	this.AddMetadataMenuItem(ThePopupMenu);
	
	//********************************************
	// See if we should add the "Attribute Table" option
	
/*	var TheCanvasMap=this.GetParent(CMMainContainer);
	
	var AttributePanelElement=TheCanvasMap.GetElement(CanvasMap.ATTRIBUTE_PANEL);
	var NumRows=this.GetNumAttributeRows();
	
	if ((AttributePanelElement!=null)&&(NumRows>0))
	{
		var AttributeMenuItemElement=document.createElement('div');
		AttributeMenuItemElement.className="CM_LayerListPopupMenuItem";
		AttributeMenuItemElement.setAttribute("id","CM_AttributeMenuItem");
		AttributeMenuItemElement.innerHTML="Attribute Table";
		
		AttributeMenuItemElement.TheLayer=this;
		AttributeMenuItemElement.ThePopupMenu=ThePopupMenu;
		AttributeMenuItemElement.AttributePanelElement=AttributePanelElement;
		
		AttributeMenuItemElement.addEventListener('click', function(event)
		{
			this.ThePopupMenu.style.visibility="hidden";
			
			var TheCanvasMap=this.TheLayer.GetParent(CMMainContainer);
			 
			this.TheLayer.UpdateAttributeTable(this.AttributePanelElement);
			
			event.stopPropagation();
		});
		ThePopupMenu.appendChild(AttributeMenuItemElement);
	}*/
}
