/******************************************************************************************************************
* CMLayerGraticule
*
* by: Jim Graham
* 
* This class required a level of complexity beyond what I initially expected.
* The problem is that we have to:
*	- Draw graticule lines based on the zoom level, projection, and area viewed
*  - Paint coordinates along the border based on where the graticules intersect with the border
*  - Paint coordinates along the outside of the graticule grid if it is visible (i.e. does not 
*    intersect with the border.
*
* graticles are made visible by setting the overall style to transparent.  The graticules must be drawn to get the intersections for the exterior and border
* @module CMLayerGraticule 
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/

CMLayerGraticule.MAX_SPACING_DEGREES=15;

/**
* Below are the additional settings definitions for STLayers.  Setings for the basic drawing style and text
* style are inherited from STItem
* @public, @settings
*/
CMLayerGraticule.SettingDefintions=
{
	Graticule: // actual graticules within the map
	{
		East: { Name:"East",Type:CMBase.DATA_TYPE_FLOAT, Default:-180 },
		West: { Name:"West",Type:CMBase.DATA_TYPE_FLOAT, Default:180 },
		North: { Name:"North",Type:CMBase.DATA_TYPE_FLOAT, Default:90 },
		South: { Name:"South",Type:CMBase.DATA_TYPE_FLOAT, Default:-90 },
		Spacing: { Name:"Spacing",Type:CMBase.DATA_TYPE_FLOAT, Default:200 }, /// desired spacing for the graticules in pixels
		DegreeSpacing: { Name:"Degree Spacing",Type:CMBase.DATA_TYPE_FLOAT, Default:-1 }, // only implemented for 3d 
		TextSpacing: { Name:"Text Spacing",Type:CMBase.DATA_TYPE_FLOAT, Default:360 }, // only implemented for 3d 
	},
	Border: // area around the outside of them map (covers the edges of the map)
	{
		// standard HTML 5 settings except the defaults may change and sometimes the available settings will change between each settings group
		strokeStyle: { Name:"Line Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(0,0,0)" },
		fillStyle: { Name:"Fill Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(255,255,255)" },
		lineWidth: { Name:"Width",Type:CMBase.DATA_TYPE_INTEGER, Default:3 },
		BorderWidth: { Name:"Border Width",Type:CMBase.DATA_TYPE_FLOAT, Default:12 },
	},
	BorderText: // coordinates along the border when the map intersects with the border
	{
		font: { Name:"Font",Type:CMBase.DATA_TYPE_FONT, Default:"14px Arial" },
		strokeStyle: { Name:"Line Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(0,0,0)" },
		lineWidth: { Name:"Width",Type:CMBase.DATA_TYPE_INTEGER, Default:1 },
		fillStyle: { Name:"Fill Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(0,0,0)" },
	},
};
//**************************************************************************************************
// Constructor
//**************************************************************************************************
/*
* Constructor
* @public, @constructs
*/
function CMLayerGraticule() 
{
	CMLayer.call(this);
	
	// 
	this.TimeSlices[0].Settings.Graticule=	
	{
	};
	this.TimeSlices[0].Settings.Border=	
	{
	};
	this.TimeSlices[0].Settings.BorderText=	
	{
	};
	//***********************************************
	// temp properties for the currnet clipping box setup in SetupBoundsPoly()
	
	this.Bounds=null;
	//this.BoundsSpacing=10;
	
	//***********************************************
	// Temp properties for finding the min/max lat/lon for the graticules
	
	// coordinates for the bounds in the current projection
	this.ClippingBoxProjectedEastings=null;
	this.ClippingBoxProjectedNorthings=null;
	
	this.ClippingBoxProjectedWestCoordinates=null;
	this.ClippingBoxProjectedEastCoordinates=null;
	
	this.ClippingBoxProjectedNorthCoordinates=null;
	this.ClippingBoxProjectedSouthCoordinates=null;
	
	//***********************************************
	// intersections between the graticules (lines) and the sides of the border
	this.TopIntersectionCoordinates=null;
	this.BottomIntersectionCoordinates=null;
	
	this.LeftIntersectionCoordinates=null;
	this.RightIntersectionCoordinates=null;
	
	//********************************************************
	this.Lines3D=null;
	
}
CMLayerGraticule.prototype=Object.create(CMLayer.prototype); // inherit prototype functions from PanelBase()

CMLayerGraticule.prototype.contructor=CMLayerGraticule; // override the constructor to go to ours
//******************************************************************
// Private definitions
//******************************************************************
/*
* These are the posible options for the intervals between values on the border.
*/
CMLayerGraticule.DegreeQuantized=[
	 30, // 30 degrees
	 15,
	 10,
	 5,
	 2,
	 1,
	 0.5, // 30 minutes
	 0.25, // 15 minutes
	 1/6, // 10 minutes
	 1/12, // 5 minutes
	 1/30, // 2 minutes
	 1/60, // 1 minute
	 1/120, // 30 seconds
	 1/240, // 15 seconds
	 1/360, // 10 seconds
	 1/720, // 5 seconds
	 1/1800, // 2 seconds
	 1/3600 // 1 second
];

//******************************************************************
// Private Functions for computing the dimensions of a set of graticules
// that match the viewing area
//******************************************************************

/**
* Find the coordinate where a general line segment intersects a vertical line segment.
* This is provided because it is much faster than finding the intersection
* of two line segments.
* @private
* @param P1X - horiziontal value of the first coordinate for the line segment
* @param P1Y - vertical value of the first coordinate for the line segment
* @param P2X - horiziontal value of the second coordinate for the line segment
* @param P2Y - vertical value of the second coordinate for the line segment
* @param RefX - horiziontal value of the vertical line segment
* @param RefTop - top of the vertical line segment
* @param RefBottom - bottom of the vertical line segment
* @param Coordinates - array that will collect the intersections with the vertical line segment
* @param Latitude - latitude to be added so we can label the intersection on the border.
* @return - JSON object with {Easting:RefX,Northing:RefTop} if an intersection was found, null otherwise
*/
CMLayerGraticule.SegmentIntersectsAVertical=function(P1X,P1Y,P2X,P2Y,RefX,RefTop,RefBottom,Coordinates,Latitude)
{
	var Result=null;
	
	if (((P1X<RefX)&&(P2X<RefX))||((P1X>RefX)&&(P2X>RefX))|| //segment is to one side
		((P1Y<RefBottom)&&(P2Y<RefBottom))||((P1Y>RefTop)&&(P2Y>RefTop))) // segement is above or below 
	{
	}
	else // segment's bounding rectangle overlaps the vertical segement
	{
		if (P2X==P1X) // vertical line
		{
			if (P1X==RefX)  // colinear
			{
	//			Result=[RefX,RefTop]; // vertical colinear with the vertical (not sure of correct Y value)
				Coordinates.push({Easting:RefX,Northing:P1Y});
			}
		}
		else if (P1X==P2X) // vertical line
		{
			// already know the line is in the same box as the horiziontal so it must intersect
			Coordinates.push({Easting:RefX,Northing:P1Y,Longitude:Longitude});
		}
		else
		{
			// find the equation of the line
			var m=(P2Y-P1Y)/(P2X-P1X); // y=mx+b, b=y-mx
			var b=P1Y-(m*P1X);
			
			// find the intersection with our vertical
			
			var y=m*RefX+b;
			if ((y<RefTop)&&(y>RefBottom)) 
			{
				Coordinates.push({Easting:RefX,Northing:y,Latitude:Latitude});
			}
		}
	}
	return(Result);
}
/**
* Find the coordinate where a general line segment intersects a horizontal line segment.
* This is provided because it is much faster than finding the intersection
* of two line segments.
* @private
* @P1X - horiziontal value of the first coordinate for the line segment
* @P1Y - vertical value of the first coordinate for the line segment
* @P2X - horiziontal value of the second coordinate for the line segment
* @P2Y - vertical value of the second coordinate for the line segment
* @RefLeft - left value of the horizontal line segment
* @RefRight - right value of the horizontal line segment
* @RefY - vertical value of the horizontal line segment
* @Coordinates - array that will collect the intersections with the vertical line segment
* @Longitude - latitude to be added so we can label the intersection on the border.  Optional
* @return - JSON object with {Easting:RefX,Northing:RefTop} if an intersection was found, null otherwise
*/
CMLayerGraticule.SegmentIntersectsAHorizontal=function(P1X,P1Y,P2X,P2Y,RefLeft,RefRight,RefY,Coordinates,Longitude)
{
	var Result=null;
	
	if (((P1X<RefLeft)&&(P2X<RefLeft))||((P1X>RefRight)&&(P2X>RefRight))|| //segment is to one side
		((P1Y<RefY)&&(P2Y<RefY))||((P1Y>RefY)&&(P2Y>RefY))) // segement is above or below 
	{
	}
	else // segment's bounding rectangle overlaps the vertical segement
	{
		if (P2Y==P1Y) // horizontal line
		{
			if (P1Y==RefY) // colinear
			{
	//			Result=[RefX,RefTop]; // vertical colinear with the vertical (not sure of correct Y value)
				
				Coordinates.push({Easting:P1X,Northing:RefY,Longitude:Longitude});
			}
		}
		else if (P1X==P2X) // vertical line
		{
			// already know the line is in the same box as the horiziontal so it must intersect
			Coordinates.push({Easting:P1X,Northing:RefY,Longitude:Longitude});
/*			if ((P1X<=RefRight)&&(P1X>=RefLeft)) 
			{
				Coordinates.push({Easting:RefLeft,Northing:RefY,Longitude:Longitude});
			}
*/		}
		else
		{
			// find the equation of the line
			var m=(P2Y-P1Y)/(P2X-P1X); // y=mx+b, b=y-mx
			var b=P1Y-(m*P1X);
			
			// find the intersection with our vertical
			
			var x=(RefY-b)/m; // x=(y-b)/m
			if ((x>=RefLeft)&&(x<=RefRight)) 
			{
				Coordinates.push({Easting:x,Northing:RefY,Longitude:Longitude});
			}
		}
	}
	return(Result);
}
/**
* Find the coordinates where a general polygon intersects a vertical line segment.
* This is provided because it is much faster than finding the intersection
* of two line segments.
* @private
* @Xs - X values for the polygons coordinates
* @Ys - Y values for the polygons coordinates
* @RefX - horiziontal value of the vertical line segment
* @RefTop - top of the vertical line segment
* @RefBottom - bottom of the vertical line segment
* @Coordinates - array that will collect the intersections with the vertical line segment
* @Longitude - latitude to be added so we can label the intersection on the border.  Optional
* @return - JSON object with {Easting:RefX,Northing:RefTop} if an intersection was found, null otherwise
*/
CMLayerGraticule.GetPolyIntersectionsWithVertical=function(Xs,Ys,RefLeft,RefTop,RefBottom,Coordinates)
{
	var Result=null;
	
	var LastIndex=Xs.length-1;
	
	for (var i=0;(i<LastIndex);i++)
	{
		Result=CMLayerGraticule.SegmentIntersectsAVertical(Xs[i],Ys[i],Xs[i+1],Ys[i+1],RefLeft,RefTop,RefBottom,Coordinates);
	}
	Result=CMLayerGraticule.SegmentIntersectsAVertical(Xs[LastIndex],Ys[LastIndex],Xs[0],Ys[0],RefLeft,RefTop,RefBottom,Coordinates);
	return(Result);
}
/**
* Find the coordinates where a general polygon intersects a horizontal line segment.
* This is provided because it is much faster than finding the intersection
* of two line segments.
* @private
* @Xs - X values for the polygons coordinates
* @Ys - Y values for the polygons coordinates
* @RefLeft - left value of the horizontal line segment
* @RefRight - right value of the horizontal line segment
* @RefY - vertical value of the horizontal line segment
* @Coordinates - array that will collect the intersections with the vertical line segment
* @Longitude - latitude to be added so we can label the intersection on the border.  Optional
* @return - JSON object with {Easting:RefX,Northing:RefTop} if an intersection was found, null otherwise
*/
CMLayerGraticule.GetPolyIntersectionsWithHorizontal=function(Xs,Ys,RefLeft,RefRight,RefY,Coordinates)
{
	var Result=null;
	
	var LastIndex=Xs.length-1;
	
	for (var i=0;(i<LastIndex);i++)
	{
		Result=CMLayerGraticule.SegmentIntersectsAHorizontal(Xs[i],Ys[i],Xs[i+1],Ys[i+1],RefLeft,RefRight,RefY,Coordinates);
	}
	Result=CMLayerGraticule.SegmentIntersectsAHorizontal(Xs[LastIndex],Ys[LastIndex],Xs[0],Ys[0],RefLeft,RefRight,RefY,Coordinates);
	
	return(Result);
}
/**
* Get the max bounds for the projector then make it smaller if specified by the user.
* @private
*//*
CMLayerGraticule.prototype.SetupBounds=function(ClippingPoly)
{
	var Xs=ClippingPoly.Xs;
	var Ys=ClippingPoly.Ys;
	
	this.Bounds={};
	
	ClippingPoly.Xs[1]=this.GetSetting("Graticule","East",Xs[1]);
	ClippingPoly.Xs[0]=this.GetSetting("Graticule","West",Xs[0]);
	ClippingPoly.Ys[0]=this.GetSetting("Graticule","North",Ys[0]);
	ClippingPoly.Ys[1]=this.GetSetting("Graticule","South",Ys[1]);
}*/
CMLayerGraticule.prototype.AddEastWestCoordinate=function(TheProjector,Latitude,BoundsWest,BoundsEast)
{
	var WestResult=TheProjector.ProjectFromGeographic(BoundsWest,Latitude);
	var EastResult=TheProjector.ProjectFromGeographic(BoundsEast,Latitude);
	
	if (WestResult!=null)
	{
		var WestCoordinate={Easting:WestResult[0],Northing:WestResult[1]};
		
		this.ClippingBoxProjectedWestCoordinates.push(WestCoordinate);
	}
	if (EastResult!=null)
	{
		var EastCoordinate={Easting:EastResult[0],Northing:EastResult[1]};
		
		this.ClippingBoxProjectedEastCoordinates.push(EastCoordinate);
	}
}
CMLayerGraticule.prototype.AddNorthSouthCoordinate=function(TheProjector,Longitude,BoundsNorth,BoundsSouth)
{
	var NorthResult=TheProjector.ProjectFromGeographic(Longitude,BoundsNorth);
	var SouthResult=TheProjector.ProjectFromGeographic(Longitude,BoundsSouth);
	
	if (NorthResult!=null)
	{
		var NorthCoordinate={Easting:NorthResult[0],Northing:NorthResult[1]};
		
		this.ClippingBoxProjectedNorthCoordinates.push(NorthCoordinate);
	}
	if (SouthResult!=null)
	{
		var SouthCoordinate={Easting:SouthResult[0],Northing:SouthResult[1]};
	
		this.ClippingBoxProjectedSouthCoordinates.push(SouthCoordinate);
	}
}
//******************************************************************
// Functions for finding the valid area to render
//******************************************************************

/**
* Approach:
* - Gets the bounds from the projector (North,South,East,West)
* - Create a temp series of coordinates that are for the north, south, east, and west edges of the specified clipping box
* - Convert the edge coordintes into a series:
*  - this.ClippingBoxProjectedEastings[]
*  - this.ClippingBoxProjectedNorthings[]
*
*  - this.ClippingBoxProjectedNorthCoordinates[]
*  - this.ClippingBoxProjectedWestCoordinates[]
*  - this.ClippingBoxProjectedNorthCoordinates[]
*  - this.ClippingBoxProjectedSouthCoordinates[]
* @private
*/
CMLayerGraticule.prototype.SetupBoundsPoly=function(ClippingPoly)
{
//	if (this.ClippingBoxProjectedEastings===null) // bounds have not been setup
	{
		// get the desired spacing (max)
		var BoundsSpacing=CMLayerGraticule.MAX_SPACING_DEGREES;
		
		// make sure we have a projector
		var TheGeo=this.GetParent(CMGeo);
		
		var TheProjector=TheGeo.GetProjector();
			
		// initialize bounds poly (coordinates to go around the bounds)
		this.ClippingBoxProjectedEastings=[];
		this.ClippingBoxProjectedNorthings=[];
		
		// 
		this.ClippingBoxProjectedWestCoordinates=[];
		this.ClippingBoxProjectedEastCoordinates=[];
		
		this.ClippingBoxProjectedNorthCoordinates=[];
		this.ClippingBoxProjectedSouthCoordinates=[];
	
		//*************************************************************************************
		// go from the south to the north adding coordinates to the east and west coordinate arrays
		
		var XMinMax=CMUtilities.GetMinMax(ClippingPoly.Xs);
		var YMinMax=CMUtilities.GetMinMax(ClippingPoly.Ys);

		var Latitude=YMinMax.Min;
		var Range=YMinMax.Max-YMinMax.Min;
		var Factor=Range/10;
		
		for (var i=0;i<=10;i++)
		{
			this.AddEastWestCoordinate(TheProjector,Latitude,XMinMax.Min,XMinMax.Max);
			
			Latitude+=Factor; // change the latitude to be on a BoundsSpacing line
		}
		
		var StartingRemainder=Latitude%BoundsSpacing;
		
		if (StartingRemainder!=0) // have some extra stuff in the south
		{
			this.AddEastWestCoordinate(TheProjector,Latitude,XMinMax.Min,XMinMax.Max);
			
			Latitude=Latitude-StartingRemainder; // change the latitude to be on a BoundsSpacing line
		}
		while (Latitude<=YMinMax.Max)
		{
			this.AddEastWestCoordinate(TheProjector,Latitude,XMinMax.Min,XMinMax.Max);
			
			Latitude+=BoundsSpacing;
		}
		var EndingRemainder=YMinMax.Max%BoundsSpacing;
		
		if (EndingRemainder!=0) // have some extra stuff in the south
		{
			Latitude=YMinMax.Max;
		
			this.AddEastWestCoordinate(TheProjector,Latitude,XMinMax.Min,XMinMax.Max);
		}
		
		//*************************************************************************************
		// go from the south to the north adding coordinates to the east and west coordinate arrays
		
		var Longitude=XMinMax.Min;
		var Range=XMinMax.Max-XMinMax.Min;
		var Factor=Range/10;
		
		for (var i=0;i<=10;i++)
		{
			this.AddNorthSouthCoordinate(TheProjector,Longitude,YMinMax.Max,YMinMax.Min);
			
			Longitude+=Factor;
		}
		var StartingRemainder=Longitude%BoundsSpacing;
		
		if (StartingRemainder!=0) // have some extra stuff in the south
		{
			this.AddNorthSouthCoordinate(TheProjector,Longitude,YMinMax.Max,YMinMax.Min);
			
			Longitude=Longitude-StartingRemainder;
		}
		while (Longitude<=XMinMax.Max) 
		{
			this.AddNorthSouthCoordinate(TheProjector,Longitude,YMinMax.Max,YMinMax.Min);
			
			Longitude+=BoundsSpacing;
		}
		var EndingRemainder=XMinMax.Max%BoundsSpacing;
		
		if (EndingRemainder!=0) // have some extra stuff in the south
		{
			Longitude=XMinMax.Max;
		
			this.AddNorthSouthCoordinate(TheProjector,Longitude,YMinMax.Max,YMinMax.Min);
		}
		
		// make the max bounds into two x and y arrays
		
		for (var i=0;i<this.ClippingBoxProjectedWestCoordinates.length;i++) // south to north along the left side
		{
			this.ClippingBoxProjectedEastings.push(this.ClippingBoxProjectedWestCoordinates[i].Easting);
			this.ClippingBoxProjectedNorthings.push(this.ClippingBoxProjectedWestCoordinates[i].Northing);
		}
		for (var i=0;i<this.ClippingBoxProjectedNorthCoordinates.length;i++) // weat to east along the top
		{
			this.ClippingBoxProjectedEastings.push(this.ClippingBoxProjectedNorthCoordinates[i].Easting);
			this.ClippingBoxProjectedNorthings.push(this.ClippingBoxProjectedNorthCoordinates[i].Northing);
		}
		for (var i=0;i<this.ClippingBoxProjectedEastCoordinates.length;i++) // north to south down the right side
		{
			var Index=this.ClippingBoxProjectedEastCoordinates.length-1-i;
			
			this.ClippingBoxProjectedEastings.push(this.ClippingBoxProjectedEastCoordinates[Index].Easting);
			this.ClippingBoxProjectedNorthings.push(this.ClippingBoxProjectedEastCoordinates[Index].Northing);
		}
		for (var i=0;i<this.ClippingBoxProjectedSouthCoordinates.length;i++) // east to west across the bototm
		{
			var Index=this.ClippingBoxProjectedSouthCoordinates.length-1-i;
			
			this.ClippingBoxProjectedEastings.push(this.ClippingBoxProjectedSouthCoordinates[Index].Easting);
			this.ClippingBoxProjectedNorthings.push(this.ClippingBoxProjectedSouthCoordinates[Index].Northing);
		}
	}
}
//******************************************************************
// Private painting functions
//******************************************************************

/**
* Simple utility to paint the meridian segments and, if they intersect with a meridian, record them
* @private
*/
CMLayerGraticule.prototype.PaintMeridianSegment=function(TheView,PX1,PY1,PX2,PY2,ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude) 
{
	TheView.PaintRefLine(PX1,PY1,PX2,PY2);
	
	CMLayerGraticule.SegmentIntersectsAHorizontal(PX1,PY1,PX2,PY2,ViewRefLeft,ViewRefRight,ViewRefTop,this.TopIntersectionCoordinates,Longitude);
	
	CMLayerGraticule.SegmentIntersectsAHorizontal(PX1,PY1,PX2,PY2,ViewRefLeft,ViewRefRight,ViewRefBottom,this.BottomIntersectionCoordinates,Longitude);
}

/**
* Simple utility to paint the meridian segments and, if they intersect with a meridian, record them
* @private
*/
CMLayerGraticule.prototype.PaintParallelSegment=function(TheView,PX1,PY1,PX2,PY2,ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude) 
{
	TheView.PaintRefLine(PX1,PY1,PX2,PY2);
	
	CMLayerGraticule.SegmentIntersectsAVertical(PX1,PY1,PX2,PY2,ViewRefLeft,ViewRefTop,ViewRefBottom,this.LeftIntersectionCoordinates,Latitude);
	
	CMLayerGraticule.SegmentIntersectsAVertical(PX1,PY1,PX2,PY2,ViewRefRight,ViewRefTop,ViewRefBottom,this.RightIntersectionCoordinates,Latitude);
}

//**************************************************************************
// CMBase Functions
//**************************************************************************

CMLayerGraticule.prototype.CMLayer_SettingsChanged=CMLayer.prototype.SettingsChanged;

CMLayerGraticule.prototype.SettingsChanged=function()
{
	this.CMLayer_SettingsChanged();
	
	this.ClippingBoxProjectedEastings=null;
}
// we want the CMItem definitions but not the CMLayer definitions
CMLayerGraticule.prototype.CMLayer_GetSettingsDefinitions=CMLayer.prototype.GetSettingsDefinitions;

CMLayerGraticule.prototype.GetSettingsDefinitions=function() 
{
	var Result=this.CMLayer_GetSettingsDefinitions();
	
	for (Key in CMLayerGraticule.SettingDefintions)
	{
		Result[Key]=CMLayerGraticule.SettingDefintions[Key];
	}

	return(Result); 
}
//***********************************************************************
// CMItem functions
//***********************************************************************
/*CMLayerGraticule.prototype.CMItem_SetVisible=CMItem.prototype.SetVisible;

CMLayerGraticule.prototype.SetVisible=function(Flag)
{
	if (this.GetVisible()!=Flag)
	{
		this.CMItem_SetVisible(Flag);
		
		if (typeof(CM3Scene)!== 'undefined')
		{
			var TheScene=this.GetParent(CM3Scene);
			
			if ((TheScene!=null)&&(this.Lines3D!=null))
			{
				var TheMesh=this.Lines3D;
				//var TheMesh=this.TheFeatureMeshes[i];
				
				if (Flag)  // add the mesh to the TheOGLScene to make it visible
				{
					TheScene.GetGeo().AddOGLObjectToGeo(TheMesh);
				}
				else // remove the meshfrom the TheOGLScene to make it invisible
				{
					TheScene.GetGeo().RemoveOGLObjectFromGeo(TheMesh);
				}
			}
		}
	}
}*/
//***********************************************************************
// Private functions to find the Min/Max Lat/Lon range and the spacing
// between the lines.
//***********************************************************************

/*
* Function to find the desired spacing between the lines
* @private
*/
CMLayerGraticule.prototype.GetSpacing=function(TheView,TheProjector,CenterLongitude,CenterLatitude,LongitudeRange)
{	
	if (LongitudeRange<0)
	{
		throw("Sorry, the LongitudeRange cannot be less than zero");
	}
	// Find the center of the min/max lat/lon (target area) 
//	var CenterLongitude=(MaxLon+MinLon)/2;
//	var CenterLatitude=(MaxLat+MinLat)/2;
	
	// if we don't reset the longitude to zero, the calcs below make one lon on one side of the international date line
	// and the other on the opposite side.  This leads to huge horizontal distances
//	var CenterLongitude=0; // jjg - this fixes the calc for the center longitude when it is near the international date tline

	// 	move east by 10% of the range of values to find two points in geographic (20% or 1/5 of the total area)
	var Lon1=CenterLongitude-(LongitudeRange/10);
	var Lon2=CenterLongitude+(LongitudeRange/10);
	
//	var Lat1=CenterLatitude-(LatitudeRange/10);
//	var Lat2=CenterLatitude+(LatitudeRange/10);
	
	// Project these two points.  
	var Projected1=TheProjector.ProjectFromGeographic(Lon1,CenterLatitude);
	var Projected2=TheProjector.ProjectFromGeographic(Lon2,CenterLatitude);
	
	// use the latitudes to avoid the issues with longitude crossing the date line
//	var Projected1=TheProjector.ProjectFromGeographic(CenterLatitude,Lat1);
//	var Projected2=TheProjector.ProjectFromGeographic(CenterLatitude,Lat2);
	
	var DesiredWidth=this.GetSetting("Graticule","Spacing",200);
	
	var SpacingIndex=0;
	
	if ((Projected1!=null)&&(Projected2!=null))
	{
		// Find the change in longitude (20% of the total) and the corresponding change in projected space
		var ChangeInLongitude=LongitudeRange/5;
		
		var ChangeInProjected=Projected2[0]-Projected1[0];
		if (ChangeInProjected<0) ChangeInProjected=-ChangeInProjected;
		
		// Convert the projected range to pixels
		var ChangeInPixels=TheView.GetPixelWidthFromRefWidth(ChangeInProjected);
		
		// the change in longitude over the change in pixels is the approximate degrees/pixels at the center of the area
		var DegreesPerPixel=ChangeInLongitude/ChangeInPixels;
		
		// the gradiule width in degreess is found by mulitplying DegreesPerPixels by the desired spacing of graticules in pixels
		var GraticuleWidthInDegrees=DegreesPerPixel*DesiredWidth; // deisred width
		
		// now find the spacing that is greater than the desired
		while ((CMLayerGraticule.DegreeQuantized[SpacingIndex]>GraticuleWidthInDegrees)&&
			(SpacingIndex+1<CMLayerGraticule.DegreeQuantized.length))
		{
			SpacingIndex++;
		}
		if ((SpacingIndex+1<CMLayerGraticule.DegreeQuantized.length)&&(SpacingIndex<0)) 
		{
			SpacingIndex--;
		}
	}
	var Spacing=CMLayerGraticule.DegreeQuantized[SpacingIndex];

	return(Spacing);
}
/**
* Find the coordinates at the intersection of the view bounds and the clipping bounds.
* - Finds where the clipping bounds intersects with the view edges
* - Adds any corners of the view bounds that are in the clipping bounds.
*
* @private
*/
CMLayerGraticule.prototype.GetVisibleBounds=function(TheView,TheProjector) 
{
	//**********************************************************************
	// setup the coordinates of the view edges
	var TheCanvasElement=TheView.GetCanvasElement();
	
	var ViewRefLeft=TheView.GetRefXFromPixelX(0);
	var ViewRefRight=TheView.GetRefXFromPixelX(TheCanvasElement.width);
	var ViewRefTop=TheView.GetRefYFromPixelY(0);
	var ViewRefBottom=TheView.GetRefYFromPixelY(TheCanvasElement.height);

	//**********************************************************************
	// Fill the ValidCoordinates with coordinates that are: intersection of the projected bounds and the view sides,
	//  view corners in the projected bounds, corners of the projected bounds that are inside the view
	//**********************************************************************
	var ValidCoordinates=[]; // coordinates that intersect projection bounds and view are placed here
	
	//**********************************************************************
	// find the intersecions between the bounds of the projection and each of the view edges
	// resulting points are placed in the ValidCoordinates array
	
	CMLayerGraticule.GetPolyIntersectionsWithVertical(this.ClippingBoxProjectedEastings,this.ClippingBoxProjectedNorthings,
			ViewRefLeft,ViewRefTop,ViewRefBottom,ValidCoordinates); // left side
	
	CMLayerGraticule.GetPolyIntersectionsWithVertical(this.ClippingBoxProjectedEastings,this.ClippingBoxProjectedNorthings,
			ViewRefRight,ViewRefTop,ViewRefBottom,ValidCoordinates); // right side
	
	CMLayerGraticule.GetPolyIntersectionsWithHorizontal(this.ClippingBoxProjectedEastings,this.ClippingBoxProjectedNorthings,
			ViewRefLeft,ViewRefRight,ViewRefTop,ValidCoordinates); // top
	
	CMLayerGraticule.GetPolyIntersectionsWithHorizontal(this.ClippingBoxProjectedEastings,this.ClippingBoxProjectedNorthings,
			ViewRefLeft,ViewRefRight,ViewRefBottom,ValidCoordinates); // bottom
	
	//**********************************************************************
	// get the view points that are in the bounds of the projector by checking if they are InsideAPolygon where the polygon is the projector bounds
	
	// add top left if inside
	if (CMUtilities.InsideAPolygon(ViewRefLeft,ViewRefTop,this.ClippingBoxProjectedEastings,this.ClippingBoxProjectedNorthings,this.ClippingBoxProjectedEastings.length))
	{
		ValidCoordinates.push({Easting:ViewRefLeft,Northing:ViewRefTop});
	}
	
	// add top right if inside
	if (CMUtilities.InsideAPolygon(ViewRefRight,ViewRefTop,this.ClippingBoxProjectedEastings,this.ClippingBoxProjectedNorthings,this.ClippingBoxProjectedEastings.length))
	{
		ValidCoordinates.push({Easting:ViewRefRight,Northing:ViewRefTop});
	}
	
	// add bottom right if inside
	if (CMUtilities.InsideAPolygon(ViewRefRight,ViewRefBottom,this.ClippingBoxProjectedEastings,this.ClippingBoxProjectedNorthings,this.ClippingBoxProjectedEastings.length))
	{
		ValidCoordinates.push({Easting:ViewRefRight,Northing:ViewRefBottom});
	}
	
	// add bottom left if inside
	if (CMUtilities.InsideAPolygon(ViewRefLeft,ViewRefBottom,this.ClippingBoxProjectedEastings,this.ClippingBoxProjectedNorthings,this.ClippingBoxProjectedEastings.length))
	{
		ValidCoordinates.push({Easting:ViewRefLeft,Northing:ViewRefBottom});
	}
	
	//**********************************************************************
	// paint all the valid coordinates that surround the area
	if (false) //
	{
		var NewStyle={fillStyle:"Blue"};
		TheView.SetStyle(NewStyle,false);
	
		for (var i=0;i<ValidCoordinates.length;i++)
		{
			TheView.PaintRefCircle(ValidCoordinates[i].Easting,ValidCoordinates[i].Northing,20);
		}
	}
	//**********************************************************************
	// project the valid coordinates back to geographic

	var GeoCoordinates=[];
	for (var i=0;i<ValidCoordinates.length;i++)
	{
		var Result=TheProjector.ProjectToGeographic(ValidCoordinates[i].Easting,ValidCoordinates[i].Northing);

		if (Result==null) 
		{
		//	CMMainContainer.Error("CMLayerGraticule.Paint() problem back projecting");
		}
		else
		{
			var GeoCoordinate={Longitude:Result[0],Latitude:Result[1]};
		
			GeoCoordinates.push(GeoCoordinate);
		}
	}
	return(GeoCoordinates);
}


/**
* Returns an object with the lat/lon range defined for the speciifed ClippingBounds.
*
* Approach:
* - SetupBoundsPoly() - sets up the arrays with the coordinates that go around the bounds
*  - this.ClippingBoxProjectedEastings[]
*  - this.ClippingBoxProjectedNorthings[]
*
* - And with arrays for the coordinates along the north, west, east, and southern edges of the bounds 
*  - this.ClippingBoxProjectedEastCoordinates[]
*  - this.ClippingBoxProjectedWestCoordinates[]
*  - this.ClippingBoxProjectedNorthCoordinates[]
*  - this.ClippingBoxProjectedSouthCoordinates[]
*
* - Create a ValidCoordinates array that contains:
*  - intersection of the projected bounds and the view sides,
*  - View coordinates that are inside the projectors valid bounds
*  - corners of the projected bounds that are inside the view
*
*
* @private
*/

CMLayerGraticule.prototype.GetLatLonRangeFromClippingBounds=function(TheView,ClippingPoly,TheProjector) 
{
	var Result=null;
	
//	var Spacing=-1;
//	if (LatLonRange!=null) Spacing=LatLonRange.Spacing;
	
	//**********************************************************************
	// setup globals for ClipMin/Max Lat/Lon
	//this.SetupBounds(ClippingPoly);
	
	//**********************************************************************
	// Setup the overall bounds for the allowable projection area
	// - Set up the arrays with the coordinates that go around the bounds
	//  - this.ClippingBoxProjectedEastings[]
	//  - this.ClippingBoxProjectedNorthings[]
	//
	// - And with arrays for the coordinates along the north, west, east, and southern edges of the bounds 
	//  - this.ClippingBoxProjectedEastCoordinates[]
	//  - this.ClippingBoxProjectedWestCoordinates[]
	//  - this.ClippingBoxProjectedNorthCoordinates[]
	//  - this.ClippingBoxProjectedSouthCoordinates[]
	//
	this.SetupBoundsPoly(ClippingPoly);
	
	var ClippingXMinMax=CMUtilities.GetMinMax(ClippingPoly.Xs);
	var ClippingYMinMax=CMUtilities.GetMinMax(ClippingPoly.Ys);
	
	if (false) // debugging to paint red border around bounds projected (looks good)
	{
		var NewStyle={strokeStyle:"Red",lineWidth:3};
		TheView.SetStyle(NewStyle,false);
	
		TheView.PaintRefPoly(this.ClippingBoxProjectedEastings,this.ClippingBoxProjectedNorthings,true,false,true);
	}

	//**********************************************************************
	// Get the bounds of the projection wihtin the view
	var GeoCoordinates=this.GetVisibleBounds(TheView,TheProjector);
	
	for (var i=0;i<GeoCoordinates.length;i++)
	{
		var TheCoordinate=GeoCoordinates[i];
		
		if (TheCoordinate==undefined)
		{
			var j=12;
		}
		if (ClippingXMinMax.Max==180) // 180 could get flipped to -180
		{
			if (TheCoordinate.Longitude<-179) TheCoordinate.Longitude=180;
		}
		if (ClippingXMinMax.Min==-180) // 180 could get flipped to +180
		{
			if (TheCoordinate.Longitude>179) TheCoordinate.Longitude=-180;
		}
	}
	//**********************************************************************
	// setup the coordinates of the view edges
	var TheCanvasElement=TheView.GetCanvasElement();
	
	var ViewRefLeft=TheView.GetRefXFromPixelX(0);
	var ViewRefRight=TheView.GetRefXFromPixelX(TheCanvasElement.width);
	var ViewRefTop=TheView.GetRefYFromPixelY(0);
	var ViewRefBottom=TheView.GetRefYFromPixelY(TheCanvasElement.height);

	//**********************************************************************
	// setup flags to determine if we are using the corners of the clipping bounds
	
	var ViewXs=[ViewRefLeft,ViewRefRight,ViewRefRight,ViewRefLeft];
	var ViewYs=[ViewRefTop,ViewRefTop,ViewRefBottom,ViewRefBottom];
	
	if ((this.ClippingBoxProjectedWestCoordinates.length>0)&&
		(this.ClippingBoxProjectedEastCoordinates.length>0))
	{
		SWCoordinate=this.ClippingBoxProjectedWestCoordinates[0];
		NWCoordinate=this.ClippingBoxProjectedWestCoordinates[this.ClippingBoxProjectedWestCoordinates.length-1];
		
		SECoordinate=this.ClippingBoxProjectedEastCoordinates[0];
		NECoordinate=this.ClippingBoxProjectedEastCoordinates[this.ClippingBoxProjectedEastCoordinates.length-1];
		
		var UseBoundsNW=null;
		var UseBoundsNE=null;
		var UseBoundsSE=null;
		var UseBoundsSW=null;
		
		try
		{
			UseBoundsNW=CMUtilities.InsideAPolygon(NWCoordinate.Easting,NWCoordinate.Northing,ViewXs,ViewYs,ViewXs.length);
			UseBoundsNE=CMUtilities.InsideAPolygon(NECoordinate.Easting,NECoordinate.Northing,ViewXs,ViewYs,ViewXs.length);
			UseBoundsSE=CMUtilities.InsideAPolygon(SECoordinate.Easting,SECoordinate.Northing,ViewXs,ViewYs,ViewXs.length);
			UseBoundsSW=CMUtilities.InsideAPolygon(SWCoordinate.Easting,SWCoordinate.Northing,ViewXs,ViewYs,ViewXs.length);
		}
		catch (err)
		{
			throw(err);
		}
		//**********************************************************************
		if (false) // paint the corners of the projector baounds that are in the view for debugging
		{
			var NewStyle={fillStyle:"Green"};
			TheView.SetStyle(NewStyle,false);
		
			if (UseBoundsNW)
			{
				TheView.PaintRefCircle(NWCoordinate.Easting,NWCoordinate.Northing,10);
			}
			if (UseBoundsNE)
			{
				TheView.PaintRefCircle(NECoordinate.Easting,NECoordinate.Northing,10);
			}
			if (UseBoundsSE)
			{
				TheView.PaintRefCircle(SECoordinate.Easting,SECoordinate.Northing,10);
			}
			if (UseBoundsSW)
			{
				TheView.PaintRefCircle(SWCoordinate.Easting,SWCoordinate.Northing,10);
			}
		}
		var NewStyle={fillStyle:"Black"};
		TheView.SetStyle(NewStyle);
		
		//**********************************************************************
		// find the min/max in geographic coordinates
		
		var MinLat;
		var MaxLat;
		var MinLon;
		var MaxLon;
		
		if (GeoCoordinates.length>0)
		{
			MinLat=GeoCoordinates[0].Latitude;
			MaxLat=GeoCoordinates[0].Latitude;
			MinLon=GeoCoordinates[0].Longitude;
			MaxLon=GeoCoordinates[0].Longitude;
			
			for (var i=1;i<GeoCoordinates.length;i++)
			{
				if (GeoCoordinates[i].Latitude<MinLat) MinLat=GeoCoordinates[i].Latitude;
				if (GeoCoordinates[i].Latitude>MaxLat) MaxLat=GeoCoordinates[i].Latitude;
				if (GeoCoordinates[i].Longitude<MinLon) MinLon=GeoCoordinates[i].Longitude;
				if (GeoCoordinates[i].Longitude>MaxLon) MaxLon=GeoCoordinates[i].Longitude;
			}
		}
		//**********************************************************************
		// expand the min/max lat/lon based on the extreme coordinates 
		
		if ((UseBoundsNW)|(UseBoundsNE))
		{
			if ((MaxLat==undefined)||(isNaN(MaxLat))||(ClippingYMinMax.Max>MaxLat)) MaxLat=ClippingYMinMax.Max;
		}
		if ((UseBoundsSW)|(UseBoundsSE))
		{
			if ((MinLat==undefined)||(isNaN(MinLat))||(ClippingYMinMax.Min<MinLat)) MinLat=ClippingYMinMax.Min;
		}
		
		if ((UseBoundsNW)|(UseBoundsSW))
		{
			if ((MinLon==undefined)||(isNaN(MinLon))||(ClippingXMinMax.Min<MinLon)) MinLon=ClippingXMinMax.Min;
		}
		if ((UseBoundsNE)|(UseBoundsSE))
		{
			if ((MaxLon==undefined)||(isNaN(MaxLon))||(ClippingXMinMax.Max>MaxLon)) MaxLon=ClippingXMinMax.Max;
		}
		
		if ((MinLat==undefined)||(MaxLat==undefined)||(MinLon==undefined)||(MaxLon==undefined))
		{
			//window.alert("Min/Max Lat/Lon are undefined");
		}
		else
		{
			
			Result={
				MinLon:MinLon,
				MaxLon:MaxLon,
				MinLat:MinLat,
				MaxLat:MaxLat,
			};
		}
	}
	return(Result);
}
//*****************************************************************************************
// Private function to get the grid of coordinates to draw
//*****************************************************************************************
/**
* Creates a 2d grid of projtected coordinates to paint based on the specified range of
* lat/lon values
*
* @private
*/
CMLayerGraticule.prototype.GetProjectedCoordinateGrid=function(TheView,LatLonRange,TheProjector) 
{
	var MinLon=LatLonRange.MinLon;
	var MaxLon=LatLonRange.MaxLon;
	var MinLat=LatLonRange.MinLat;
	var MaxLat=LatLonRange.MaxLat;
	var Spacing=LatLonRange.Spacing;
	
	var Coordinates=[];
	var YIndex=0;
	
	// start at the top
	var Latitude=MaxLat;
	while (Latitude>=MinLat)
	{
		Coordinates[YIndex]=[];
		var XIndex=0;
		
		var Longitude=MinLon;
		while (Longitude<=MaxLon) // draw lines of longitude (6 degrees each)
		{
			// get the easting and northing
			//if (Longitude==0) Longitude=0.000000001;
			var Result=TheProjector.ProjectFromGeographic(Longitude,Latitude);
			
			if (Result!=null)
			{
				var Coordinate1={Easting:Result[0],Northing:Result[1]};
				
				// save the lat/lon for later
				Coordinate1.Longitude=Longitude;
				Coordinate1.Latitude=Latitude;
				
				// save the coordinate in the grid
				Coordinates[YIndex][XIndex]=Coordinate1;
			}
			
			// move the longitude right by Spacing or the remainder if we are at the top
			if (XIndex==0) // just added the first line at the left side
			{
				if (Longitude!=MaxLon) // did not just draw the right line
				{
					// if there is a remainder, move to an integral line based on Spacing
					var Remainder=Longitude%Spacing;
					if (Remainder!=0)
					{
						Longitude=Math.ceil(Longitude/Spacing)*Spacing;
					}
					else // otherwise, move to the next Longitude right based on Spacing
					{
						Longitude+=Spacing;
						
						if (Longitude>MaxLon) // overshot on the first column of cells
						{
							Longitude=MaxLon; // make sure we draw them
						}
					}
				}
			}
			else // not at the first line
			{
				if (Longitude!=MaxLon) // did not just draw the right line
				{
					Longitude+=Spacing; // move right based on Spacing
					
					if (Longitude>MaxLon) // went past the bottom, move to the bottom
					{
						Longitude=MaxLon;
					}
				}
				else // special case when we just drew the right line
				{
					Longitude+=Spacing; // forces an exit
				}
			}
			XIndex++;
		}
		
		// move the latitude down by Spacing or the remainder if we are at the top
		if (YIndex==0) // at the top
		{
			// if there is a remainder, move to an integral line based on Spacing
			var Remainder=Latitude%Spacing;
			if (Remainder!=0)
			{
				Latitude=Math.floor(Latitude/Spacing)*Spacing;
			}
			else // otherwise, move to the next latitude down based on Spacing
			{
				Latitude-=Spacing;
			}
		}
		else // not at the top line
		{
			if (Latitude!=MinLat) // did not just draw the bottom line
			{
				Latitude-=Spacing; // move down based on Spacing
				
				if (Latitude<MinLat) // went past the bottom, move to the bottom
				{
					Latitude=MinLat;
				}
			}
			else // special case when we just draw the bottom line
			{
				Latitude-=Spacing; // forces an exit
			}
			
		}
		YIndex++;
	}
	return(Coordinates);
}
//***********************************************************************
// Proteced Painting functions
//***********************************************************************
/**
* Paint one set of graticules based on a clipping bounds in geographic
*
* 
* @protected
*/
CMLayerGraticule.prototype.PaintCoordinateGrid=function(TheView,LatLonRange,Coordinates,TheProjector) 
{
	var MinLon=LatLonRange.MinLon;
	var MaxLon=LatLonRange.MaxLon;
	var MinLat=LatLonRange.MinLat;
	var MaxLat=LatLonRange.MaxLat;
	var Spacing=LatLonRange.Spacing;
	
	//
	
	var TheCanvasElement=TheView.GetCanvasElement();
	
	var BorderWidth=this.GetSetting("Border","BorderWidth");
	
	var ViewRefLeft=TheView.GetRefXFromPixelX(BorderWidth);
	var ViewRefRight=TheView.GetRefXFromPixelX(TheCanvasElement.width-BorderWidth);
	var ViewRefTop=TheView.GetRefYFromPixelY(BorderWidth);
	var ViewRefBottom=TheView.GetRefYFromPixelY(TheCanvasElement.height-BorderWidth);

	//******************************************************
	// draw the graticule lines
	//******************************************************
	// this gets the main style for the gratidule lines and the text in the map
	var TheStyle=this.GetStyle(TheView);
	
	if (TheStyle!=undefined) TheView.SetStyle(TheStyle,false);
	
	// if we got any coordinates, draw the grid of lines
	if (Coordinates.length!=0)
	{
		//******************************************************
		// draw the lines of longitude (meridians)
		
		var Longitude=MinLon;
		
		var NumColumns=Coordinates[0].length;
		var NumRows=Coordinates.length;
		for (var ColumnIndex=0;ColumnIndex<NumColumns;ColumnIndex++) // draw lines of longitude (6 degrees each)
		{
			if (Coordinates.length>=2)
			{
				for (var RowIndex=0;RowIndex<NumRows-1;RowIndex++)
				{
					var Coordinate1=Coordinates[RowIndex][ColumnIndex];
					var Coordinate2=Coordinates[RowIndex+1][ColumnIndex];
					
					if ((Coordinate1!=undefined)&&(Coordinate2!=undefined))
					{
						var Lat1=Coordinate1.Latitude;
						var Lon1=Coordinate1.Longitude;
						
						var Lat2=Coordinate2.Latitude;
						var Lon2=Coordinate2.Longitude;
						
						var LatRange=Lat2-Lat1;
						var LatFactor=LatRange/10;
						
						var Lat=Lat1;
						var StartCoordinate=TheProjector.ProjectFromGeographic(Lon1,Lat);
						
						for (var i=0;i<10;i++)
						{
							Lat+=LatFactor;
							
							var EndCoordinate=TheProjector.ProjectFromGeographic(Lon1,Lat);
							
							if (EndCoordinate!=undefined)
							{
								this.PaintMeridianSegment(TheView,StartCoordinate[0],StartCoordinate[1],
									EndCoordinate[0],EndCoordinate[1],ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude);
								
								StartCoordinate=EndCoordinate;
							}
						}
					}
				}
			}
/*			if (Coordinates.length<=1) // do nothing
			{
			}
			else if ((Coordinates.length==2)) // straight line
			{
				TheView.PaintRefLine(Coordinates[0][ColumnIndex].Easting,Coordinates[0][ColumnIndex].Northing,
					Coordinates[1][ColumnIndex].Easting,Coordinates[1][ColumnIndex].Northing);
			}
			else // at least 3, maybe more
			{
				try 
				{
					var PreviousCoordinate=Coordinates[0][ColumnIndex];
					var Coordinate1=Coordinates[1][ColumnIndex];
					var Coordinate2=Coordinates[2][ColumnIndex];
					var NextCoordinate=null;
				}
				catch(err) {
					CMMainContainer.Error(err.message);
				}
				
				//*********************************************************************************
				// Paint the first part of the graticule as a 3-point Bezier curve
				
				if ((PreviousCoordinate!=undefined)&&(Coordinate1!=undefined)&&(Coordinate2!=undefined))
				{
					// get the Bezier points for the start of the curve
					var Points=CMUtilityBezier.GetThreePoint3D(10,
							PreviousCoordinate.Easting,PreviousCoordinate.Northing,0,
							Coordinate1.Easting,Coordinate1.Northing,0,
							Coordinate2.Easting,Coordinate2.Northing,0);
					
					// paint from the parallel to the first coordinate on the Bezier curve
					this.PaintMeridianSegment(TheView,PreviousCoordinate.Easting,PreviousCoordinate.Northing,
							Points[0][0],Points[1][0],ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude);
					
					// paint the rest of the segements in the Bezier curve
					for (var i=0;i<Points[0].length-1;i++)
					{
						this.PaintMeridianSegment(TheView,Points[0][i],Points[1][i],Points[0][i+1],Points[1][i+1],
							ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude);
					}
					
					// paint the segment from the bezier to the coordinate
					this.PaintMeridianSegment(TheView,Coordinate1.Easting,Coordinate1.Northing,
							Points[0][Points[0].length-1],Points[1][Points[0].length-1],
							ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude);
				}
				var PreviousCoordinate=Coordinate1;
				
				//*********************************************************************************
				// Paint the middle part of the graticule as a 4-point Bezier curve
				
				for (var RowIndex=1;RowIndex<NumRows-2;RowIndex++)
				{
					var NextCoordinate=Coordinates[RowIndex+2][ColumnIndex];
					
					if ((PreviousCoordinate!=undefined)&&(NextCoordinate!=undefined)&&(Coordinate1!=undefined)&&(Coordinate2!=undefined))
					{
						var Points=CMUtilityBezier.GetFourPoint2D(10,
							PreviousCoordinate.Easting,PreviousCoordinate.Northing,
							Coordinate1.Easting,Coordinate1.Northing,
							Coordinate2.Easting,Coordinate2.Northing,
							NextCoordinate.Easting,NextCoordinate.Northing);
						
						// connect to the previous curve
						this.PaintMeridianSegment(TheView,
								Coordinate1.Easting,Coordinate1.Northing,
								Points[0][0],Points[1][0],
								ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude);
					
						// draw the middle of the curve
						for (var i=0;i<Points[0].length-1;i++)
						{
							this.PaintMeridianSegment(TheView,
									Points[0][i],Points[1][i],Points[0][i+1],Points[1][i+1],
									ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude);
						}
						// draw the end of the curve to the current point
						this.PaintMeridianSegment(TheView,
								Points[0][Points[0].length-1],Points[1][Points[0].length-1],
								Coordinate2.Easting,Coordinate2.Northing,
								ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude);
					}
					PreviousCoordinate=Coordinate1;
					Coordinate1=Coordinate2;
					Coordinate2=NextCoordinate;
				}
				//*********************************************************************************
				// Paint the last part of the graticule as a 3-point Bezier curve
				
				if ((PreviousCoordinate!=undefined)&&(Coordinate1!=undefined)&&(Coordinate2!=undefined))
				{
					// get the points on the Bezier curve
					var Points=CMUtilityBezier.GetThreePoint3D(10,
							Coordinate2.Easting,Coordinate2.Northing,0,
							Coordinate1.Easting,Coordinate1.Northing,0,
							PreviousCoordinate.Easting,PreviousCoordinate.Northing,0);
					
					// draw from the coordinate to the first point on the curve
					this.PaintMeridianSegment(TheView,
							Coordinate1.Easting,Coordinate1.Northing,
							Points[0][Points[0].length-1],Points[1][Points[0].length-1],
							ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude);
					
					// draw the points on the curve
					for (var i=Points[0].length-1;i>0;i--)
					{
						this.PaintMeridianSegment(TheView,
								Points[0][i],Points[1][i],Points[0][i-1],Points[1][i-1],
								ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude);
					}
					// paint from the Bezier curve to the last coordinate
					this.PaintMeridianSegment(TheView,
							Points[0][0],Points[1][0],Coordinate2.Easting,Coordinate2.Northing,
							ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Longitude);
				}
			}
	*/		Longitude+=Spacing;
		}
		
		//******************************************************
		// draw the lines of latitude (parallels)
		
		var NumPoints=9;
		var Latitude=MaxLat;
		
		for (var RowIndex=0;RowIndex<NumRows;RowIndex++)
		{
			if (Coordinates.length>=2)
			{
				for (var ColumnIndex=0;ColumnIndex<NumColumns-1;ColumnIndex++)
				{
					var Coordinate1=Coordinates[RowIndex][ColumnIndex];
					var Coordinate2=Coordinates[RowIndex][ColumnIndex+1];
					
					if ((Coordinate1!=undefined)&&(Coordinate2!=undefined))
					{
						var Lat1=Coordinate1.Latitude;
						var Lon1=Coordinate1.Longitude;
						
						var Lat2=Coordinate2.Latitude;
						var Lon2=Coordinate2.Longitude;
						
						var LonRange=Lon2-Lon1;
						var LonFactor=LonRange/10;
						
						var Lon=Lon1;
						var StartCoordinate=TheProjector.ProjectFromGeographic(Lon1,Lat1);
						
						for (var i=0;i<10;i++)
						{
							Lon+=LonFactor;
							
							var EndCoordinate=TheProjector.ProjectFromGeographic(Lon,Lat1);
							
							this.PaintParallelSegment(TheView,StartCoordinate[0],StartCoordinate[1],EndCoordinate[0],EndCoordinate[1],
									ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude);
	//						TheView.PaintRefLine(StartCoordinate[0],StartCoordinate[1],
	//							EndCoordinate[0],EndCoordinate[1]);
							
							StartCoordinate=EndCoordinate;
						}
					}
				}
			}
/*			if (Coordinates[RowIndex].length<=1) // do nothing
			{
			}
			else if (Coordinates[RowIndex].length==2) // straight line
			{
				TheView.PaintRefLine(Coordinates[RowIndex][0].Easting,Coordinates[RowIndex][0].Northing,
					Coordinates[RowIndex][1].Easting,Coordinates[RowIndex][1].Northing);
			}
			else
			{
				var PreviousCoordinate=Coordinates[RowIndex][0];
				var Coordinate1=Coordinates[RowIndex][1];
				var Coordinate2=Coordinates[RowIndex][2];
				var NextCoordinate=null;
				
				//*********************************************************************************
				// Paint the first part of the graticule as a 3-point Bezier curve
				
				if ((PreviousCoordinate!=undefined)&&(Coordinate1!=undefined)&&(Coordinate2!=undefined))
				{
					// get the Bezier points for the start of the curve
					var Points=CMUtilityBezier.GetThreePoint3D(10,
							PreviousCoordinate.Easting,PreviousCoordinate.Northing,0,
							Coordinate1.Easting,Coordinate1.Northing,0,
							Coordinate2.Easting,Coordinate2.Northing,0);
					
					// draw the line from the first line of longitude to the start of the Beizer points
					this.PaintParallelSegment(TheView,PreviousCoordinate.Easting,PreviousCoordinate.Northing,Points[0][0],Points[1][0],
							ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude);
					
					// draw the Bezier curve
					for (var i=0;i<Points[0].length-1;i++)
					{
						this.PaintParallelSegment(TheView,Points[0][i],Points[1][i],Points[0][i+1],Points[1][i+1],
								ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude);
					}
					// draw segment form the end of the Bezier points to the second meridian
					this.PaintParallelSegment(TheView,Points[0][Points[0].length-1],Points[1][Points[0].length-1],
							Coordinate1.Easting,Coordinate1.Northing,
							ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude);
				}
				//*********************************************************************************
				// Paint the middle part of the graticule as a 4-point Bezier curve
				
				for (var ColumnIndex=1;ColumnIndex<NumColumns-2;ColumnIndex++) // draw lines of longitude (6 degrees each)
				{
					var NextCoordinate=Coordinates[RowIndex][ColumnIndex+2];
					
					if ((PreviousCoordinate!=undefined)&&(NextCoordinate!=undefined)&&(Coordinate1!=undefined)&&(Coordinate2!=undefined))
					{
						// get the next Bezier curve
						var Points=CMUtilityBezier.GetFourPoint2D(10,
							PreviousCoordinate.Easting,PreviousCoordinate.Northing, // for start slope
							Coordinate1.Easting,Coordinate1.Northing, // start point
							Coordinate2.Easting,Coordinate2.Northing, // end point
							NextCoordinate.Easting,NextCoordinate.Northing); // for end slope
						
						// meridan to the first Beizer point
						this.PaintParallelSegment(TheView,
								Coordinate1.Easting,Coordinate1.Northing,Points[0][0],Points[1][0],
								ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude);
					
						// draw this curve
						for (var i=0;i<Points[0].length-1;i++)
						{
							this.PaintParallelSegment(TheView,
									Points[0][i],Points[1][i],Points[0][i+1],Points[1][i+1],
									ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude);
						}
						
						// from the last Bezier point to the next meridian
						this.PaintParallelSegment(TheView,
								Points[0][Points[0].length-1],Points[1][Points[0].length-1],Coordinate2.Easting,Coordinate2.Northing,
								ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude);
					}
					// shift the coordinates
					PreviousCoordinate=Coordinate1;
					Coordinate1=Coordinate2;
					Coordinate2=NextCoordinate;
				}
				//*********************************************************************************
				// Paint the last part of the graticule as a 3-point Bezier curve
				
				if ((PreviousCoordinate!=undefined)&&(Coordinate1!=undefined)&&(Coordinate2!=undefined))
				{
					var Points=CMUtilityBezier.GetThreePoint3D(10,
							Coordinate2.Easting,Coordinate2.Northing,0,
							Coordinate1.Easting,Coordinate1.Northing,0,
							PreviousCoordinate.Easting,PreviousCoordinate.Northing,0);
					
					// from meridian to the first Bezier point
					this.PaintParallelSegment(TheView,
							Coordinate1.Easting,Coordinate1.Northing,Points[0][Points[0].length-1],Points[1][Points[0].length-1],
							ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude);
							
					for (var i=Points[0].length-1;i>=1;i--)
					{
						this.PaintParallelSegment(TheView,
								Points[0][i],Points[1][i],Points[0][i-1],Points[1][i-1],
								ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude);
					}
					this.PaintParallelSegment(TheView,
							Points[0][0],Points[1][0],Coordinate2.Easting,Coordinate2.Northing,
							ViewRefLeft,ViewRefRight,ViewRefTop,ViewRefBottom,Latitude);
				}
			}
*/			Latitude-=Spacing;
		}
	}

}
/*
* Paint the coordinate labels along the grid lines
*/
CMLayerGraticule.prototype.PaintGridLabels=function(TheView,LatLonRange,Coordinates,ClippingPoly) 
{
	var MinLon=LatLonRange.MinLon;
	var MaxLon=LatLonRange.MaxLon;
	var MinLat=LatLonRange.MinLat;
	var MaxLat=LatLonRange.MaxLat;
	
	//********************************************************************************
	// draw the lat lons that are outside the bounds but do not intersect with the border

	if (Coordinates.length!=0)
	{
		var TextStyle=this.GetSettingGroup("Text");
		TheView.SetStyle(TextStyle,false);
		
		var FontSize=this.SetupLabelFont(TheView,-1); // setup the font but not from a feature
		var NumColumns=Coordinates[0].length;
		var NumRows=Coordinates.length;
		
		var RefFontHeight=TheView.GetRefHeightFromPixelHeight(FontSize);
		var RefFontCharWidth=-RefFontHeight/2;
	
		var XMinMax=CMUtilities.GetMinMax(ClippingPoly.Xs);
		var YMinMax=CMUtilities.GetMinMax(ClippingPoly.Ys);

		// paint latitudes along the west border
		if (MinLon==XMinMax.Min)
		{
			for (var Row=0;Row<NumRows;Row++)
			{
				var TheCoordinate=Coordinates[Row][0];
				
				if (TheCoordinate!=undefined)
				{
					var PixelWidth=TheView.GetTextWidthInPixels(Text);
					
					//var RefWidth=TheView.GetRefWidthFromPixelWidth(4);
					
					try
					{
						var Text=CMUtilities.GetDMSFromDD(TheCoordinate.Latitude,false,true);
					
						TheView.PaintRefText(Text,TheCoordinate.Easting-RefFontCharWidth,TheCoordinate.Northing+RefFontHeight/2,FontSize,"right",0);
					}
					catch(e)
					{
						throw(e);
					}
				}
			}
		}
		// paint latitudes along the east border
//		var RefWidth=TheView.GetRefWidthFromPixelWidth(4);
					
		if (MaxLon==XMinMax.Max)
		{
			for (var Row=0;Row<NumRows;Row++)
			{
				var TheCoordinate=Coordinates[Row][NumColumns-1];
				
				if (TheCoordinate!=undefined)
				{
					var Text=CMUtilities.GetDMSFromDD(TheCoordinate.Latitude,false,true);
					
					TheView.PaintRefText(Text,TheCoordinate.Easting+RefFontCharWidth,TheCoordinate.Northing+RefFontHeight/2,FontSize,"",0)
				}
			}
		}
		// paint longitudes along the south border
		if (MinLat==YMinMax.Min)
		{
			for (var Column=0;Column<NumColumns;Column++)
			{
				var TheCoordinate=Coordinates[NumRows-1][Column];
				
				if (TheCoordinate!=undefined)
				{
					var Text=CMUtilities.GetDMSFromDD(TheCoordinate.Longitude,true,true);
			
					TheView.PaintRefText(Text,TheCoordinate.Easting,TheCoordinate.Northing+RefFontHeight,FontSize,"center",0)
				}
			}
		}
		// paint longitudes along the north border
		if (MaxLat==YMinMax.Max)
		{
			for (var Column=0;Column<NumColumns;Column++)
			{
				var TheCoordinate=Coordinates[0][Column];
				
				if (TheCoordinate!=undefined)
				{
					var Text=CMUtilities.GetDMSFromDD(TheCoordinate.Longitude,true,true);
					
					TheView.PaintRefText(Text,TheCoordinate.Easting,TheCoordinate.Northing-RefFontHeight/4,FontSize,"center",0)
				}
			}
		}
	}
}
/**
* Paint the border around the outside of the map for the coordinates when the graticules intersect with
* the edge of the map
*/
CMLayerGraticule.prototype.PaintBorder=function(TheView) 
{
	//*********************************************************
	// draw the border width coordinates in it.
	
	var BorderWidth=this.GetSetting("Border","BorderWidth",10);
	
	if (BorderWidth>0)
	{
		var TheCanvasElement=TheView.GetCanvasElement();
		
		var CanvasWidth=TheCanvasElement.width;
		var CanvasHeight=TheCanvasElement.height;
		
		// the border is filled only
		var FillStyle=this.GetSetting("Border","fillStyle","rgba(255,255,255)");
		var NewStyle={fillStyle:FillStyle,strokeStyle:"rgba(0,0,0,0)"};
		TheView.SetStyle(NewStyle,false);
		
		TheView.PaintRect(0,BorderWidth,0,CanvasHeight); // left
		TheView.PaintRect(0,CanvasWidth,0,BorderWidth); // top
		TheView.PaintRect(CanvasWidth-BorderWidth,CanvasWidth,0,CanvasHeight); // right
		TheView.PaintRect(0,CanvasWidth,CanvasHeight-BorderWidth,CanvasHeight); // bottom
		
		// the neat line is just stroked
		var NeatLineStyle=this.GetSetting("Border","strokeStyle","rgba(0,0,0)");
		var NewStyle={fillStyle:"rgba(0,0,0,0)",strokeStyle:NeatLineStyle};
		TheView.SetStyle(NewStyle,false);
		
		TheView.PaintRect(BorderWidth,CanvasWidth-BorderWidth,BorderWidth,CanvasHeight-BorderWidth); // bottom
	}
}
/**
* Paint the coordinate labels around the edge of the map
*/
CMLayerGraticule.prototype.PaintBorderLabels=function(TheView) 
{
	var BorderWidth=this.GetSetting("Border","BorderWidth",10);
	
	if (BorderWidth>0)
	{
		// draw the coordinates in the border
		
		var BorderTextStyle=this.GetSettingGroup("BorderText");
		var FontSize=this.SetupLabelFont(TheView,-1); // setup the font but not from a feature
		
		TheView.SetStyle(BorderTextStyle,false);
		
		// draw coordinates in the top border
		
		TheView.ResetCollisions();
				
		var RefFontSize=TheView.GetRefHeightFromPixelHeight(FontSize);
		
		var RefOffsetToBorder=RefFontSize*0.2; // find the offset to the bottom of the border from be text baseline
		
		if (this.TopIntersectionCoordinates!=null)
		{
			for (var i=0;i<this.TopIntersectionCoordinates.length;i++)
			{
				var TheCoordinate=this.TopIntersectionCoordinates[i];
				
				var Text=CMUtilities.GetDMSFromDD(TheCoordinate.Longitude,true,true);
				
				TheView.PaintRefText(Text,TheCoordinate.Easting,TheCoordinate.Northing-RefOffsetToBorder,FontSize,"center",0)
			}
		}
		// draw coordinates in the bottom border
		
		var RefOffset2=TheView.GetRefHeightFromPixelHeight(FontSize);
				
		if (this.BottomIntersectionCoordinates!=null)
		{
			for (var i=0;i<this.BottomIntersectionCoordinates.length;i++)
			{
				var TheCoordinate=this.BottomIntersectionCoordinates[i];
				
				var Text=CMUtilities.GetDMSFromDD(TheCoordinate.Longitude,true,true);
				
				TheView.PaintRefText(Text,TheCoordinate.Easting,TheCoordinate.Northing+(RefOffset2*0.9),FontSize,"center",0)
			}
		}
		// draw coordinates in the left border
				
		if (this.LeftIntersectionCoordinates!=null)
		{
			for (var i=0;i<this.LeftIntersectionCoordinates.length;i++)
			{
				var TheCoordinate=this.LeftIntersectionCoordinates[i];
				
				var Text=CMUtilities.GetDMSFromDD(TheCoordinate.Latitude,false,true);
				
				TheView.PaintRefText(Text,TheCoordinate.Easting+RefOffsetToBorder,TheCoordinate.Northing,FontSize,"center",-Math.PI/2)
			}
		}
		
		// draw coordinates in the right border
		
		if (this.RightIntersectionCoordinates!=null)
		{
			for (var i=0;i<this.RightIntersectionCoordinates.length;i++)
			{
				var TheCoordinate=this.RightIntersectionCoordinates[i];
				
				var Text=CMUtilities.GetDMSFromDD(TheCoordinate.Latitude,false,true);
				
				TheView.PaintRefText(Text,TheCoordinate.Easting-RefOffsetToBorder,TheCoordinate.Northing,FontSize,"center",Math.PI/2)
			}
		}
	}
}
//******************************************************************
// CMLayer Functions
//******************************************************************
/**
* Paint one set of graticules based on a clipping bounds in geographic
*
* Approach:
* - GetLatLonRangeFromClippingBounds
* - GetProjectedCoordinateGrid
*
* @protected
*/
CMLayerGraticule.prototype.Paint=function(TheView) 
{
	// 2D painting
	if ((this.IsVisible(TheView))&&(TheView instanceof CMView2D))
	{
		// make sure we have a projector
		var TheGeo=this.GetParent(CMGeo);
		
		var TheProjector=TheGeo.GetProjector();
		if (TheProjector==null) TheProjector=new CMProjector(); // geographic
		
		TheProjector.Projection=null;
		TheProjector.Initialize();
		
		//*******************************************************
		// Find the lat/long min/maxes that for each clipping bounds that appears in the view
		var ClippingPolys=TheProjector.GetClippingPolys();
		
		if (ClippingPolys!=undefined)
		{
			var LatLonRanges=[];
			for (var i=0;i<ClippingPolys.length;i++)
			{
				var ClippingPoly=ClippingPolys[i]; // Xmin,Ymin,XMax,YMax
				
				var LatLonRange=this.GetLatLonRangeFromClippingBounds(TheView,ClippingPoly,TheProjector);
				
				LatLonRanges.push(LatLonRange);
			}
			//**********************************************************************************************
			// Find the center of the largest clipping bounds in the view and the total range of lat/lon values
			//var TotalLatitudeRange=-1;
			var TotalLonRange;
			var MaxLonRange=0; // maximum width of the map in the current projection
			var LonAtMaxLonRange=0; // lat and lon at the maximum width
			var LatAtMaxLonRange=0;
			
			for (var i=0;i<ClippingPolys.length;i++)
			{
				var LatLonRange=LatLonRanges[i];
				
				if (LatLonRange!=null)
				{
					var NewLatRange=LatLonRange.MaxLat-LatLonRange.MinLat;
					var NewLonRange=LatLonRange.MaxLon-LatLonRange.MinLon;
					
					if (NewLonRange>MaxLonRange) // works for -1 to initialize or when smaller
					{
						MaxLonRange=NewLonRange;
						LonAtMaxLonRange=(LatLonRange.MaxLon+LatLonRange.MinLon)/2;
						LatAtMaxLonRange=(LatLonRange.MaxLat+LatLonRange.MinLat)/2;
					}
					//TotalLatitudeRange+=NewLatRange;
					if (TotalLonRange==undefined) TotalLonRange=NewLonRange;
					else TotalLonRange+=NewLonRange;
				}
			}
			
			//**********************************************************************
			// Find the spacing between the lines based on the resolution of the map (returns a value from the DegreeQuantized[] array
			//if (Spacing==-1)
	
			if (TotalLonRange<0) 
			{
				throw("Sorry, the TotalLonRange cannot be less than zero");
			}
			else
			{
				Spacing=this.GetSpacing(TheView,TheProjector,LonAtMaxLonRange,LatAtMaxLonRange,TotalLonRange);
			
				console.log("Spacing="+Spacing);
				
				var TheStyle=this.GetStyle();
			
				for (var i=0;i<ClippingPolys.length;i++)
				{
					var LatLonRange=LatLonRanges[i];
					
					if (LatLonRange!=null)
					{
						var ClippingPoly=ClippingPolys[i];
						
						var XMinMax=CMUtilities.GetMinMax(ClippingPoly.Xs);
						var YMinMax=CMUtilities.GetMinMax(ClippingPoly.Ys);
						
						LatLonRange.Spacing=Spacing;
						
						//**********************************************************************
						// find the lat/lons that are an even mulitple of the spacing
						// In other words, move the grid out until we hit meridians and parallels that fit the spacing.
						
						LatLonRange.MinLon=Math.floor(LatLonRange.MinLon/Spacing-1)*Spacing;
						LatLonRange.MaxLon=Math.ceil(LatLonRange.MaxLon/Spacing+1)*Spacing;
						LatLonRange.MinLat=Math.floor(LatLonRange.MinLat/Spacing-1)*Spacing;
						LatLonRange.MaxLat=Math.ceil(LatLonRange.MaxLat/Spacing+1)*Spacing;
						
						if (LatLonRange.MinLon<XMinMax.Min) LatLonRange.MinLon=XMinMax.Min;
						if (LatLonRange.MaxLon>XMinMax.Max) LatLonRange.MaxLon=XMinMax.Max;
						if (LatLonRange.MinLat<YMinMax.Min) LatLonRange.MinLat=YMinMax.Min;
						if (LatLonRange.MaxLat>YMinMax.Max) LatLonRange.MaxLat=YMinMax.Max;
					}
				}
				//*********************************************************************
				// Paint the graticules which includes where they intersect with the view bounds
				
				this.BottomIntersectionCoordinates=[];
				this.TopIntersectionCoordinates=[];
				this.LeftIntersectionCoordinates=[];
				this.RightIntersectionCoordinates=[];
				
				var TextVisible=this.GetSetting("Text","Visible",true);
	
				for (var i=0;i<ClippingPolys.length;i++)
				{
					var LatLonRange=LatLonRanges[i];
					
					if (LatLonRange!=null)
					{
						var ClippingPoly=ClippingPolys[i];
						
						if (LatLonRange!=null) // returns null if the grid is not in the view
						{
							var ProjectedCoordinateGrid=this.GetProjectedCoordinateGrid(TheView,LatLonRange,TheProjector);
							
							// setup an array to catch any points of intersection that cross the border
							this.PaintCoordinateGrid(TheView,LatLonRange,ProjectedCoordinateGrid,TheProjector);
							
							if (TextVisible)
							{
								TheView.ResetCollisions();
					
								// paint the labels within the grid
								
								this.PaintGridLabels(TheView,LatLonRange,ProjectedCoordinateGrid,ClippingPoly);
							}
						}
					}
				}
				//***********************************************************************************
				// After the graticules have been painted, we can paint the border and border labels if desired
				
				// paint the boreder with it's labels
				this.PaintBorder(TheView);
		
				// paint the boreder with it's labels
				this.PaintBorderLabels(TheView);
			}
		}
		// restore the original style
		TheStyle=TheView.SetStyle(TheStyle);
	}
}

