/******************************************************************************************************************
* CMProj4WebService
*
* Projector class that uses a web service that communicates with the Proj4 engine to create projection grids.
* The class uses the ProjInfo.js file for information on the available projection methods and parameters.
*
* @extends CMProj4Grid
* @module CMProj4WebService
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/
CMProj4WebService.MESSAGE_GRID_RECIEVED=CMBase.GetUniqueNumber();

/**
* Below are the settings definitions.
* @public, @settings
*/
CMProj4WebService.SettingDefintions=
{
	Projection:
	{
		LonMin: { Name:"LonMin",Type:CMBase.DATA_TYPE_FLOAT, Default:0, Min:-180,Max:180 },
		LonMax: { Name:"LonMax",Type:CMBase.DATA_TYPE_FLOAT, Default:0, Min:-180,Max:180 },
		LatMin: { Name:"LatMin",Type:CMBase.DATA_TYPE_FLOAT, Default:0, Min:-90,Max:90 },
		LatMax: { Name:"LatMax",Type:CMBase.DATA_TYPE_FLOAT, Default:0, Min:-90,Max:90 },
		
		// the following are all proj4 parameters
		lon_0: { Name:"Longitude Of Origin",Type:CMBase.DATA_TYPE_FLOAT, Default:0, Min:-180,Max:180 }, // typically, longitude of origin
		lon_1: { Name:"First Meridian",Type:CMBase.DATA_TYPE_FLOAT, Default:-40, Min:-180,Max:180 }, // first longitude
		lon_2: { Name:"Second Merdian",Type:CMBase.DATA_TYPE_FLOAT, Default:40, Min:-180,Max:180 },// second longitude
		lonc: { Name:"Lonigidude of central point",Type:CMBase.DATA_TYPE_FLOAT, Default:0, Min:-180,Max:180 },// 
		
		lat_ts: { Name:"Latitude Of True Scale",Type:CMBase.DATA_TYPE_FLOAT, Default:0, Min:-90,Max:90 }, // latitude of true scale
		lat_0: { Name:"Latitude Of Origin",Type:CMBase.DATA_TYPE_FLOAT, Default:0, Min:-90,Max:90 }, // typically, latitude of origin
		lat_1: { Name:"First Parallel",Type:CMBase.DATA_TYPE_FLOAT, Default:20, Min:-90,Max:90 }, // first latitude for conics
		lat_2: { Name:"Second Parallel",Type:CMBase.DATA_TYPE_FLOAT, Default:40, Min:-90,Max:90 }, // second latitude for conics
		
		h: { Name:"Height of Satellite above the earth",Type:CMBase.DATA_TYPE_FLOAT, Default:35785831.0 , Min:1,Max:1000000000 },
		n: { Name:"n",Type:CMBase.DATA_TYPE_FLOAT, Default:2 , Min:1,Max:5 }, // undocumented
		m: { Name:"m",Type:CMBase.DATA_TYPE_FLOAT, Default:3 , Min:1,Max:5 },// undocumented
		M: { Name:"M",Type:CMBase.DATA_TYPE_FLOAT, Default:0.5 , Min:0.001,Max:5 },// greater than 0
		W: { Name:"W",Type:CMBase.DATA_TYPE_FLOAT, Default:1 , Min:0.001,Max:5 },// greater than 0
		south: { Name:"South",Type:CMBase.DATA_TYPE_BOOLEAN, Default:false },
		lsat: { Name:"Satellite",Type:CMBase.DATA_TYPE_ENUMERATED, Options:[1,2,3,4,5], Default:1 },
		path: { Name:"Path",Type:CMBase.DATA_TYPE_INTEGER, Default:1, Min:1,Max:253},
		alpha: { Name:"Alpha",Type:CMBase.DATA_TYPE_FLOAT, Default:1, Min:0,Max:360}, // angle of rotational pole
		tilt: { Name:"Tilt",Type:CMBase.DATA_TYPE_FLOAT, Default:1, Min:0,Max:360}, // 
		azi: { Name:"Azimuth",Type:CMBase.DATA_TYPE_FLOAT, Default:1, Min:0,Max:360}, // 
		
		o_lat_p: { Name:"Latitude of New Pole",Type:CMBase.DATA_TYPE_FLOAT, Default:90, Min:-90,Max:90}, // moves latitude for oblique projections
	
		//
		Type: { Name:"Type",Type:CMBase.DATA_TYPE_STRING, Default:'Cylindrical' }, // 
		Method: { Name:"Method",Type:CMBase.DATA_TYPE_STRING, Default:'Robinson' }, // only those projection methods that work well are exposed at this time.

		LongitudeStep: { Name:"Longitude Step",Type:CMBase.DATA_TYPE_INTEGER, Default:5 }, // 
		LatitudeStep: { Name:"Latitude Step",Type:CMBase.DATA_TYPE_INTEGER, Default:5 }, // 
	},
};

//***************************************************************************************
// Constructor
//***************************************************************************************
/*
* Constructor
* @public, @constructs
*/
function CMProj4WebService() 
{
	CMProjector.call(this);
	
	// jjg - move to settings
	this.SemiMajor=6371000;

	// Properties
	this.Error=null;
}
CMProj4WebService.prototype=Object.create(CMProj4Grid.prototype); // inherit prototype functions from CMProjector()

CMProj4WebService.prototype.contructor=CMProj4WebService; // override the constructor to go to ours

//***************************************************************************************
//	CMBase functions
//***************************************************************************************

CMProj4WebService.prototype.GetSettingsDefinitions=function() 
{
	var Result={};
	
	for (Key in CMProj4WebService.SettingDefintions)
	{
		Result[Key]=CMProj4WebService.SettingDefintions[Key];
	}

	return(Result); 
}
CMProj4WebService.prototype.CMProj4Grid_SetSettings=CMProj4Grid.prototype.SetSettings;

CMProj4WebService.prototype.SetSettings=function(NewSettings)
{
/*	for (var GroupKey in NewSettings)
	{
		if (GroupKey=="Projection")
		{
			for (var SettingKey in NewSettings[GroupKey])
			{
				if (SettingKey!="lon_0")
				{
					// dump the current grid so a new one is loaded
					this.SetProjectionGrid(undefined);
				}
			}
		}
	}
*/	this.CMProj4Grid_SetSettings(NewSettings);
}
CMProj4WebService.prototype.CMProj4Grid_SettingsChanged=CMProj4Grid.prototype.SettingsChanged;

CMProj4WebService.prototype.SettingsChanged=function() 
{
	this.CMProj4Grid_SettingsChanged();
	
	this.SendMessageToListeners(CMBase.MESSAGE_SETTINGS_CHANGED);
}

//***************************************************************************************
//	CMProjector functions
//***************************************************************************************
//CMProj4WebService.prototype.CMProj4Grid_Initialize=CMProj4Grid.prototype.Initialize;

CMProj4WebService.prototype.Initialize=function()
{
	//this.CMProj4Grid_Initialize();
	
//	if (this.Proj4WebService_Initialized==false)
	{
		if (this.GetProjectionGrid()==undefined)
		{
			if (this.PendingRequest==undefined) // do not have a pending request
			{
				var ProjectionName=this.GetSetting("Projection","Method");
				var LongitudeStep=this.GetSetting("Projection","LongitudeStep",5);
				var LatitudeStep=this.GetSetting("Projection","LatitudeStep",5);
				var LonMax=this.GetSetting("Projection","LonMax",180);
				var LonMin=this.GetSetting("Projection","LonMin",-180);
				var LatMax=this.GetSetting("Projection","LatMax",90);
				var LatMin=this.GetSetting("Projection","LatMin",-90);
				
				var LatNewPole=this.GetSetting("Projection","o_lat_p",90);
				
				var Index=CMProj4WebService.GetProjectionIndexFromSettingsName(ProjectionName);
				
				var ProjectionCode=ProjInfo.ProjectionArray[Index].Code;
				var LongName=ProjInfo.ProjectionArray[Index].LongName;
				
				var Parameters=ProjInfo.ProjectionArray[Index].Parameters;
				
				var URL="/Archive/WebServices/GetProjGridFile.py"+
					"?ProjectionCode="+ProjectionCode+
					"&LongName="+LongName+
					"&LonMax="+LonMax+
					"&LonMin="+LonMin+
					"&LatMax="+LatMax+
					"&LatMin="+LatMin+
					"&LatNewPole="+LatNewPole+
					"&LonStep="+LongitudeStep+
					"&LatStep="+LatitudeStep;
				
				for (var  i=0;i<Parameters.length;i++)
				{
					var Name=Parameters[i].Name;
					
					var Value=this.GetSetting("Projection",Name,0);
					
					if ((Name=="south"))  // ignore for now
					{
						//URL+="&south"; // jjg not connected yet
					}
/*					else if (Name=="lon_0") // handled inside the CMProj4Grid class
					{
					}
*/					else
					{
						URL+="&"+Name+"="+Value;
					}
				}
				
				// Create the request
				var TheRequest=new XMLHttpRequest(); // wait staff at the resturant
				TheRequest.open("GET",URL,true); // the URL is what we ordered
				TheRequest.TheURL=URL;
				TheRequest.TheProjector=this;
					
				TheRequest.onreadystatechange=function() 
				{
					if (this.readyState == 4)  // done
					{
						if (this.status == 200) // OK
						{
							var TheText=TheRequest.responseText;
			
							try
							{
								// convert the data to JavaScript object
								var TheGeoJSONObject=JSON.parse(TheText);
							
								// make sure all invalid values are NaNs.
								
								for (var Row=0;Row<TheGeoJSONObject.NumRows;Row++)
								{
									for (var Column=0;Column<TheGeoJSONObject.NumColumns;Column++)
									{
										if (TheGeoJSONObject.Eastings[Row][Column]=="NaN") TheGeoJSONObject.Eastings[Row][Column]=NaN;
										if (TheGeoJSONObject.Northings[Row][Column]=="NaN") TheGeoJSONObject.Northings[Row][Column]=NaN;
										if (TheGeoJSONObject.Longitudes[Row][Column]=="NaN") TheGeoJSONObject.Longitudes[Row][Column]=NaN;
										if (TheGeoJSONObject.Latitudes[Row][Column]=="NaN") TheGeoJSONObject.Latitudes[Row][Column]=NaN;
									}
								}
								// save the grid
								this.TheProjector.SetProjectionGrid(TheGeoJSONObject);
							}
							catch(err) 
							{
								alert(err.message+" "+TheText);
							}
								
							this.TheProjector.SendMessageToListeners(CMProj4WebService.MESSAGE_GRID_RECIEVED);
						}
						else alert("HTTP error "+this.status+" "+this.statusText+" ("+this.TheURL+")");
						
						this.TheProjector.PendingRequest=undefined;
					}
				}
				
				this.PendingRequest=TheRequest;
				TheRequest.send();		
			}
		}
	}
}

//***************************************************************************************
// CMProjector Functions
//***************************************************************************************
CMProj4WebService.prototype.GetMethods=function()
{
	var ProjectionMethods=[];
	
	var Type=this.GetSetting("Projection","Type","Cylindrical");
	if (Type=="Other") Type="";
	
	for (var i=0;i<ProjInfo.ProjectionArray.length;i++)
	{
		var Flag=false;
		
		if (ProjInfo.ProjectionArray[i].Type==Type)
		{
			ProjectionMethods.push(ProjInfo.ProjectionArray[i].LongName);
		}
	}
	return(ProjectionMethods);
}
//***************************************************************************************
// CMProjector Functions
//***************************************************************************************

/**
* Returns clipping bounds that can be used to clip geographic data for the current
* projection settings
* @public
* @returns - 
*//*
CMProj4WebService.prototype.CMProj4Grid_GetClippingPolys=CMProj4Grid.prototype.GetClippingPolys;

CMProj4WebService.prototype.GetClippingPolys=function()
{
	this.Initialize();

	var Result=this.CMProj4Grid_GetClippingPolys();
	
	var TheProjectionGrid=this.GetProjectionGrid();
		
	if (TheProjectionGrid!=undefined) // have a grid
	{
		// offset the longitude which can be less than -180 or greater than 180 for animation
		var LongitudeOfOrigin=this.GetSetting("Projection","lon_0",0);
		
		var XMin=TheProjectionGrid.LongitudeMin+LongitudeOfOrigin;
		var XMax=TheProjectionGrid.LongitudeMax+LongitudeOfOrigin;
			
		var Pad=0.00001; // this is required to keep drawn polygons from breaking across the clipping areas
		
		Result=CMProjector.GetGeographicClippingPolys(XMin,XMax,TheProjectionGrid.LatitudeMin,TheProjectionGrid.LatitudeMax,Pad);
	}
	return(Result);
}*/
//***************************************************************************************
// CMProjector functions that had to be overriden to create a circle for clipping orthographic projections
//***************************************************************************************
/**
* Return a region that can be used to draw the outline of the Geo based on the current projection
* @public
* @returns A Region that can be drawn with TheView.PaintRefRegion(RegionToDraw,false,false,true)
*/
CMProj4WebService.prototype.CMProjector_GetRegionToDraw=CMProj4Grid.prototype.GetRegionToDraw;
CMProj4WebService.prototype.GetRegionToDraw=function()
{
	var RegionToDraw=[];
	
	var ProjectionName=this.GetSetting("Projection","Method");
	
	if (ProjectionName=="Orthographic")
	{
		var PolyToDraw=CMUtilities.GetRegularPolygon(100,this.SemiMajor*2,0,0,0);
		var AreaToDraw=[PolyToDraw];
		RegionToDraw.push(AreaToDraw);
	}
	else if (this.GetProjectionGrid()!=undefined)
	{
		RegionToDraw=this.CMProjector_GetRegionToDraw();
	}
	
	return(RegionToDraw);
}

/**
* Return a region that can be used to fill the area of the Geo based on the current projection
* @public
* @returns A Region that can be filled with TheView.PaintRefRegion(RegionToFill,false,true,false);
*/
CMProj4WebService.prototype.CMProjector_GetRegionToFill=CMProjector.prototype.GetRegionToFill;

CMProj4WebService.prototype.GetRegionToFill=function()
{
	var RegionToFill=[];
	
	var ProjectionName=this.GetSetting("Projection","Method");
	
	if (ProjectionName=="Orthographic")
	{
		var PolyToDraw=CMUtilities.GetRegularPolygon(100,this.SemiMajor*2,0,0,0);
		var AreaToDraw=[PolyToDraw];
		RegionToFill.push(AreaToDraw);
	}
	else if (this.GetProjectionGrid()!=undefined)
	{
		RegionToFill=this.CMProjector_GetRegionToFill();
	}
	return(RegionToFill);
}

/**
* Gets the maximum rectangular box for the current projection
* Used to zoom to the entire bounds of a projection
*/
CMProj4WebService.prototype.CMProjector_GetProjectedBounds=CMProjector.prototype.GetProjectedBounds;

CMProj4WebService.prototype.GetProjectedBounds=function()
{
	var Result=null;
	
	var ProjectionName=this.GetSetting("Projection","Method");
	
	if (ProjectionName=="Orthographic")
	{
		Result={
			XMin:-6371000,
			XMax:6371000,
			YMin:-6371000,
			YMax:6371000
		};
	}
	else if (this.GetProjectionGrid()!=undefined)
	{
		Result=this.CMProjector_GetProjectedBounds();
	}
	return(Result);
}
CMProj4WebService.prototype.GetClass=function() { return(CMProj4WebService); }

//***************************************************************************************
// static public CMProj4WebService Functions to return information from the ProjInfo dictionary
//***************************************************************************************
/**
* @static
* @public
*/
CMProj4WebService.GetProjectionCodeFromProjectionName=function(LongName)
{
	var Index=CMProj4WebService.GetProjectionIndexFromSettingsName(LongName)
	
	var Code=ProjInfo.ProjectionArray[Index]["Code"];
	
	return(Code);
}
/**
* @static
* @public
*/
CMProj4WebService.GetProjectionIndexFromProjectionCode=function(ProjectionCode)
{
	var Index=-1;
	
	for (var i=0;(i<ProjInfo.ProjectionArray.length)&&(Index==-1);i++)
	{
		if (ProjInfo.ProjectionArray[i]["Code"]==ProjectionCode) Index=i;
	}
	
	return(Index);
}
/**
* @static
* @public
*/
CMProj4WebService.GetProjectionIndexFromSettingsName=function(SettingsName)
{
	var Index=-1;
	
	for (var i=0;(i<ProjInfo.ProjectionArray.length)&&(Index==-1);i++)
	{
		if (ProjInfo.ProjectionArray[i]["LongName"]==SettingsName) Index=i;
	}
	
	return(Index);
}

/**
* Returns projection information specified in the ProjInfo object (typically from including ProjInfo.js)
* @public
* @static
*/
CMProj4WebService.GetControlInformation=function(ProjectionCode)
{
	var Controls=null;
	
	var ProjectionIndex=CMProj4WebService.GetProjectionIndexFromProjectionCode(ProjectionCode);
	
	if (ProjInfo.ProjectionArray[ProjectionIndex].Parameters!=undefined)
	{
		Controls=ProjInfo.ProjectionArray[ProjectionIndex].Parameters;
	}
	return(Controls);
}
/**
* Returns a geographic bounds that will make the specified projection look too
* @public
* @static
*/
CMProj4WebService.GetPreferedGeographicBounds=function(ProjectionCode)
{
	var Result=null;
	
	var ProjectionIndex=CMProj4WebService.GetProjectionIndexFromProjectionCode(ProjectionCode);
	
	Result={
		LonMax:ProjInfo.ProjectionArray[ProjectionIndex].LonMax,
		LonMin:ProjInfo.ProjectionArray[ProjectionIndex].LonMin,
		LatMax:ProjInfo.ProjectionArray[ProjectionIndex].LatMax,
		LatMin:ProjInfo.ProjectionArray[ProjectionIndex].LatMin,
	};
	return(Result);
}
/**
* Return one of the elements in the ProjInfo projection array.
* @public
* @static
*/
CMProj4WebService.GetProjectionInfoFromProjectionCode=function(ProjectionCode,Key)
{
	var Result=null;
	
	var ProjectionIndex=CMProj4WebService.GetProjectionIndexFromProjectionCode(ProjectionCode);
	
	Result=ProjInfo.ProjectionArray[ProjectionIndex][Key];
	
	return(Result);
}

