/******************************************************************************************************************
* CMProjector Base class.
* This class defines the main functions for projectors to convert between projected and
* geographic coordinates.
*
* This class also provides a projector that can be used to keep geographic data as geographic
* data.  In otherwords, it does not touch the data and assumes the data is in geographic.  This
* allows it to work with layers like the graticule layer.
*
* @module CMProjector
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/
//******************************************************************
// Constructors
//******************************************************************
/*
* Constructor for a projector.  This is a null class that just passes
* the coordinate values back without changing them.  It effectively
* implements a flattened "geographic" projection.
* @protected, @constructs
*/
function CMProjector() 
{
	CMBase.call(this);	
}

CMProjector.prototype=Object.create(CMBase.prototype); // inherit prototype functions from CMBase()

CMProjector.prototype.contructor=CMProjector; // override the constructor to go to ours

//***************************************************************************************
// Private static functions
//***************************************************************************************
/**
* Return the range of x coordinate values if it is less then the range for the y coordinate values,
* otherwise, return the y range
* @private
*/
CMProjector.GetMinPointDistance=function(Xs,Ys)
{
	var XMinMax=CMUtilities.GetMinMax(Xs);
	var YMinMax=CMUtilities.GetMinMax(Ys);

	var XFactor=(XMinMax.Max-XMinMax.Min)/200;
	var YFactor=(YMinMax.Max-YMinMax.Min)/200;

	var MinDistance=XFactor;
	if (XFactor>YFactor) MinDistance=YFactor;
	
	return(MinDistance);
}
//***************************************************************************************
// Protected Functions to be used by subclasses to project data after clipping to bounds
//***************************************************************************************

/**
* Projects a single poly (i.e. an array of arrays of coordinates - 2 deep) (lineString, linearRing)
* but without clipping.  
* @proected
*/
CMProjector.prototype.ProjectPolyFromGeographicWithoutClip=function(ThePoly)
{
	var NewPoly=[];
	
	var Xs=ThePoly[0];
	var Ys=ThePoly[1];
	
	var NewXs=[];
	var NewYs=[];
	
	for (var j=0; j < Xs.length; j++) 
	{
		try
		{
			var Coordinate=this.ProjectFromGeographic(Xs[j],Ys[j]);
			
			if (CMUtilities.IsDefined(Coordinate))
			{
				NewXs.push(Coordinate[0]);
				NewYs.push(Coordinate[1]);
			}
		}
		catch (Error)
		{
			throw(Error);
		}
	}
	return([NewXs,NewYs]);
}
/**
* Projects an array of related polys (i.e. an array of arrays of arrays - 3 deep) (multiLineString, polygon)
* @proected
*/
CMProjector.prototype.ProjectAreaFromGeographicWithoutClip=function(TheArea)
{
	var NewArea=[];
	
	for (var j=0; j < TheArea.length; j++) 
	{
		NewArea[j]=this.ProjectPolyFromGeographicWithoutClip(TheArea[j]);
	}
	return(NewArea);
}
/**
* Projects an array of related polys (i.e. an array of arrays of arrays of arrays - 4 deep) (multiPolygon)
* @proected
*/
CMProjector.prototype.ProjectRegionFromGeographicWithoutClip=function(TheRegion)
{
	var NewRegion=[];
	
	for (var j=0; j < TheRegion.length; j++) 
	{
		NewRegion[j]=this.ProjectAreaFromGeographicWithoutClip(TheRegion[j]);
	}
	return(NewRegion);
}

//******************************************************************
// CMProjector functions for subclasses to override
//******************************************************************
/**
* Subclasses should override this to setup for projecting.
* @override
* @protected
*/
CMProjector.prototype.Initialize=function()
{
}

/**
* Primary function to convert from a projected coordinate to geographic
* @override
* @public
* @param Easting - east-west projected coordinate value
* @param Northing - north-south projected coordinate value
* @param Elevation - vertical coordinate value in meters
*/
CMProjector.prototype.ProjectToGeographic=function(Easting,Northing,Elevation)
{
	var Result=[Easting,Northing,Elevation];
	
	return(Result);
}
/**
* Primary function to convert from a projected system to geographic
* @override
* @public
* @param Longitude - east-west geographic coordinate value
* @param Latitude - north-south geographic coordinate value
* @param Elevation - vertical coordinate value in meters
*/
CMProjector.prototype.ProjectFromGeographic=function(Longitude,Latitude,Elevation)
{
	var Result=[Longitude,Latitude,Elevation];
	return(Result);
}
//******************************************************************
// Protected functions for subclasses to override as needed
//******************************************************************
CMProjector.prototype.GetMethods=function()
{
	var SettingsDefinitions=this.GetSettingsDefinitions();
	var ProjectionMethods=SettingsDefinitions.Projection.Method.Options;
	return(ProjectionMethods);
}

/**
* Return the clipping polygons in geographic coordinates for the projection.
* Override to provide clipping to keep geometries correct for projections.
* No longer gets called
* @override
* @public
*/
CMProjector.prototype.GetClippingPolys=function()
{
	//alert("CMProjector.prototype.GetClippingPolys() needs to be overriden");
	
	var Left=-180;
	var Right=180;
	var Top=90;
	var Bottom=-90
	
	var ClippingPoly={
		Xs:[Left,Right,Right,Left],
		Ys:[Top,Top,Bottom,Bottom],
		Outside:[true,true,true,true]
	};
	Result=[ClippingPoly];
	
	return(Result);
}
/**
* Return a region that can be used to draw the outline of the Geo based on the current projection.
* This is based only on the "Outside" bounding areas.
* @protected
* @returns A Region that can be drawn with TheView.PaintRefRegion(RegionToDraw,false,false,true)
*/
CMProjector.prototype.GetRegionToDraw=function()
{
	var RegionToDraw=[];
	
	// fill the area
	var ClippingPolys=this.GetClippingPolys();
	
	if (ClippingPolys==null) // no clipping
	{
		var X2s=[-180,180,180,-180];
		var Y2s=[90,90,-90,-90];
		
		var NewArea=this.ProjectAreaFromGeographicWithoutClip([[X2s,Y2s]]);
		
		RegionToDraw.push(NewArea);
	}
	else
	{
		// stroke the outside areas
		for (var i=0;i<ClippingPolys.length;i++)
		{
			var ClippingPoly=ClippingPolys[i];
			
			var Xs=ClippingPoly.Xs;
			var Ys=ClippingPoly.Ys;
			var Outside=ClippingPoly.Outside;
	
			// create a series of polylines for each edge of the bounds when the Outside flag is true
			
			for (var j=0;j<Outside.length;j++)
			{
				if (Outside[j]) // line is on the outside of the bounds (top)
				{
					var NextIndex=j+1;
					if (NextIndex==Outside.length) NextIndex=0;
					
					var X1s=[Xs[j],Xs[NextIndex]];
					var Y1s=[Ys[j],Ys[NextIndex]];
					
					var NewArea=[[X1s,Y1s]];
					
					var MinDistance=CMProjector.GetMinPointDistance(Xs,Ys);
			
					NewArea=CMUtilities.AddPointsToArea(NewArea,CMDatasetVector.TYPE_POLYGONS,MinDistance); // TYPE_POLYLINES
					
					NewArea=this.ProjectAreaFromGeographicWithoutClip(NewArea);
					
					RegionToDraw.push(NewArea);
				}
			}
		}
	}
	return(RegionToDraw);
}
/**
* Return a region that can be used to fill the valid areas of a projection.  This can be different from
* the drawing bounds because it includes polygons inside the bounds.
* @protected
* @returns A Region that can be drawn with TheView.PaintRefRegion(...)
*/
CMProjector.prototype.GetRegionToFill=function()
{
	var RegionToFill=[];
	
	// fill the area
	var BBoxes=this.GetClippingPolys();
	
	if (BBoxes==null)
	{
		var X2s=[-180,180,180,-180];
		var Y2s=[90,90,-90,-90];
		
		var NewArea=this.ProjectAreaFromGeographicWithoutClip([[X2s,Y2s]]);
		
		RegionToFill.push(NewArea);
	}
	else
	{
		// fill the areas
		for (var i=0;i<BBoxes.length;i++)
		{
			var BBox=BBoxes[i];
			
			var Xs=BBox.Xs;
			var Ys=BBox.Ys;
			var Outside=BBox.Outside;
	
			// clip to each of the bounding boxes and project the resulting areas
			//var ClippedArea=CMUtilities.ClipArea(TheArea,BBoxes[i],Closed);
			
			var NewArea=[[Xs,Ys]];
			
			var MinDistance=CMProjector.GetMinPointDistance(Xs,Ys);
			
			NewArea=CMUtilities.AddPointsToArea(NewArea,CMDatasetVector.TYPE_POLYGONS,MinDistance);
			
			NewArea=this.ProjectAreaFromGeographicWithoutClip(NewArea);
			
			RegionToFill.push(NewArea);
			//TheView.PaintRefArea(NewArea,true,true,false); // area, closed, fill, no stroke
		}
	}
	return(RegionToFill);
}
/**
* Return the X/Y Min/Max projected range of values.
* @protected
* @returns - Dictionary with XMin, XMax, YMin, YMax
*/
CMProjector.prototype.GetProjectedBounds=function()
{
	var Result=null;
	
	var ClippingBoxes=this.GetClippingPolys();
	
	if (ClippingBoxes!=null)
	{
		var Initialized=false;
		var XMin=0;
		var YMin=0;
		var XMax=0;
		var YMax=0;
		
		for (var i=0;i<ClippingBoxes.length;i++)
		{
			var ClippingBox=ClippingBoxes[i];
			
			var NewArea=[[ClippingBox.Xs,ClippingBox.Ys]];
			
			NewArea=CMUtilities.AddPointsToArea(NewArea,CMDatasetVector.TYPE_POLYGONS,5);
			
			NewArea=this.ProjectAreaFromGeographicWithoutClip(NewArea);
			
			var Xs=NewArea[0][0];
			var Ys=NewArea[0][1];
			
			for (var j=0;j<Xs.length;j++)
			{
				if (Initialized==false)
				{
					XMax=Xs[j];
					XMin=Xs[j];
					YMax=Ys[j];
					YMin=Ys[j];
					
					Initialized=true;
				}
				else
				{
					if (Xs[j]<XMin) XMin=Xs[j];
					if (Xs[j]>XMax) XMax=Xs[j];
					if (Ys[j]<YMin) YMin=Ys[j];
					if (Ys[j]>YMax) YMax=Ys[j];
				}
			}
		}
		Result={
			XMin:XMin,
			XMax:XMax,
			YMin:YMin,
			YMax:YMax
		};
	}
	return(Result);
}
//******************************************************************
// Public functions to project.  These should be overriden if
// clipping is required.
//******************************************************************

/**
* Override to provide clipping to keep geometries correct for projections.
* @override
* @public
* @param ThePoly - Array of two dimensions, ThePoly[0]=Xs[] and ThePoly[1]=Ys[], optionally, ThePoly[2]=Zs[]
* @returns - An array with the new Xs, Ys, and Zs.
*/
CMProjector.prototype.ProjectPolyFromGeographic=function(ThePoly)
{
	// get the current arrays and setup the new arrays
	var Xs=ThePoly[0];
	var Ys=ThePoly[1];
	var Zs=null;
	
	var NewXs=[];
	var NewYs=[];
	var NewZs=null;
	
	if (ThePoly.length>2) 
	{
		Zs=ThePoly[2];
		NewZs=[];
	}
	// projection the coordinates
	var Z=undefined;
	for (var i=0;i<Xs.length;i++)
	{
		if (Zs!=null) Z=Zs[i];
		
		var Result=this.ProjectFromGeographic(Xs[i],Ys[i],Z);
		
		NewXs.push(Result[0]);
		NewYs.push(Result[1]);
		if (Zs!=null) NewZs.push(Result[2]);
	}
	var NewPoly=[NewXs,NewYs];
	
	return(NewPoly);
}
/**
* Projects an Area, an array of polys, to geographic.
* @override
* @public
* @param TheArea - Array of polygons (2D or 3D)
* @param TheType - CMDatasetVector.TYPE_POINTS, CMDatasetVector.TYPE_POLYLINES, or CMDatasetVector.TYPE_POLYGONS, 
* @returns - An array of new polys
*/
CMProjector.prototype.ProjectAreaFromGeographic=function(TheArea,TheType)
{
	this.Initialize();
	
	var NewArea=[];
	 
	// find the two bounding boxes
	var BBoxes=this.GetClippingPolys();
	
	if (BBoxes==null)
	{
		NewArea=this.ProjectAreaFromGeographicWithoutClip(TheArea);
	}
	else
	{
		for (var i=0;i<BBoxes.length;i++)
		{
			var ClippingBox=BBoxes[i];
			
			// clip to each of the bounding boxes and project the resulting areas
			
			// jjg - this will need to change with more complex bounds
			var TurfBBox=[ClippingBox.Xs[0],ClippingBox.Ys[2],ClippingBox.Xs[2],ClippingBox.Ys[0]]; // minX, minY, maxX, maxY
			
			var ClippedArea=CMUtilities.ClipArea(TheArea,TurfBBox,TheType);
			
			ClippedArea=CMUtilities.AddPointsToArea(ClippedArea,TheType,5);
			
			ClippedArea=this.ProjectAreaFromGeographicWithoutClip(ClippedArea);
			
			for (var j=0;j<ClippedArea.length;j++) NewArea.push(ClippedArea[j]);
		}
	}
	return(NewArea);
}
/**
* Projects a Region, an array of areas, to geographic.
* @override
* @public
* @param TheRegion - Array of Areas (2D or 3D)
* @param TheType - CMDatasetVector.TYPE_POINTS, CMDatasetVector.TYPE_POLYLINES, or CMDatasetVector.TYPE_POLYGONS, 
* @returns - An array of new Areas
*/
CMProjector.prototype.ProjectRegionFromGeographic=function(TheRegion,TheType)
{
	this.Initialize();
	
	var NewRegion=[];
	 
	for (var i=0;i<TheRegion.length;i++)
	{
		var NewArea=this.ProjectAreaFromGeographic(TheRegion[i],TheType);
		
		NewRegion.push(NewArea);
	}

	return(NewRegion);
}
/**
* Projects an array of Regions to geographic.  This is the same as the spatial data
* for a single feature.
* @override
* @public
* @param TheRegion - Array of Regions (2D or 3D)
* @param TheType - CMDatasetVector.TYPE_POINTS, CMDatasetVector.TYPE_POLYLINES, or CMDatasetVector.TYPE_POLYGONS, 
* @returns - An array of new Regions
*/
CMProjector.prototype.ProjectRegionsFromGeographic=function(TheRegions,TheType)
{
	this.Initialize();
	
	var NewRegions=[];
	 
	for (var i=0;i<TheRegions.length;i++)
	{
		var NewRegion=this.ProjectRegionFromGeographic(TheRegions[i],TheType);
		
		NewRegions.push(NewRegion);
	}

	return(NewRegions);
}
//***************************************************************************************
// Protected static functions for help with bounds.
//***************************************************************************************
/**
* creates a set of rectangular clipping bounds for the specified geographic area.
* If the XMin<-180 or the XMax>180, mulitple boxes will be created.
*
* @protected, @static
*/
CMProjector.GetGeographicClippingPolys=function(XMin,XMax,YMin,YMax,Pad)
{
	Result=[];
	
	var Top=YMax;
	var Bottom=YMin;
	
	// find the two bounding boxes
	if (XMin<-180)
	{
		if (XMax<-180) // just have one clipping box on the left
		{
			var Left=XMin+360+Pad;
			var Right=XMax+360-Pad;
			
			var ClippingBox={
				Xs:[Left,Right,Right,Left],
				Ys:[Top,Top,Bottom,Bottom],
				Outside:[true,true,true,true]
			};
			Result.push(ClippingBox);
		}
		else // XMax>-180 and is <180, have two boxes, -180/180 is shared edge
		{
			var Left=XMin+360+Pad;
			var Right=180;
			
			var ClippingBox={
				Xs:[Left,Right,Right,Left],
				Ys:[Top,Top,Bottom,Bottom],
				Outside:[true,false,true,true]
			};
			Result.push(ClippingBox);

			var Left=-180;
			var Right=XMax-Pad;
			
			var ClippingBox={
				Xs:[Left,Right,Right,Left],
				Ys:[Top,Top,Bottom,Bottom],
				Outside:[true,true,true,false]
			};
			Result.push(ClippingBox);
		}
	}
	else if (XMax>180) // XMax must be greater than 180
	{
		if (XMin>180) // both are greater than 180, one clipping box on the right
		{
			var Left=XMin-360+Pad;
			var Right=XMax-360-Pad;
			
			var ClippingBox={
				Xs:[Left,Right,Right,Left],
				Ys:[Top,Top,Bottom,Bottom],
				Outside:[true,true,true,true]
			};
			Result.push(ClippingBox);
		}
		else // XMin<180 and XMax > 180
		{
			if (true)
			{
				var Left=XMin+Pad;
				var Right=180;
				
				var ClippingBox={
					Xs:[Left,Right,Right,Left],
					Ys:[Top,Top,Bottom,Bottom],
					Outside:[true,false,true,true]
				};
				Result.push(ClippingBox);
				
				//
				var Left=-180;
				var Right=XMax-360-Pad;
				
				var ClippingBox={
					Xs:[Left,Right,Right,Left],
					Ys:[Top,Top,Bottom,Bottom],
					Outside:[true,true,true,false]
				};
				Result.push(ClippingBox);
			}
			else
			{
				//
				var Left=XMin+Pad;
				var Right=180;
				
				var ClippingBox={
					Xs:[Left,Right,Right,Left],
					Ys:[Top,Top,Bottom,Bottom],
					Outside:[true,false,true,true]
				};
				Result.push(ClippingBox);
				
				//
				var Left=-180;
				var Right=XMax-360-Pad;
				
				var ClippingBox={
					Xs:[Left,Right,Right,Left],
					Ys:[Top,Top,Bottom,Bottom],
					Outside:[true,true,true,false]
				};
				Result.push(ClippingBox);
			}
		}
	}
	else // both are within the valid range
	{
		var Left=XMin+Pad;
		var Right=XMax-Pad;
		
		var ClippingBox={
			Xs:[Left,Right,Right,Left],
			Ys:[Top,Top,Bottom,Bottom],
			Outside:[true,true,true,true]
		};
		Result.push(ClippingBox);
	}

	return(Result);
}

//***************************************************************************************
// Public static functions for Unit conversions
//***************************************************************************************
/*
* Convert an angle in degrees to radians
* @public, @static
* @param - DegreeAngle
* @returns - angle in radians
*/
CMProjector.DegreesToRadians=function(DegreeAngle)
{
	var RadianAngle=DegreeAngle*(Math.PI/180);
	
	return(RadianAngle);
}
/*
* Convert an angle in radians to degrees
* @public, @static
* @param - RadianAngle
* @returns - angle in degrees
*/
CMProjector.RadiansToDegrees=function(RadianAngle) 
{
	DegreeAngle= RadianAngle/(Math.PI/180);
	
	return(DegreeAngle);
}
