/**
* CMScaleBar
* Class to render a scale bar into the scene.  The scale bar will be automatically 
* repainted with the scale based on its position in the scene.  
* The algorithm is to find the two points at either end of the rectangle provided
* for the scale bar and convert these points to geographic coordinates.  Then,
* a great arc calculation is performed to find the distance between the points.
* The scale displayed is then found by reducing this distance to an even multiple
* of 1, 2, or 5.  If this distance does not fit in the area provided (including  the
* the labels), the next smaller distance is computed.  This continues until a distance
* is found that fits.
* 	
* Note that the scale bar is placed into a DIV in the map container.  Thus is uses aLinkcolor
* CSS style (CM_ScaleBar) for the style of the containing box.
*
* @module CMScaleBar
* @Copyright HSU, Jim Graham, 2019
*/

//***************************************************************************************
// Constructors
//***************************************************************************************
/**
* Definitions for the units.  For ISO kilomters or meters will be displayed based on the
* size of the scale bar.  For English, miles and feet are displaed.
* @enum
*/
CMScaleBar.UNITS_ISO=0; // meters, kilometers
CMScaleBar.UNITS_ENGLISH=1; // feet, miles


/**
* Below are the settings definitions.
* @public, @settings
*/
CMScaleBar.SettingDefintions=
{
	ScaleBar:
	{
		ClassName: { Name:"Class Name",Type:CMBase.DATA_TYPE_STRING, Default:"CM_ScaleBar" },
		
		// standard HTML 5 settings except the defaults may change and sometimes the available settings will change between each settings group
		strokeStyle: { Name:"Line Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(0,0,0)" },
		lineWidth: { Name:"Line Width",Type:CMBase.DATA_TYPE_INTEGER, Default:1 },
		lineCap: { Name:"Line Cap",Type:CMBase.DATA_TYPE_ENUMERATED, Options:['butt','round','square'],Default:'round' },
		lineJoin: { Name:"Line Join",Type:CMBase.DATA_TYPE_ENUMERATED, Options:['bevel','round','miter'],Default:'round' },
		fillStyle: { Name:"Fill Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(255,255,255)" },
		shadowColor: { Name:"Shadow Color",Type:CMBase.DATA_TYPE_COLOR, Default:"rgb(0,0,0)" },
		shadowBlur: { Name:"Shadow Blur",Type:CMBase.DATA_TYPE_FLOAT, Default:1 },
		shadowOffsetX: { Name:"Shadow X",Type:CMBase.DATA_TYPE_FLOAT, Default:1 },
		shadowOffsetY: { Name:"Shadow Y",Type:CMBase.DATA_TYPE_FLOAT, Default:1 }
	},
	UnitText:
	{
		Text: { Name:"Text",Type:CMBase.DATA_TYPE_STRING, Default:null },
		font: { Name:"Font",Type:CMBase.DATA_TYPE_FONT, Default:"12px Arial" },
		strokeStyle: { Name:"Line Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(0,0,0)" },
		fillStyle: { Name:"Fill Style",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(255,255,255)" },
		lineWidth: { Name:"Width",Type:CMBase.DATA_TYPE_INTEGER, Default:1 },
		lineCap: { Name:"Line Cap",Type:CMBase.DATA_TYPE_ENUMERATED, Options:['butt','round','square'],Default:'round' },
		lineJoin: { Name:"Line Join",Type:CMBase.DATA_TYPE_ENUMERATED, Options:['bevel','round','miter'],Default:'round' }
	},
	Factors:
	{
		UnitFontHeightFactor: { Name:"Corner Width",Type:CMBase.DATA_TYPE_FLOAT, Default:0.35 }, // proportion of the height of the scale bar used for units
		LabelFontHeightFactor: { Name:"Corner Height",Type:CMBase.DATA_TYPE_FLOAT, Default:0.3 },
		ScaleBarHeightFactor: { Name:"Corner Height",Type:CMBase.DATA_TYPE_FLOAT, Default:0.3 },
		ScaleBarBaseLineFromBottomFactor: { Name:"Corner Height",Type:CMBase.DATA_TYPE_FLOAT, Default:0.2 }, // proportional distance of the bottom of the scale bar from the bottom of the background
		
		Units: { Name:"Units",Type:CMBase.DATA_TYPE_ENUMERATED, Options:[CMScaleBar.UNITS_ISO,CMScaleBar.UNITS_ENGLISH],Default:CMScaleBar.UNITS_ISO },
		Margin: { Name:"Margin",Type:CMBase.DATA_TYPE_FLOAT, Default:4 }, // margin around the scale bar
	}
};
	

/**
* Creates a new scale bar with the specified dimensions in the map.
* @public, @constructs
*/
function CMScaleBar(X,Y,Width,Height) 
{
	CMItem.call(this);
	
/*	this.TimeSlices[0].Settings.Rectangle=	
	{
		Coordinates:
		{
			Xs:[0,10], // min,max
			Ys:[0,10] // Min,max
		}
	};
	this.TimeSlices[0].Settings.RoundedRectangle=
	{
		RoundedCornerWidth:3,
		RoundedCornerHeight:3
	};
*/	this.TimeSlices[0].Settings.UnitText=
	{
	};
	this.TimeSlices[0].Settings.ScaleBar=
	{
	};
	this.TimeSlices[0].Settings.Factors=
	{
		UnitFontHeightFactor:0.35,
		LabelFontHeightFactor:0.3,
		ScaleBarHeightFactor:0.3,
		ScaleBarBaseLineFromBottomFactor:0.2,
		
		Units:CMScaleBar.UNITS_ISO,
		Margin:4,
	};
	
//	this.BottomSticky=null; // MoveFlag=null; // null for not sticky, true for move, false for size
//	this.RightSticky=null;
};

CMScaleBar.prototype=Object.create(CMItem.prototype); // inherit prototype functions from CMBase()

CMScaleBar.prototype.contructor=CMScaleBar; // override the constructor to go to ours

//***************************************************************************************
// private funtions
//***************************************************************************************
/**
* Strips the "px" from a style value (i.e. converts "100px" to "100")
*/
CMScaleBar.RemovePixels=function(Style)
{
	var Result="";
	
	if (Style!=null)
	{
		var Tokens=Style.split(" ");
		for (var i=0;i<Tokens.length;i++)
		{
			if (Tokens[i].indexOf("px")==-1) 
			{
				if (Result!="") Result+=" ";
				Result+=Tokens[i];
			}
		}
	}
	return(Result);
}

//***************************************************************************************
// Unit conversions
//***************************************************************************************

CMScaleBar.prototype.SetupScaleBar=function()
{
	if (this.TheContainer==undefined)
	{
		var TheMainContainer=this.GetParent(CMMainContainer);
		var TheCanvasContainer=TheMainContainer.GetElement(CMMainContainer.CANVAS_CONTAINER);
		
		// add the DIV container
		var TheContainer=document.createElement("div");
		TheContainer.className=this.GetSetting("ScaleBar","ClassName","CM_ScaleBar");
		
		var Width=TheContainer.offsetWidth;
		
		TheCanvasContainer.appendChild(TheContainer);
		
		this.TheContainer=TheContainer;

		// add the canvas to the container
		this.TheCanvas=document.createElement('CANVAS');
		TheContainer.appendChild(this.TheCanvas);
		
		// setup the scalebar's view with the canvas element
		this.TheView=new CMView2D();
		this.TheView.Setup(this.TheCanvas);
	}
}

//******************************************************************
// Item functions for subclasses to override
//******************************************************************
CMLog10 = Math.log10 || function(x) {
  return Math.log(x) * Math.LOG10E;
};

CMScaleBar.prototype.Paint=function(TheMapView)
{
	this.SetupScaleBar();
	
	var TheScene=this.GetParent(CMScene);
	
	var TheView=TheScene.GetView(0);
	
	TheView=this.TheView;
	//****************************************************************
	// Find the distance across the scale bar
	
	// convert the pixel coordinates to reference (map) coordinates)
	
	var PixelWidth=this.TheContainer.offsetWidth;
	var PixelHeight=this.TheContainer.offsetHeight;
	
	if (this.TheCanvas.width!=PixelWidth) this.TheCanvas.width=PixelWidth;
	if (this.TheCanvas.height!=PixelHeight) this.TheCanvas.height=PixelHeight;
		
	var MapPixelLeft=this.TheContainer.offsetLeft;
	var MapPixelRight=MapPixelLeft+PixelWidth;
	
	var MiddlePixelY=this.TheContainer.offsetTop+(PixelHeight/2);
	
	// get the map ref coordinate for the left and right sides of the scale bar
	var Lon1=TheMapView.GetRefXFromPixelX(MapPixelLeft);
	var Lat1=TheMapView.GetRefYFromPixelY(MiddlePixelY);
	
	var Lon2=TheMapView.GetRefXFromPixelX(MapPixelRight);
	var Lat2=Lat1; // latitudes are the same
	
	// convert the map coordinates to geographic coordinates
	
	var TheProjector=this.GetParent(CMMainContainer).GetProjector();
	
	if (TheProjector!=null)
	{
		var Result1=TheProjector.ProjectToGeographic(Lon1,Lat1);
		
		var Result2=TheProjector.ProjectToGeographic(Lon2,Lat2);
		
		if ((Result1!=null)&&(Result2!=null))
		{
			Lon1=Result1[0];
			Lat1=Result1[1];
			Lon2=Result2[0];
			Lat2=Result2[1];
		}
	}
	// convert the coordinates to radians
	
	Lon1=Lon1/180*Math.PI;
	Lat1=Lat1/180*Math.PI;
	Lon2=Lon2/180*Math.PI;
	Lat2=Lat2/180*Math.PI;
	
	// get the unit setting
	
	var FactorSettings=this.GetSettingGroup("Factors");
	
	var Units=FactorSettings["Units"];
	
	// find the create arc length
	
	var DeltaAngle=Math.acos(Math.sin(Lat1)*Math.sin(Lat2)+Math.cos(Lat1)*Math.cos(Lat2)*Math.cos(Math.abs(Lon2-Lon1)));
	
	var Distance=DeltaAngle*6371;// distance across entire scale bar area in km
	
	if (Units==CMScaleBar.UNITS_ENGLISH) Distance*=0.621371;
	
	//****************************************************************************
	// find the distance to use for the scale bar
	
	var UnitFontHeightFactor=FactorSettings["UnitFontHeightFactor"];
	var LabelFontHeightFactor=FactorSettings["LabelFontHeightFactor"];
	var ScaleBarHeightFactor=FactorSettings["ScaleBarHeightFactor"];
	var ScaleBarBaseLineFromBottomFactor=FactorSettings["ScaleBarBaseLineFromBottomFactor"];
	
	var Margin=FactorSettings["Margin"];

	// Find the width of a zero (for the left side)
	
	var TheLabelFontSetting=this.GetSetting("Text","font","12px Arial");
	
	TheLabelFontSetting=CMScaleBar.RemovePixels(TheLabelFontSetting);
	
	var LabelFontSize=""+(PixelHeight*LabelFontHeightFactor);
	var TheLabelFont=LabelFontSize+"px "+TheLabelFontSetting;
	
	// get the size of a character from the view
	
	var TheContext=TheView.TheContext;
	
	TheContext.font=TheLabelFont;
	
	var ZeroText=""+0;
	var ZeroWidth=TheContext.measureText(ZeroText).width;
	 
	// find the available width
	
	var AvailablePixels=PixelWidth-(ZeroWidth/2)-(Margin*2);
	
	// compute the Reference distance for each pixel (important) and 
	// the ScaleBarWidthInRefUnits that fits in the available width
	
	var RefDistancePerPixel=Distance/PixelWidth;
	
	var ScaleBarWidthInRefUnits=RefDistancePerPixel*AvailablePixels;
	
	// find the starting full bar width that is a mulitple of 10
	// this may be over 1 (10,000, 1000, 100, 10) or under 1 (0.1, 0.001)
	
	var N=CMLog10(ScaleBarWidthInRefUnits);
	N=Math.ceil(N);
	
	ScaleBarWidthInRefUnits=Math.pow(10,N);
	
	// draw the scale bar
	
	var Interval=1;
	var NumDigits=N+1;
	
	// find the initial width for the unit string
	
	var UnitString="km";
	if (Units==CMScaleBar.UNITS_ENGLISH) UnitString="miles";
	
	var TheUnitFontSetting=this.GetSetting("UnitText","font","12px Arial");
	
	TheUnitFontSetting=CMScaleBar.RemovePixels(TheUnitFontSetting);
	
	var UnitFontSize=""+(PixelHeight*UnitFontHeightFactor);
	var TheUnitFont=UnitFontSize+"px "+TheUnitFontSetting;
	
	TheContext.font=TheUnitFont;
	
//	var TheTextStyle=this.GetStyle(TheView,0,"Text"); // for the values along the ticks
	
	// find the initial unit width
//	TheView.SetStyle(TheUnitStyle);
	
	var UnitWidth=TheContext.measureText(UnitString).width;
	
	// reduce until it fits
	
	var ScaleBarWidthInPixels;
	
	var Fits=false;
	var Count=0;
	while ((Fits==false)&&(Count<100))
	{
		// see if we need to convert to meters
		
		if ((ScaleBarWidthInRefUnits<1)&&((UnitString=="km")||(UnitString=="miles"))) // 0.5
		{
			if (UnitString=="km")
			{
				ScaleBarWidthInRefUnits*=1000; // 500 meters
				UnitString="m";
				RefDistancePerPixel=RefDistancePerPixel*1000;
			}
			else
			{
				ScaleBarWidthInRefUnits*=10000; // 2640 feet in 1/2 mile
				UnitString="ft";
				RefDistancePerPixel=RefDistancePerPixel*5280;
			}
			// unit width changed
			
			TheContext.font=TheUnitFont;
			var UnitWidth=TheContext.measureText(UnitString).width;
//			TheContext.font=LabelFontStyle; // put the label font style back
		}
		// find the width of the last label
		
		var LabelText=""+ScaleBarWidthInRefUnits;
		
		TheContext.font=TheLabelFont;
		var LabelWidth=TheContext.measureText(LabelText).width;
		
		// find the width of the scale bar
		
		ScaleBarWidthInPixels=ScaleBarWidthInRefUnits/RefDistancePerPixel;
		
		// find the full width either by 1/2 the label width or the unit width, whichever is larger
		
		var FullWidth=ScaleBarWidthInPixels;
		
		if ((LabelWidth/2)>UnitWidth) FullWidth+=(LabelWidth/2);
		else FullWidth+=UnitWidth;
		
		// if the full width of the proposed scale bar fits, we're done
		
		if (FullWidth<AvailablePixels)
		{
			Fits=true;
		}
		else if (((UnitString=="m")||(UnitString=="ft"))&&(ScaleBarWidthInRefUnits<=0.01))
		{
			Fits=true;
		}
		else // reduce the interval 
		{
			if (Interval==1) // must be 1 going to 5 and down a digit
			{
				Interval=5;
				ScaleBarWidthInRefUnits=ScaleBarWidthInRefUnits/2;
			}
			else if (Interval==5)
			{
				Interval=2;
				ScaleBarWidthInRefUnits=ScaleBarWidthInRefUnits*2/5;
			}
			else // must be 2, going to 1
			{
				Interval=1;
				ScaleBarWidthInRefUnits=ScaleBarWidthInRefUnits/2;
			}
		}
		Count+=1;
	}
	
	//*********************************************************************
	// Draw the scale bar
	
	//this.CMItemRect_Paint(TheView);
	
	// setup the coordinates for the bar
	
	var ScaleBarWidthInPixels=ScaleBarWidthInRefUnits/RefDistancePerPixel;
	
	var ScaleBarLeft=Margin+(ZeroWidth/2);
	var ScaleBarRight=ScaleBarLeft+ScaleBarWidthInPixels;
	var ScaleBarY=PixelHeight-(PixelHeight*ScaleBarBaseLineFromBottomFactor);
	
	// get the settings
	
	var TheStyleSettings=this.GetStyle(TheView,0,"Style");
	var TheLabelSettings=this.GetStyle(TheView,0,"Text");
	var TheUnitTextSettings=this.GetStyle(TheView,0,"UnitText");
	var TheScaleBarSettings=this.GetStyle(TheView,0,"ScaleBar");
	
	// draw the backgorund
	
//	TheView.SetStyle(TheStyleSettings);
	
	TheContext.clearRect(0, 0, this.TheCanvas.width, this.TheCanvas.height);
/*	var RoundedCornerWidth=this.GetSetting("RoundedRectangle","RoundedCornerWidth");
	var RoundedCornerHeight=this.GetSetting("RoundedRectangle","RoundedCornerHeight");
	
	TheView.SetStyle(TheStyleSettings);
	
	TheView.PaintRoundedRect(MapPixelLeft,MapPixelRight,MinY,MinY+PixelHeight,RoundedCornerWidth,RoundedCornerHeight);
*/	
	TheView.RestoreStyle();
	
	// draw the units
	
	TheView.SetStyle(TheUnitTextSettings);
	
	TheContext.font=TheUnitFont;
	TheContext.fillText(UnitString,ScaleBarLeft+ScaleBarWidthInPixels+Margin,ScaleBarY);
	
	TheView.RestoreStyle();
	
	// draw the bar
	
	TheView.SetStyle(TheScaleBarSettings);
	
	TheContext.beginPath();
	TheContext.moveTo(ScaleBarLeft,ScaleBarY);
	TheContext.lineTo(ScaleBarRight,ScaleBarY);
	TheContext.stroke();
	
	// draw the ticks
	
	var ScaleBarHeight=PixelHeight*ScaleBarHeightFactor;
	
	TheContext.beginPath();
	TheContext.moveTo(ScaleBarLeft,ScaleBarY);
	TheContext.lineTo(ScaleBarLeft,ScaleBarY-ScaleBarHeight);
	TheContext.stroke();
	
	TheContext.beginPath();
	TheContext.moveTo(ScaleBarRight,ScaleBarY);
	TheContext.lineTo(ScaleBarRight,ScaleBarY-ScaleBarHeight);
	TheContext.stroke();
	
	TheView.RestoreStyle();
	
	// write out the labels
	
	TheView.SetStyle(TheLabelSettings);
	
	TheContext.font=TheLabelFont;
	
	var LabelBaseLine=ScaleBarY-ScaleBarHeight-Margin;
	
	TheContext.fillText("0",ScaleBarLeft-(ZeroWidth/2),LabelBaseLine);
	 
	var LabelText=""+ScaleBarWidthInRefUnits;
	var LabelWidth=TheContext.measureText(LabelText).width;
	 
	TheContext.fillText(LabelText,ScaleBarRight-LabelWidth/2,LabelBaseLine);
	
	TheView.RestoreStyle();
}

