/******************************************************************************************************************
* CMTile
*
* @module CMTile
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/
//******************************************************************
// Constants
//******************************************************************

// Definitions for the corners for polygons drawn in the tile

CMTile.UPPER_RIGHT=-1;
CMTile.LOWER_RIGHT=-2;
CMTile.LOWER_LEFT=-3;
CMTile.UPPER_LEFT=-4;

//******************************************************************
// Global Variables
//******************************************************************

// A count of the total number of tiles that have been loaded for debugging and performance monitoring
CMTile.NumTilesLoaded=0;

// static variables for the loader and the material (the material does not change the appearance of the tiles so it is just created once)
CMTile.TheLoader=null;
CMTile.TheMaterial=null;

//******************************************************************
// Tile Constructor
//******************************************************************
function CMTile(TheDataset,ZoomLevel,GlobalColumn,GlobalRow)
{
	this.GlobalRow=GlobalRow;
	this.GlobalColumn=GlobalColumn;
	this.ZoomLevel=ZoomLevel;

	this.TheData=null; // the original tile.js information with vector data
	this.TheRaster=null; // raster image
	this.TheRasterRequest=null;
	
	this.ChildTiles=null;
	this.TheDataset=TheDataset;
	
	this.LoadStatus=CMDataset.LOAD_STATUS_NONE;
	this.TheRequest=null;
	
	this.PaintTileInfo=false; // for debugging
	
	// 3D Variables
	this.TheMesh=null;
	this.MeshIsInGeo=false;
}
//***********************************************************************
// Private static functions
//***********************************************************************
/**
* Creates a grid of pixels to match the specified extent
* @static
* @private
*/
CMTile.GetSubSample=function(RequestedExtent,DEMTile,Exaggeration) 
{
	var DEMData=DEMTile.GetData();
	
	var DEMRefBounds=DEMData.RasterData.Bounds;
	
	// get the pixels, num columns and rows from the DEM
	var DEMPixelMatrix=DEMData.RasterData.Data;
	var DEMMaskMatrix=DEMData.RasterData.Mask;
	
	var DEMNumColumns=DEMPixelMatrix[0].length;
	var DEMNumRows=DEMPixelMatrix.length;
	
	// find the ref dimensions of the DEM tile
	var DEMRefWidth=DEMRefBounds.XMax-DEMRefBounds.XMin;
	var DEMRefHeight=DEMRefBounds.YMax-DEMRefBounds.YMin;
	
	// find the width of the columns and rows in the dem
	var DEMColumnWidth=DEMRefWidth/DEMNumColumns;
	var DEMRowHeight=DEMRefHeight/DEMNumRows;
	
	// find the ref dimensions of the requested area 
	var RequestedRefWidth=RequestedExtent.XMax-RequestedExtent.XMin;
	var RequestedRefHeight=RequestedExtent.YMax-RequestedExtent.YMin; // note this is positive
	
	// find the offset to the left side of the requested area in the DEM tile
	var LeftRefOffset=RequestedExtent.XMin-DEMRefBounds.XMin;
	
	// find the left-most column in the DEM tile
	var DEMStartColumn=CMUtilities.Truncate(LeftRefOffset/DEMColumnWidth);
	
	// find the bottom row in the DEM tile from the offset to the bottom of the requested area in the DEM tile
	var BottomRefOffset=RequestedExtent.YMin-DEMRefBounds.YMin;
	var DEMStartRow=CMUtilities.Truncate(BottomRefOffset/DEMRowHeight);
	
	// for now we always generate 64 rows and columns in the output
	var OutputNumRows=64;
	var OutputNumColumns=64;
	
	var SampleRate=RequestedRefWidth/(DEMColumnWidth*OutputNumColumns);
	
	var OutputColumnWidth=RequestedRefWidth/OutputNumColumns;
	var OutputRowHeight=RequestedRefHeight/OutputNumRows;
	
	//
	var ThePixels=[];
	var TheMask=[];
	
	for (var y=0;y<=OutputNumRows;y++)
	{
		ThePixels[y]=[];
		TheMask[y]=[];
		
		var Y=(RequestedExtent.YMin+(y*OutputRowHeight));
		
		for (var x=0;x<=OutputNumColumns;x++)
		{
			var X=(RequestedExtent.XMin+(x*OutputColumnWidth));
			 
			try
			{
				var TempY=256-(DEMStartRow+y*SampleRate);
				TempY=Math.floor(TempY);
				if (TempY>255) TempY=255;
				if (TempY<0) TempY=0;
				
				var TempX=DEMStartColumn+x*SampleRate;
				TempX=Math.floor(TempX);
				if (TempX>255) TempX=255;
				if (TempX<0) TempX=0;
				
				var Z=DEMPixelMatrix[TempY][TempX];
				
				Z=Z*Exaggeration; // exageration setting for this
				//Z=-Z*10;
				if (Z==-0) Z=0;
				
				if (Z!=0)
				{
					var j=12;
				}
				ThePixels[y][x]=[];
				ThePixels[y][x][0]=X;
				ThePixels[y][x][1]=Y;
				ThePixels[y][x][2]=Z;
				
				TheMask[y][x]=DEMMaskMatrix[TempY][TempX];
			}
			catch(Error)
			{
				var j=12;
				//alert(Error);
			}
		}
	}
	var TheResult=[ThePixels,TheMask];
	// compute intersectin of two extents
	
	return(TheResult);

}
//******************************************************************
// Private CMTile functions
//******************************************************************

/**
* Return the coordinates for a single polygon.  This involves converting the
* edges and sides of tiles into polygons and returning their coordinates.
* @private
*/
CMTile.prototype.GetPolygonCoordinates=function(TheEdges,TheArea,PolyIndex,PaintedEdges) 
{
	Result=null;
	
	var Factor=1/Math.pow(2,this.ZoomLevel); // width of one pixel
	var TileRefWidth=256*Factor;
	
	var RefX=(this.GlobalColumn*TileRefWidth);
	var RefY=(this.GlobalRow*TileRefWidth);
	
	var TileRefLeft=RefX
	var TileRefRight=RefX+TileRefWidth;
	var TileRefTop=RefY+TileRefWidth;
	var TileRefBottom=RefY;
	
	//
	
	var ThePolys=TheArea.Polys;
	
	var ThePoly=ThePolys[PolyIndex];
		
	if ((ThePoly.Closed)&&(ThePoly.EdgeIndexes.length==1)) // paint closed polys (both edges and fills)
	{
		var EdgeIndex=ThePoly.EdgeIndexes[0];
		
		try
		{
			var TheEdge=TheEdges[EdgeIndex];
			
			Result={ Xs:TheEdge.Xs,Ys:TheEdge.Ys };
		}
		catch( e)
		{
			throw(e);
		}
		if (PaintedEdges!=null) PaintedEdges[EdgeIndex]=true;
	}
	else // fill unclosed polys
	{
		var Xs=[];
		var Ys=[];
		var NumCoordinates=0;
		
		for (var PolyEdgeIndex=0;PolyEdgeIndex<ThePoly.EdgeIndexes.length;PolyEdgeIndex++)
		{
			var EdgeIndex=ThePoly.EdgeIndexes[PolyEdgeIndex];
			var EdgeDirection=true; // forward
			
			if (ThePoly.EdgeDirections!=undefined) EdgeDirection=ThePoly.EdgeDirections[PolyEdgeIndex];
			
			if (EdgeIndex>=0) // real edge
			{
				var EdgeXs=TheEdges[EdgeIndex].Xs;
				var EdgeYs=TheEdges[EdgeIndex].Ys;
				
				for (var k=0;k<EdgeXs.length;k++)
				{
					if (EdgeDirection)
					{
						Xs[NumCoordinates]=EdgeXs[k];
						Ys[NumCoordinates]=EdgeYs[k];
					}
					else
					{
						Xs[NumCoordinates]=EdgeXs[EdgeXs.length-k-1];
						Ys[NumCoordinates]=EdgeYs[EdgeXs.length-k-1];
					}
					NumCoordinates++;
				}
				if (PaintedEdges!=null) PaintedEdges[EdgeIndex]=true;
			}
			else // add a corner
			{
				switch (EdgeIndex)
				{
				case CMTile.UPPER_RIGHT:
					Xs[NumCoordinates]=TileRefRight
					Ys[NumCoordinates]=TileRefTop
					NumCoordinates++;
					break;
				case CMTile.LOWER_RIGHT:
					Xs[NumCoordinates]=TileRefRight
					Ys[NumCoordinates]=TileRefBottom
					NumCoordinates++;
					break;
				case CMTile.LOWER_LEFT:
					Xs[NumCoordinates]=TileRefLeft
					Ys[NumCoordinates]=TileRefBottom
					NumCoordinates++;
					break;
				case CMTile.UPPER_LEFT:
					Xs[NumCoordinates]=TileRefLeft
					Ys[NumCoordinates]=TileRefTop
					NumCoordinates++;
					break;
				}
			}
		}
		Result={ Xs:Xs,Ys:Ys };
//											TheView.PaintRefPoly(Xs,Ys,true);
//		}
	}
	return(Result);
}
//******************************************************************
// Private CMTile functions
//******************************************************************
/**
* Called by parent tiles (or dataset) until an set of data is found
* that overlaps with the requested tile
* @private
*/
CMTile.prototype.GetDataForTextureTile=function(TheView,TheTextureTile,Exaggeration)
{
	var Result=null;
	
	if (this.LoadStatus==CMDataset.LOAD_STATUS_NONE)
	{
		this.LoadTile(TheView);
	}
	else if (this.LoadStatus==CMDataset.LOAD_STATUS_LOADED)
	{
		var RequestedExtent=TheTextureTile.GetTileExtent();
		var ThisExtent=this.GetTileExtent();
		
		if (CMUtilities.BoundsIncludes(ThisExtent,RequestedExtent))
		{
			if (this.DatasetRaster!=null)
			{
				Result=CMTile.GetSubSample(RequestedExtent,this,Exaggeration);
			}
		}
		else						  
		{			
			for (var Row=0;(Row<2)&&(Result==false);Row++)
			{
				for (Column=0;(Column<2)&&(Result==false);Column++)
				{
					var TheChildTile=this.Tiles[Row][Column];
					
					Result=TheChildTile.GetDataForTextureTile(TheView,TheTextureTile,Exaggeration);
				}
			}
		}
	}
	return(Result);
}
//******************************************************************
// Protected CMTile Get/Set functions 
//******************************************************************
/**
* Allows a terrain tile to access the JSON data for a DEM that is stored in
* this tile
* @protected
*/
CMTile.prototype.GetData=function() 
{
	return(this.TheData);
}
/**
* Sets the file extenion used by GetTileImageFilePath()
* @protected
*/
CMTile.prototype.SetFileExtension=function(FileExtension) 
{
	this.Extension=FileExtension;
}
//******************************************************************
// Protected CMTile functions
//******************************************************************
/**
* Hide the mesh when the layer goes invisible
* @protected
*/
CMTile.prototype.HideMesh=function(TheLayer) 
{
	if (this.MeshIsInGeo) // need to remove this tile from mesh
	{
		var TheGeo=TheLayer.GetParent(CMGeo);
		TheGeo.RemoveOGLObjectFromGeo(this.TheMesh);
		this.MeshIsInGeo=false;
	}
}

/**
* Hides this tiles child tiles.  If called with HideSelf true, will also hide this tiles data
* and then will call child tiles to hide themselves.  Called by the dataset and then each
* parent tile calls its children.
*
* @protected
*/
CMTile.prototype.HideTiles=function(TheLayer,HideSelf) 
{ 
	if (HideSelf) this.HideMesh(TheLayer);

	if (this.ChildTiles!=null)
	{
		for (var RowIndex=0;RowIndex<2;RowIndex++)
		{
			if (this.ChildTiles[RowIndex]!=null)
			{
				for (var ColumnIndex=0;ColumnIndex<2;ColumnIndex++)
				{
					if (this.ChildTiles[RowIndex][ColumnIndex]!=null)
					{
						var TheChildTile=this.ChildTiles[RowIndex][ColumnIndex];
						
						TheChildTile.HideTiles(TheLayer,true);
					}
				}
			}
		}
	}
}

//******************************************************************
// CMTile static functions used by subclasses (e.g. CMTileOpenFormat)
// To display tiles in 3D
//******************************************************************
/**
* Get the material for displaying a tile in 3D.  The material is 
* unique for each tile and manages loading the tile
* @protected
*/
CMTile.GetNewMaterial=function(TheFileName) 
{
	var texloader = new THREE.TextureLoader();
	texloader.crossOrigin = '';

	var Opacity=1;
	var Color=0x00ffff;
	var Transparent=false;
	var Side=THREE.FrontSide;
	
	var SolidColorMaterial=new THREE.MeshLambertMaterial({
		  color: Color,
		  side:Side ,
		  opacity:Opacity,
		  transparent :false
	});	
	//console.log("CMTileOpenFormat: Loading: "+TheFileName);
	
	SolidColorMaterial.map=texloader.load(TheFileName);

	return(SolidColorMaterial);
}
/**
* Provides the distance from the View to the Tile extent
* Should probably be in the Geo class as the distance varies
* based on the projection.
*
* Extent - extent of the tile in geographic
* @protected
*/
CMTile.GetDistanceTo3DView=function(TheView,Extent,TheLayer) 
{
	var TheGeo=TheLayer.GetParent(CMGeo);
	var CameraPosition=TheView.GetPosition();
	
	// find the center of the tile
	var	Lon=(Extent.XMin+Extent.XMax)/2;
	var Lat=(Extent.YMin+Extent.YMax)/2;
	var Elevation=0;
	
	if (true)
	{
		// get the location of the tile in 3D coordinates
		var Result=TheGeo.ProjectFromGeographic(Lon, Lat, Elevation);
//		var Vector1=CMTile.GetGlobalCoordinate(TheLayer,OX,OY,0); // project to geographic and then to the current projection (could be spherical)
		
		OX=Result[0];
		OY=Result[1];
		OZ=Result[2];
	}
	var DX=CameraPosition.x-OX;
	var DY=CameraPosition.y-OY; // y is negative
	var DZ=CameraPosition.z-OZ;
	
	var Distance=Math.sqrt((DX*DX)+(DY*DY)+(DZ*DZ));
	
	return(Distance);
}


//******************************************************************
// Protected CMTile functions
//******************************************************************

/**
* Utility function to get the extent of this tile in the tile's spatial reference system.
*
* Note that this function and the CMTileOpenFormat function are slightly different
* in tha the y direction is handled differently
*
* @protected
*/
CMTile.prototype.GetTileExtent=function() 
{
	var Factor=1/Math.pow(2,this.ZoomLevel); // width of one pixel
	var TileRefWidth=256*Factor;
	
	var RefX=(this.GlobalColumn*TileRefWidth);
	var RefY=(this.GlobalRow*TileRefWidth); // rows for BlueSpray tiles are 0 on top and negative numbers going down
	
	var Extent={
		"XMin":RefX,
		"XMax":RefX+TileRefWidth,
		"YMax":RefY+TileRefWidth, // top of the tile
		"YMin":RefY // bottom of the tile
	}
	return(Extent);
}

/**
* Utility function to get the extent of this tile in geographic coordinates
*
* @protected
*/
CMTile.prototype.GetTileExtentAsGeographic=function() 
{
	var TheProjector=this.TheDataset.GetProjector();

	var Extent=this.GetTileExtent();
	
	if (TheProjector!=null)
	{
		var Result1=TheProjector.ProjectToGeographic(Extent.XMin,Extent.YMin);
		var Result2=TheProjector.ProjectToGeographic(Extent.XMax,Extent.YMax);
		
		Extent.XMin=Result1[0];
		Extent.YMin=Result1[1];
		Extent.XMax=Result2[0];
		Extent.YMax=Result2[1];
	}
	return(Extent);
}
/**
* Utility function to get the extent of this tile in global coordinates.
* This means the extent is projected from its native spatial reference,
* to geographic, and then to the global spatial reference using the 
* projector in the Geo, if any.
*
* @protected
*/
CMTile.prototype.GetTileWidthAsGlobal=function(TheLayer)
{
	var Width=0;
	
	var TheGeo=TheLayer.GetParent(CMGeo);

	var TheProjector=TheGeo.GetProjector();
	
	var TheExtent=this.GetTileExtentAsGeographic();
	
	if (TheProjector!=null)
	{
		var Result1=TheProjector.ProjectFromGeographic(TheExtent.XMin,TheExtent.YMin,0);
		var Result2=TheProjector.ProjectFromGeographic(TheExtent.XMax,TheExtent.YMax,0);
				
		var DX=Result1[0]-Result2[0];
		var DY=Result1[1]-Result2[1];
		var DZ=Result1[2]-Result2[2];
		
		Width=Math.sqrt(DX*DX+DY*DY+DZ*DZ)*2/2;
	}
	else
	{
		var DX=TheExtent.XMax-TheExtent.XMin;
		var DY=TheExtent.YMax-TheExtent.YMin;
		
		Width=Math.sqrt(DX*DX+DY*DY)*2/2;
	}
	return(Width);
}

//******************************************************************
// Protected CMTile functions overriden by CMTileOpenFormat
//******************************************************************

/**
* Returns the full path to this tiles file
* @protected
*/
CMTile.prototype.GetTileImageFilePath=function() 
{
	var Extension=this.TheDataset.GetRasterFileExtension();
	
	var FileName="Tiles_"+this.ZoomLevel+"/"+(this.GlobalColumn)+"/"+(this.GlobalRow)+Extension;
	
	var ThePath=this.TheDataset.URL+FileName;
	
	return(ThePath);
}
/**
* Function to load a single raster tile.  This is called when a tile with an associated raster
* file (PNG) is recieved.
* @protected
*/
CMTile.prototype.LoadTileRaster=function() 
{
	// setup the new Image object to hold the raster (Image)
	
	this.TheRaster=new Image(); 
	this.TheRaster.TheTile=this;
	
	// set the URL of the image which will trigger the request
	
	var ThePath=this.GetTileImageFilePath();
	
	// build the request object
	this.TheRasterRequest=
	{
		LoadStatus:CMDataset.LOAD_STATUS_NONE,
		Type:CMDataset.REQUEST_TYPE_IMAGE,
		TheImage:this.TheRaster,
		src:ThePath,
		TheFunction:function()
		{ 
			this.TheTile.TheDataset.SendMessageToListeners(CMDataset.MESSAGE_DATASET_TILE_LOADED,this.TheTile);
		}
	};
	this.TheRasterRequest.TheTile=this;
	
	//console.log("CMTile requesting tile: "+ThePath);
	
	// make the request
	CMDataset.MakeRequest(this.TheRasterRequest);
}
/*
* Sets up a grid-based goemetry to hold the coordinates for the projected space
* @protected
*/
CMTile.prototype.SetupGeometry=function(TheLayer) 
{
	var TheGeo=TheLayer.GetParent(CM3Geo);
	
	var TheGeographicExtent=this.GetTileExtentAsGeographic();
	
	var XMin=TheGeographicExtent.XMin;
	var XMax=TheGeographicExtent.XMax;
	var YMin=TheGeographicExtent.YMin;
	var YMax=TheGeographicExtent.YMax;
	
	var Width=XMax-XMin;
	var Height=YMax-YMin;
	
	var StartU=0; // x direction factors
	var EndU=1;
	var StartV=0;
	var EndV=1;
	
	if (XMin<-180) 
	{
		StartU=(-180-XMin)/Width;
		XMin=-180;
	}
	if (XMax>180) 
	{
		EndU=1-(XMax-180)/Width;
		XMax=180;
	}
	if (YMin<-90) 
	{
		StartV=(-90-YMin)/Height;
		YMin=-90;
	}
	if (YMax>90) 
	{
		EndV=1-(YMax-90)/Height;
		YMax=90;
	}
	var RowRefHeight=(YMax-YMin)/16;
	var ColumnRefWidth=(XMax-XMin)/16;
	
	var TheArray=[];
	
	for (var Row=0;Row<=16;Row++)
	{
		var TheRow=[];
		TheArray[Row]=TheRow;
		
		for (var Column=0;Column<=16;Column++)
		{
			TheRow[Column]=[];
			
			// convert tile coordinates to geographic
			var Lon=XMin+(Column*ColumnRefWidth);
			var Lat=YMin+(Row*RowRefHeight);
			
			// convert geographic coordinates to spherical
			var Elevation=0;//TheElevationGroup.Offset;
			
			// project the coordinate
			var Vector1=TheGeo.ProjectFromGeographic(Lon, Lat, Elevation);
			
			// setup the grid to contain the projected coordinate
			TheRow[Column]=Vector1;
			//TheRow[Column][1]=Vector1.y;
			//TheRow[Column][2]=Vector1.z;
		}
	}
	
	// setup the THREEE geomery which will become part of the mesh
	var NumColumns=TheArray[0].length;
	var NumRows=TheArray.length;
	
	var TheGeometry=new CM3SurfaceGeometry(NumColumns-1,NumRows-1,null,StartU,EndU,StartV,EndV);
	
	// set the positions within the grid
	TheGeometry.SetPositions(TheArray);
				
	return(TheGeometry);
}


//******************************************************************
// Painting Functions
//******************************************************************
/*
* Check if we should paint this tile or it's four child tiles.
* 
* This is a little complicated because we have to
* decide if we are going to paint this tile or paint it's children.
* If we are painting this tile, then we either need to paint the vectors
* fill the tile with a single color (uniform raster pixel or tile in the
* middle of a filled polygon
*
* Algorithm to decide if painting this tile:
*
* 	if PaintDebugTile) paint this tile
* 	else
* 		if this tile's pixel width is less than or equal to 256 paint this tile
* 		else 
* 			if any of the children have not been loaded paint this tile (and load the children)
*			if any of the children are still loading, paint this tile
*			count the number of child tiles that are
*				loaded and
*				vector data or raster data with a color or a loaded raster tile
*				or empty tiles
*			if the number is less than 4, paint this tile
*
* @protected
*/

CMTile.prototype.CheckPaintTile=function(TheLayer,TheView,ThisStepTileRefWidth) 
{
	var PaintTile=false; 

	// see if we should paint because we are debugging a tile
	if (this.TheDataset.PaintDebugTile)
	{
		if ((this.TheDataset.DebugZoomLevel==this.ZoomLevel)&&
			(this.TheDataset.DebugGlobalColumn==this.GlobalColumn)&&
			(this.TheDataset.DebugGlobalRow==this.GlobalRow))
		{
			PaintTile=true; // paint the tile being debugged
		}
	}
	else // otherwise, check if this tile needs to be painted
	{
		if (PaintTile==false) // see if we should 
		{
			// If this tile has a high enough resolution, just paint it
			if ((typeof(CM3View)!="undefined")&&(TheView instanceof CM3View))
			{
				var Distance=CMTile.GetDistanceTo3DView(TheView,this.GetTileExtentAsGeographic(),TheLayer);
				
				if (Distance>ThisStepTileRefWidth*3) PaintTile=true;
//				PaintTile=true; // have to determine distance from View to Tile
			}
			else // see if we are zoomed out enough that this tile's resolution is good enough
			{
				var ThisStepsTilePixelWidth=TheView.GetPixelWidthFromRefWidth(ThisStepTileRefWidth);
	
				if (ThisStepsTilePixelWidth<=256) PaintTile=true; // resolution is low, paint this tile
			}
			// If there are no children, paint this tile
			if (this.ChildTiles==null)  PaintTile=true; // tile has no children, paint this tile
			
			// If we're not going to paint this tile, check to see if we should paint, or load, the child tiles
			if (PaintTile==false) // see if we have 4 child tiles ready to be painted (or are empty)
			{
				if (this.TheData.NumChildTiles>0) // this tile does have child tiles
				{
					NumLoadedChildTiles=0;
					
					// go through the 4 child tiles
					for (var RowIndex=0;RowIndex<2;RowIndex++)
					{
						for (var ColumnIndex=0;ColumnIndex<2;ColumnIndex++)
						{
							if (this.TheData.ChildTiles[RowIndex][ColumnIndex]>0) // if th
							{
								var TheChildTile=this.ChildTiles[RowIndex][ColumnIndex];

								if (TheChildTile.LoadStatus==CMDataset.LOAD_STATUS_NONE) 
								{
									// tile child tile has not been loaded, load it now for the next repaint
									TheChildTile.LoadTile(TheView);
									// have to paint this tile
									PaintTile=true;
								}
								else if (TheChildTile.LoadStatus==CMDataset.LOAD_STATUS_LOADING)
								{
									// cihld tile has not finished loading, have to paint this tile
									PaintTile=true;
								}
								else 
								{
									// 
									if ("Features" in this.TheData) // vector data
									{
										NumLoadedChildTiles++;
									}
									else if ("FillColor" in this.TheData) // 
									{
										NumLoadedChildTiles++; // just a color
									}
									else if ("RasterExtension" in this.TheData) //raster data
									{
										if (this.TheRasterRequest.LoadStatus==CMDataset.LOAD_STATUS_LOADED)
										{
											NumLoadedChildTiles++;
										}
										else if (this.TheRasterRequest.LoadStatus==CMDataset.LOAD_STATUS_CANCELED)
										{
											this.LoadTileRaster(".png");
										}
									}
									else if ("NumClusters" in this.TheData)
									{
										NumLoadedChildTiles++;
									}
									else if ("RasterData" in this.TheData)
									{
										NumLoadedChildTiles++;
									}
								}
							}
							else if (this.TheData.ChildTiles[RowIndex][ColumnIndex]==0) // empty tile
							{
								NumLoadedChildTiles++;
							}
						}
					}
					// unless this tile is completely covered by child tiles, paint it	
					if (NumLoadedChildTiles<4) PaintTile=true;
				}
			}
		}
	}
	return(PaintTile);
}


//******************************************************************
// Public CMTile functions
//******************************************************************

/**
* Override the function for obtaining tiles.
* Create a single image tile and attempt to load it.
* Called by CMTile and by the CMDatasetPyramid.
* @public
*/
CMTile.prototype.LoadTile=function() 
{
	var FileName="Tiles_"+this.ZoomLevel+"/"+(this.GlobalColumn)+"/"+(this.GlobalRow)+".js";
	
	var URL=this.TheDataset.URL+FileName;

	var TheRequest=new XMLHttpRequest();
	
	this.TheRequest=TheRequest;
	
	TheRequest.open("GET",URL,true);
	//TheRequest.TheView=TheView;
	TheRequest.TheURL=URL;
	TheRequest.TheDataset=this.TheDataset;
	TheRequest.TheTile=this;
	TheRequest.FileName=FileName;
	
	this.LoadStatus=CMDataset.LOAD_STATUS_LOADING;
	
	TheRequest.onreadystatechange=function() 
	{
		if( this.readyState == 4)  // done
		{
//			alert("status="+this.status);
			if( this.status == 200) // OK
			{
				// get the response text and parse it (text contains the information for the tile)
				
				var TheText=this.responseText;

				var TheGeoJSONObject=JSON.parse(TheText);
			
				// setup the tile with the specified data
				
				var TheTile=this.TheTile; // get the tile just to make the code more readable
				
				CMTile.NumTilesLoaded++; // keep track of the total number of tiles loaded
				
				TheTile.TheData=TheGeoJSONObject; // gset the data from the server into the tile
				
				// we are now loaded and ready to be painted
				
				TheTile.LoadStatus=CMDataset.LOAD_STATUS_LOADED;

				if ("RasterExtension" in TheTile.TheData) // raster data
				{
					TheTile.LoadTileRaster(".png"); // try to immeidately load the raster
				}
				if ("RasterData" in TheTile.TheData) // raster data
				{
					TheTile.DatasetRaster=new CMDatasetRaster();
					
					var MaxPixelValues=this.TheDataset.MaxPixelValues;
					var MinPixelValues=this.TheDataset.MinPixelValues;
					
					TheTile.TheData.RasterData.MaxPixelValues=MaxPixelValues;
					TheTile.TheData.RasterData.MinPixelValues=MinPixelValues;
					
					TheTile.DatasetRaster.SetData(TheTile.TheData.RasterData);
				}
				//***********************************************************************************
				// create the children, if any
				
				if (TheTile.TheData.NumChildTiles>0) // tile has children
				{
					for (var RowIndex=0;RowIndex<2;RowIndex++)
					{
						for (var ColumnIndex=0;ColumnIndex<2;ColumnIndex++)
						{
							// negative for one feature ID, >0 for mulitple and must load a child tile
							if (TheTile.TheData.ChildTiles[RowIndex][ColumnIndex]>0)
							{
								if (TheTile.ChildTiles==null) // initialize the array
								{
									TheTile.ChildTiles=[];
									TheTile.ChildTiles[0]=[null,null];
									TheTile.ChildTiles[1]=[null,null];
								}
								
								var ChildZoomLevel=TheTile.ZoomLevel+1;
								var ChildColumn=(TheTile.GlobalColumn*2)+ColumnIndex;
								var ChildRow=(TheTile.GlobalRow*2)+(1-RowIndex);
								
								// create the tile, loads are initiated on paint
								var TheChildTile=new CMTile(this.TheDataset,ChildZoomLevel,ChildColumn,ChildRow);
								
								if (this.PaintTileInfo) TheChildTile.SetPaintTileInfo(this.PaintTileInfo);
								
								TheTile.ChildTiles[RowIndex][ColumnIndex]=TheChildTile;
							}
						}
					}
				}
				// zoom to bounds if desired and repaint
				
				TheTile.TheDataset.SendMessageToListeners(CMDataset.MESSAGE_DATASET_TILE_LOADED,TheTile);
				
				//if (this.OnLoadFunction!=undefined) this.OnLoadFunction(TheTile,this.TheView);
				
/*				if (this.ZoomToBounds)
				{
					this.TheView.ZoomToBounds(this.TheDataset.GetBounds());
				}
				this.TheDataset.GetParent(CMScene).Repaint(); 
*/			}
			else alert("HTTP error "+this.status+" "+this.statusText+" ("+this.TheURL+")");
		}
	}
	TheRequest.send();
}

//***********************************************************************
// Public CMTile functions
//***********************************************************************
/**
* Paint a single tile.  Adds appropriate content to the Geo based on the content type.
* @public
*/
CMTile.prototype.PaintTile=function(TheLayer,TheView,SelectedOnly,ThisStepTileRefWidth) 
{
	var NumPainted=0; // 1 if this tile painted it's contents, 0 otherwise
		
	var TileRefWidth=ThisStepTileRefWidth;
	var RefX=(this.GlobalColumn*TileRefWidth);
	var RefY=(this.GlobalRow*TileRefWidth);
	
	var Extent=this.GetTileExtent();
	
	var PaintTile=true; // true to paint this tile, false to paint children
	
	// check for 3D
	if ((typeof(CM3View)!="undefined")&&(TheView instanceof CM3View)) // have a 3d view
	{
		if ((TheLayer.GetVisible())&&(this.TheData!=null))
		{
			// jjg - cheating a bit here and finding the disance from one corner of the tile
			// to the other corner (should do this in the projector where arc disances can be computed)
			
			var TileRefWidth2=this.GetTileWidthAsGlobal(TheLayer);
			
			PaintTile=this.CheckPaintTile(TheLayer,TheView,TileRefWidth2);
			
			if (PaintTile) // painting this tile
			{
				//*************************************************************************
				// paint a raster from a raster file (e.g. PNG or JPG)
				if ("RasterExtension" in this.TheData)  // raster file
				{
					if (this.MeshIsInGeo==false) // the mesh is not in the Geo, we need to add it
					{
						if (this.TheMesh==null) // the mesh is not available, we need to create it
						{
							if (TheLayer instanceof CM3LayerTerrain) // try to get the DEM info
							{
								var DEMDataset=TheLayer.GetDEMDataset();
								var Exaggeration=TheLayer.GetSetting("Elevation","Exaggeration",1);
								
								var TheMatrixes=DEMDataset.GetDataForTextureTile(TheView,this,Exaggeration);
								
								if (TheMatrixes!=null)
								{
									var TheArray=TheMatrixes[0];
									var TheMask=TheMatrixes[1];
									
									//  
									var NumColumns=TheArray[0].length;
									var NumRows=TheArray.length;
									
									var TheGeometry=new CM3SurfaceGeometry(NumColumns-1,NumRows-1,TheMask);
									
									TheGeometry.SetPositions(TheArray);
									
									// setup the material
									var TheFileName=this.GetTileImageFilePath();
									
									// setup the material
									var SolidColorMaterial=CMTile.GetNewMaterial(TheFileName);
								
									// setup the mesh
									this.TheMesh=new THREE.Mesh(TheGeometry,SolidColorMaterial);
								}
							}
							else // regular layer without terrain
							{
								var TheFileName=this.GetTileImageFilePath(); // first for debugging
								
								// try to get the DEM info
								var TheGeometry=this.SetupGeometry(TheLayer);
								
								// setup the material
								var SolidColorMaterial=CMTile.GetNewMaterial(TheFileName);
							
								// setup the mesh
								this.TheMesh=new THREE.Mesh(TheGeometry,SolidColorMaterial);
							}
							
						}
						// add the mesh to the geo
						
						if (this.TheMesh!=null)
						{
							var TheGeo=TheLayer.GetParent(CMGeo);
							TheGeo.AddOGLObjectToGeo(this.TheMesh);
							
//							var helper = new THREE.VertexNormalsHelper( this.TheMesh, 20, 0x00ff00, 10 );
//							helper.rotateX(-Math.PI/2);							
//							TheGeo.AddOGLObjectToGeo(helper);
							
							this.MeshIsInGeo=true;
						}
					}
				}
			}
			// 
			if (PaintTile==false) // remove this tile from the Geo, if needed, and paint the children
			{
				if (this.MeshIsInGeo) // need to remove this tile from mesh
				{
					var TheGeo=TheLayer.GetParent(CMGeo);
					TheGeo.RemoveOGLObjectFromGeo(this.TheMesh);
					this.MeshIsInGeo=false;
				}
			}
		}
	}
	else if(CMUtilities.BoundsOverlap(Extent,TheView.GetBounds()))
	{
		if ((this.TheData!=null)) // data will be null until recieved
		{
			var TileRefLeft=RefX
			var TileRefRight=RefX+TileRefWidth;
			var TileRefTop=RefY+TileRefWidth;
			var TileRefBottom=RefY;
			
			// paint the child
			
			PaintTile=this.CheckPaintTile(TheLayer,TheView,ThisStepTileRefWidth);
			
			var NumChildrenPainted=0;
			if (PaintTile) // paint this tile instead of the children
			{
				//*************************************************************************
				// vector data
				if (("Features" in this.TheData)) // vector data 
				{
					if ((this.ZoomLevel==4)&&(this.GlobalColumn==11)&&(this.GlobalRow==3))
					{
						var j=12;
					}
					// Get the features from the data
					var TheFeatures=this.TheData.Features;
				
					// setup the PaintedEdges array
					var PaintedEdges=null;
					var TheEdges=this.TheData.Edges;
/*					if (this.TheData.Edges!=undefined)
					{
						var TheEdges=this.TheData.Edges;
						
						for (var j=0;j<TheEdges.length;j++) PaintedEdges[j]=false;
					}
*/					
					// draw each feature in this tile
					for (var i=0; i < TheFeatures.length; i++) 
					{
						var ViewBounds=TheView.GetBounds();
						
			//			if (CMUtilities.BoundsOverlap(ViewBounds,this.FeatureBounds[i]))
						{
							var TheFeature=TheFeatures[i];
							
							var TheAreas=TheFeature.Areas;
							
							for (var AreaIndex=0;AreaIndex<TheAreas.length;AreaIndex++)
							{
								var TheArea=TheAreas[AreaIndex];
								
								//var TheEdges=TheArea.Edges;
								var ThePolys=TheArea.Polys;
								
								for (var PolyIndex=0;PolyIndex<ThePolys.length;PolyIndex++)
								{
									var ThePoly=ThePolys[PolyIndex];
									
									if (ThePoly.Closed)
									{
										var PolygonCoordinates=this.GetPolygonCoordinates(TheEdges,TheArea,PolyIndex,PaintedEdges);
										
										if (PolygonCoordinates.Xs.length>2)
										{
											var Poly=[PolygonCoordinates.Xs,PolygonCoordinates.Ys];
											var AreaCoordinates=[Poly];
											
											var TheType=CMDatasetVector.TYPE_POLYLINES;
											if (ThePolys[PolyIndex].Closed) TheType=CMDatasetVector.TYPE_POLYGONS;
											
											TheLayer.PaintRefArea(TheView,i,AreaCoordinates,TheType,false,false);
										}
									}
									// painting edges here and below
//									TheView.PaintRefPoly(PolygonCoordinates.Xs,PolygonCoordinates.Ys,ThePolys[PolyIndex].Closed,false);
								}
							} // for AreaIndex
						}
					} // for i<TheFeatures.length
					
					// paint the edges
					if (TheEdges!=undefined) // happens when we have features that surround the tile
					{
						for (var EdgeIndex=0;EdgeIndex<TheEdges.length;EdgeIndex++)
						{
	//						if (PaintedEdges[EdgeIndex]==false)
							{
								var Xs=TheEdges[EdgeIndex].Xs;
								var Ys=TheEdges[EdgeIndex].Ys;
								
								if (Xs.length>2)
								{
									var Poly=[Xs,Ys];
									var AreaCoordinates=[Poly];
									
									TheLayer.PaintRefArea(TheView,i,AreaCoordinates,CMDatasetVector.TYPE_POLYLINES,false,false)
								}
							}
						}
					}
				}
				//*************************************************************************
				// paint a raster from a raster file (e.g. PNG or JPG)
				if ("RasterExtension" in this.TheData)  // raster file
				{
					if (this.TheRasterRequest!=null)
					{
						if (this.TheRasterRequest.LoadStatus==CMDataset.LOAD_STATUS_LOADED)
						{
							var ImageRefWidth=TileRefWidth;
							var ImageRefHeight=-TileRefWidth;
							
							var ImageRefX=RefX;
							var ImageRefY=RefY+TileRefWidth;
							
							//console.log("Painting tile: Zoom: "+this.ZoomLevel+" ("+this.GlobalColumn+","+this.GlobalRow+")");

							TheView.PaintRefImageScaled(this.TheRaster,ImageRefX,ImageRefY,ImageRefWidth,ImageRefHeight);
						}
						else if (this.TheRasterRequest.LoadStatus==CMDataset.LOAD_STATUS_CANCELED)
						{
							this.LoadTileRaster();
						}
					}
				}
				//*************************************************************************
				// just fill in the raster
				if ("FillColor" in this.TheData) // just paint the color
				{
					var TheColor=this.TheData.TheColor;
				
					if (TheColor!==null)
					{
						var Test="rgb("+TheColor[0]+","+TheColor[1]+","+TheColor[2]+")";
						
						var TheStyle2={
							"fillStyle":""+Test+"",
							"lineWidth":0
						};
						if (TheStyle2!=null) TheView.SetStyle(TheStyle2);
						
						TheView.PaintRefBounds(Extent);
						
						if (TheStyle2!=null) TheView.RestoreStyle();
					}
				}
				//*************************************************************************
				// raster data in text form
				if ("RasterData" in this.TheData) // raster data
				{
					this.DatasetRaster.Paint(TheView);
				}
				//*************************************************************************
				// clustered point data
				if ("NumClusters" in this.TheData)
				{
					var NumClusters=this.TheData.NumClusters;
					var ClusterCoordinates=this.TheData.ClusterCoordinates;
					
					for (var i=0;(i <NumClusters); i++) 
					{
						var X=ClusterCoordinates[0][i];
						var Y=ClusterCoordinates[1][i];
						
						TheView.PaintRefCircle(X,Y,10);
					}
				}
			}
		}
	}
	// if nothing was painted above and we have data, try to paint the children
	if (this.TheData!=null) // data will be null until recieved
	{
		if (PaintTile) // 
		{
			//*************************************************************************
			// paint the debugging information
			
			if ((this.PaintTileInfo))
			{
				var TheStyle2={
					"font":"20px Arial",
					"fillStyle":"red",
					"lineWidth":1,
					"strokeColor": "red",
					"strokeStyle":"#000",
				};
				if (TheStyle2!=null) TheView.SetStyle(TheStyle2);
					
				var CenterX=RefX+TileRefWidth/2;
				var CenterY=RefY+TileRefWidth/2;
				
				var TheText=this.ZoomLevel+"_"+this.GlobalColumn+"_"+this.GlobalRow;
				
				TheView.PaintRefText(TheText,CenterX,CenterY,20);
				
				TheView.PaintRefLine(RefX,RefY,RefX,RefY+TileRefWidth); // left
				TheView.PaintRefLine(RefX+TileRefWidth,RefY,RefX+TileRefWidth,RefY+TileRefWidth); // right
				
				TheView.PaintRefLine(RefX,RefY,RefX+TileRefWidth,RefY); // top 
				TheView.PaintRefLine(RefX,RefY+TileRefWidth,RefX+TileRefWidth,RefY+TileRefWidth); // bottom
				
				if (TheStyle2!=null) TheView.RestoreStyle();
			}
		}
		if (PaintTile==false) // paint the child tiles in this tile
		{
			if (this.TheData.NumChildTiles>0)
			{
				for (var RowIndex=0;RowIndex<2;RowIndex++)
				{
					for (var ColumnIndex=0;ColumnIndex<2;ColumnIndex++)
					{
						// negative for one feature ID, >0 for mulitple and must load a child tile
						
						if (this.TheData.ChildTiles[RowIndex][ColumnIndex]>0)
						{
							var TheChildTile=this.ChildTiles[RowIndex][ColumnIndex];
							
							// the tiles in the next step are 1/2 the reference width of the tiles in this step
							
							var NextStepTileRefWidth=ThisStepTileRefWidth/2;
							
							NumChildrenPainted+=TheChildTile.PaintTile(TheLayer,TheView,SelectedOnly,NextStepTileRefWidth);
						}
					}
				}
			}
		}
		NumPainted++;
		
		// if the children of this tile painted, we do not need to paint
		
		if (NumChildrenPainted<4)
		{
			var Test=12;
		}
	}
	
	return(NumPainted);
}
//******************************************************************
// Public CMItem functions
//******************************************************************
/**
* Adds this tile to the specific boundary. Called by CMDatasetPyramid 
* @public
*/
CMTile.prototype.AddToFeatureBounds=function(FeatureIndex,Bounds) 
{
	if ("Features" in this.TheData) // vector data
	{
		var TheFeatures=this.TheData.Features;
		var TheEdges=this.TheData.TheEdges;
		
		var Found=false;
		for (var i=0;(i < TheFeatures.length)&&(Found==false); i++) 
		{
			var TheFeature=TheFeatures[i];
			
			if (TheFeature.LayerFeatureIndex==FeatureIndex)
			{
				Found=true;
				
				var TheAreas=TheFeature.Areas;
				
				for (var AreaIndex=0;(AreaIndex<TheAreas.length);AreaIndex++)
				{
					var TheArea=TheAreas[AreaIndex];
					
					var ThePolys=TheArea.Polys;
					
					for (var PolyIndex=0;(PolyIndex<ThePolys.length);PolyIndex++)
					{
						var PolygonCoordinates=this.GetPolygonCoordinates(TheEdges,TheArea,PolyIndex,null);
					
						var PolygonBounds=CMUtilities.GetPolygonBounds(PolygonCoordinates.Xs,
							PolygonCoordinates.Ys,PolygonCoordinates.Xs.length);
						
						Bounds=CMUtilities.AddToBounds(Bounds,PolygonBounds);
					}
				}
			}
		}
	}
	if ("NumClusters" in this.TheData) // point data
	{
		var NumClusters=this.TheData.NumClusters;
		var ClusterCoordinates=this.TheData.ClusterCoordinates;
		
		for (var i=0;(i <NumClusters)&&(Found==false); i++) 
		{
			var X=ClusterCoordinates[0][i];
			var Y=ClusterCoordinates[1][i];
			
			var PointBounds={
				XMax:X,
				XMin:X,
				YMax:Y,
				YMin:Y
			};
			
			Bounds=CMUtilities.AddToBounds(Bounds,PointBounds);
		}
	}
 	return(Bounds);
};

//******************************************************************
// Public CMItem Event Functions
//******************************************************************
CMTile.prototype.In=function(TheView,RefX,RefY,RefTolerance) 
{
	var FeatureIndex=-1;
	
	if ((this.TheData!=null)&&("Features" in this.TheData)) // have vector data
	{
		var ViewBounds=TheView.GetBounds();
	
		var Extent=this.GetTileExtent();
		
		var ExpandedExtent=CMUtilities.CloneBounds(Extent);
		
		CMUtilities.ExpandBounds(ExpandedExtent,RefTolerance);
		
		if ((CMUtilities.BoundsOverlap(ViewBounds,ExpandedExtent)))
		{
			var TheFeatures=this.TheData.Features;
			var TheEdges=this.TheData.Edges;
			for (var i=0;(i < TheFeatures.length)&&(FeatureIndex==-1); i++) 
			{
				
	//			if (CMUtilities.BoundsOverlap(ViewBounds,this.FeatureBounds[i]))
				{
					var TheFeature=TheFeatures[i];
					
					var TheAreas=TheFeature.Areas;
					
					for (var AreaIndex=0;(AreaIndex<TheAreas.length)&&(FeatureIndex==-1);AreaIndex++)
					{
						var TheArea=TheAreas[AreaIndex];
						
						var ThePolys=TheArea.Polys;
						
						for (var PolyIndex=0;(PolyIndex<ThePolys.length)&&(FeatureIndex==-1);PolyIndex++)
						{
							var ThePoly=TheArea.Polys[PolyIndex];
							
							var PolygonCoordinates=this.GetPolygonCoordinates(TheEdges,TheArea,PolyIndex,null);
						
							var Inside=false;
							
							if (ThePoly.Closed)
							{
								Inside=CMUtilities.InsideAPolygon(RefX,RefY,PolygonCoordinates.Xs,PolygonCoordinates.Ys,PolygonCoordinates.Xs.length);
							}
							else
							{
								//var Tolerance=TheView.GetRefWidthFromPixelWidth(6); // jjg= tolerance needs to be passed in 
								Inside=CMUtilities.InPolyline(RefX,RefY,PolygonCoordinates.Xs,PolygonCoordinates.Ys,RefTolerance);
							}
							if (Inside) 
							{
								FeatureIndex=TheFeature.LayerFeatureIndex;
							}
						}
					}
				}
			}
			if (FeatureIndex==-1) // check with loaded children
			{
				if (this.TheData.NumChildTiles>0)
				{
					for (var RowIndex=0;RowIndex<2;RowIndex++)
					{
						for (var ColumnIndex=0;ColumnIndex<2;ColumnIndex++)
						{
							if (this.TheData.ChildTiles[RowIndex][ColumnIndex]>0)
							{
								//FeatureIndex=this.ChildTiles[RowIndex][ColumnIndex].In(TheView,RefX,RefY,RefTolerance);
							}
						}
					}
				}
			}
		}
	}
	return(FeatureIndex);
};
CMTile.prototype.MouseDown=function(TheView,RefX,RefY,RefTolerance) 
{
	var Used=false;
	
	if ((this.TheData!=null)&&
		 ((TheView.GetTool()==CMView.TOOL_INFO)||(TheView.GetTool()==CMView.TOOL_SELECT)))
	{
		if (("Features" in this.TheData)||("NumClusters" in this.TheData)) // vector data
		{
			var FeatureIndex=this.In(TheView,RefX,RefY,RefTolerance);
			
			if (FeatureIndex!=-1)
			{
				this.TheDataset.ShowInfoWindow(FeatureIndex,TheView,RefX,RefY);
		
				Used=true;
			}
			else // call children?
			{
			}
		}
		else if (("RasterExtension" in this.TheData)) // raster data
		{
			if (this.TheRaster!=null)
			{
				var Factor=Math.pow(2,this.ZoomLevel);
				var TileRefWidth=this.TheDataset.TileRefWidth/Factor;
				
				var TileRefX=(this.GlobalColumn*TileRefWidth);
				var TileRefY=(this.GlobalRow*TileRefWidth);
				
				var PixelX=parseInt((RefX-TileRefX)/TileRefWidth*256);
				var PixelY=parseInt((TileRefY-RefY)/TileRefWidth*256);
				
				if ((PixelX>=0)&&(PixelX<256)&&(PixelY>=0)&&(PixelY<256)) // in tile
				{
					var TheHTML="Pixel X:"+PixelX+" Y: "+PixelY;
				
					var Index=PixelY*(256*3)+(PixelX*3);
					
					// 
					
					 var canvas = document.createElement("canvas");
					canvas.width = this.TheRaster.width;
					canvas.height =  this.TheRaster.height;
					
					// Copy the image contents to the canvas
					var ctx = canvas.getContext("2d");
					ctx.drawImage( this.TheRaster, 0, 0);
					
					var ImageData=ctx.getImageData(PixelX,PixelY,1,1);
					
					TheHTML+=" Value:"+ImageData.data[0]+","+ImageData.data[1]+","+ImageData.data[2];
					
					var InfoWindow=TheView.CreateInfoWindow("CMTile.InfoWindow",RefX,RefY,200,30,TheHTML);
				
					CMMainContainer.SetPopupWindow(InfoWindow);
					
					Used=true;
				}
			}
		}
		//
		if (Used==false)
		{
			if (this.TheData.NumChildTiles>0)
			{
				for (var RowIndex=0;RowIndex<2;RowIndex++)
				{
					for (var ColumnIndex=0;ColumnIndex<2;ColumnIndex++)
					{
						if (this.TheData.ChildTiles[RowIndex][ColumnIndex]>0)
						{
							Used=this.ChildTiles[RowIndex][ColumnIndex].MouseDown(TheView,RefX,RefY,RefTolerance);
						}
					}
				}
			}
		}
	}
	return(Used);
};


//******************************************************************
// Public CMTile Get/Set functions 
//******************************************************************
/**
* Sets whether the tile will paint it's debugging information.
* The parent dataset has a setting for this.
* @public
*/
CMTile.prototype.SetPaintTileInfo=function(NewPaintTileInfo) 
{
	this.PaintTileInfo=NewPaintTileInfo;
}
