/******************************************************************************************************************
* CMTileOpenFormat
*
* @module CMTileOpenFormat
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/
//******************************************************************
// Global Variables
//******************************************************************

// A count of the total number of tiles that have been loaded for debugging and performance monitoring
CMTileOpenFormat.NumTilesLoaded=0;

//******************************************************************
// Tile Constructor
//******************************************************************
function CMTileOpenFormat(TheDataset,ZoomLevel,GlobalColumn,GlobalRow)
{
	CMTile.call(this);
	
	// Settings?
	this.PaintTileInfo=false; // for debugging (jjg - move to settings)
	
	// Properties
	this.GlobalRow=GlobalRow;
	this.GlobalColumn=GlobalColumn;
	this.ZoomLevel=ZoomLevel;

	this.TheRaster=null; // raster image
	this.TheRasterRequest=null;
	
	this.ChildTiles=null;
	this.TheDataset=TheDataset;
	
	this.MeshIsInGeo=false;
	this.TheMesh=null;
}
CMTileOpenFormat.prototype=Object.create(CMTile.prototype); // inherit prototype functions from PanelBase()

CMTileOpenFormat.prototype.contructor=CMTileOpenFormat; // override the constructor to go to ours

//******************************************************************
// Private CMTileOpenFormat functions
//******************************************************************

/**
* Converts the specified coordinate into global coordinate using the 
* current projection.  This function is specific for the OpenFormat
* because it assumes the tiles are in the Google Mercator spatial reference.
* @private
*/
CMTileOpenFormat.GetGlobalCoordinate=function(TheLayer,OX,OY,Elevation)
{
	var TheGeo=TheLayer.GetParent(CMGeo);
	var TheGoogleMapsProjector=new CMProjectorGoogleMaps();
	TheGoogleMapsProjector.SetZoomLevel(18);

	var Result=TheGoogleMapsProjector.ProjectToGeographic(OX,OY);

	var Lon=Result[0];
	var Lat=Result[1];
	
	// project the coordinate
	
	var Result=TheGeo.ProjectFromGeographic(Lon, Lat, Elevation);
	
	return(Result);
}

//******************************************************************
// Protected CMTile functions
//******************************************************************
/**
* Returns the full path to this tile's file
* @protected
*/
CMTileOpenFormat.prototype.GetTileImageFilePath=function() 
{
	var ThePath=null;
	
	var ZoomLevel=this.ZoomLevel;
	var Row=this.GlobalRow;
	var Column=this.GlobalColumn;
	
	var FileExtension=this.TheDataset.GetRasterFileExtension();
	
	var ZoomLevel=18+ZoomLevel;//+this.ZoomLevelOffset;
	
	var StepColumn=Column;
	var StepRow=Row;
	
	var CoordinateValueOrder=this.TheDataset.GetCoordinateValueOrder();
	
	if ((StepColumn>=0)&&(StepRow>=0)) // only load valid tiles
	{
		var FileName="";
		
		for (var i=0;i<3;i++)
		{
			if (i!=0) FileName+="/";
			
			if (CoordinateValueOrder[i]=="x") { FileName+=StepColumn; }
			if (CoordinateValueOrder[i]=="y") { FileName+=StepRow; }
			if (CoordinateValueOrder[i]=="z") { FileName+=ZoomLevel; }
		}
		if (FileExtension!=null) FileName+="."+FileExtension;
		
		var ThePath=this.TheDataset.URL+FileName;
	}
	return(ThePath);
}

CMTileOpenFormat.prototype.GetTileExtent=function() 
{
	var Factor=1/Math.pow(2,this.ZoomLevel); // width of one pixel
	var TileRefWidth=256*Factor; // width of a tile
	
	var RefX=(this.GlobalColumn*TileRefWidth);
	var RefY=-(this.GlobalRow*TileRefWidth); // The rows go from top to bottom so the row is 0 on top, large positive number on the bottom
	
	var Extent={
		"XMin":RefX,
		"XMax":RefX+TileRefWidth,
		"YMax":RefY, // top of the tile
		"YMin":RefY-TileRefWidth // bottom of the tile
	}
	return(Extent);
}
/**
* OpenFormat tiles are always in GoogleMercator projection so we don't
* need to have a projector in the Geo and we may not so we create one
* and use it here
*/
CMTileOpenFormat.TheProjector=null;

CMTileOpenFormat.prototype.GetTileExtentAsGeographic=function() 
{
	if (CMTileOpenFormat.TheProjector==null)
	{
		//var TheProjector=this.TheDataset.GetProjector();
		CMTileOpenFormat.TheProjector=new CMProjectorGoogleMaps();
		CMTileOpenFormat.TheProjector.SetZoomLevel(18);
	}
	var Extent=this.GetTileExtent();
	
	if (CMTileOpenFormat.TheProjector!=null)
	{
		var Result1=CMTileOpenFormat.TheProjector.ProjectToGeographic(Extent.XMin,Extent.YMin);
		var Result2=CMTileOpenFormat.TheProjector.ProjectToGeographic(Extent.XMax,Extent.YMax);
		
		Extent.XMin=Result1[0];
		Extent.YMin=Result1[1];
		Extent.XMax=Result2[0];
		Extent.YMax=Result2[1];
	}
	return(Extent);
}

//******************************************************************
// Protected CMTile functions
//******************************************************************

/*
* Check if we should paint this tile or it's four child tiles.
* 
* This is a little complicated because we have to
* decide if we are going to paint this tile or paint it's children.
* If we are painting this tile, then we either need to paint the vectors
* fill the tile with a single color (uniform raster pixel or tile in the
* middle of a filled polygon
*
* Algorithm to decide if painting this tile:
*
* 	if PaintDebugTile) paint this tile
* 	else
* 		if this tile's pixel width is less than or equal to 256 paint this tile
* 		else 
* 			if any of the children have not been loaded paint this tile (and load the children)
*			if any of the children are still loading, paint this tile
*			count the number of child tiles that are
*				loaded and
*				vector data or raster data with a color or a loaded raster tile
*				or empty tiles
*			if the number is less than 4, paint this tile
*
* @private
*/

CMTileOpenFormat.prototype.CheckPaintTile=function(TheView,ThisStepTileRefWidth,TheLayer) 
{
	var PaintTile=false; 

	// see if we should paint because we are debugging a tile
	if (this.TheDataset.PaintDebugTile)
	{
		if ((this.TheDataset.DebugZoomLevel==this.ZoomLevel)&&
			(this.TheDataset.DebugGlobalColumn==this.GlobalColumn)&&
			(this.TheDataset.DebugGlobalRow==this.GlobalRow))
		{
			PaintTile=true; // paint the tile being debugged
		}
	}
	else // otherwise, check if this tile needs to be painted
	{
		if (PaintTile==false) // see if we should 
		{
			var PaintChildren=false;
			
			// If this tile has a high enough resolution, just paint it
			if ((typeof(CM3View)!="undefined")&&(TheView instanceof CM3View))
			{
				var Distance=CMTile.GetDistanceTo3DView(TheView,this.GetTileExtentAsGeographic(),TheLayer);
				
				if (Distance>ThisStepTileRefWidth*3) PaintTile=true;
			}
			else
			{
				var ThisStepsTilePixelWidth=TheView.GetPixelWidthFromRefWidth(ThisStepTileRefWidth);
	
				if (ThisStepsTilePixelWidth<=256) PaintTile=true; // resolution is low, paint this tile
				else if (ThisStepsTilePixelWidth<=512)
				{
					PaintChildren=true;
				}
			}
			if ((PaintTile==false))
			{
				// If there are no children, paint this tile
				if (this.ZoomLevel==0)  
				{
					PaintTile=true; // at the bottom, render this tile
				}
				// If we're not going to paint this tile, check to see if we should paint, or load, the child tiles
				else // see if we have 4 child tiles ready to be painted (or are empty)
				{
					//if (this.TheData.NumChildTiles>0) // this tile does have child tiles
					{
						NumLoadedChildTiles=0;
						
						// allocate the child tiles if they are not already there.
						if (this.ChildTiles==null) // initialize the child tiles array
						{
							this.ChildTiles=[];
							this.ChildTiles[0]=[null,null];
							this.ChildTiles[1]=[null,null];
						}
						
						// go through the 4 child tiles
						for (var RowIndex=0;RowIndex<2;RowIndex++)
						{
							for (var ColumnIndex=0;ColumnIndex<2;ColumnIndex++)
							{
								// initialize the child tile if needed
								if (this.ChildTiles[RowIndex][ColumnIndex]==null)
								{
									var ChildZoomLevel=this.ZoomLevel+1;
									var ChildColumn=(this.GlobalColumn*2)+ColumnIndex;
									var ChildRow=(this.GlobalRow*2)+(1-RowIndex);
									
									// create the tile, loads are initiated on paint
									var TheChildTile=new CMTileOpenFormat(this.TheDataset,ChildZoomLevel,ChildColumn,ChildRow);
									
									if (this.PaintTileInfo) TheChildTile.SetPaintTileInfo(this.PaintTileInfo);
									
									this.ChildTiles[RowIndex][ColumnIndex]=TheChildTile;
								}
								// if we are painting children, make sure the images are loaded
								if (PaintChildren) // have an existing child tile
								{
									var TheChildTile=this.ChildTiles[RowIndex][ColumnIndex];
	
									if (TheChildTile.TheRasterRequest==null)
									{
										TheChildTile.LoadTileRaster();
									}
									else 
									{
										if (TheChildTile.TheRasterRequest.LoadStatus==CMDataset.LOAD_STATUS_NONE) 
										{
											// tile child tile has not been loaded, load it now for the next repaint
											TheChildTile.LoadTileRaster();
											// have to paint this tile
											//PaintTile=true;
										}
										else if (TheChildTile.TheRasterRequest.LoadStatus==CMDataset.LOAD_STATUS_LOADING)
										{
											// cihld tile has not finished loading, have to paint this tile
											//PaintTile=true;
										}
										else 
										{
											if (TheChildTile.TheRasterRequest.LoadStatus==CMDataset.LOAD_STATUS_LOADED)
											{
												NumLoadedChildTiles++;
											}
											else if (TheChildTile.TheRasterRequest.LoadStatus==CMDataset.LOAD_STATUS_CANCELED)
											{
												this.LoadTileRaster(); // try again
											}
										}
									}
								}
							}
						}
						// unless this tile is not completely covered by child tiles, paint it	
						if (NumLoadedChildTiles<4) 
						{
							//PaintTile=true;
						}
					}
					//PaintTile=false;
				}
			}
		}
	}
	return(PaintTile);
}

/*
* Sets up a grid-based goemetry to hold the coordinates for the projected space
* @protected
*/
CMTileOpenFormat.prototype.SetupGeometry=function(TheLayer) 
{
	var TheGeo=TheLayer.GetParent(CM3Geo);
	
	var Extent=this.GetTileExtent();
	
	var TheArray=[];
	
	var XMin=Extent.XMin;
	var XMax=Extent.XMax;
	var YMin=Extent.YMin;
	var YMax=Extent.YMax;
	
	var RowRefHeight=(YMax-YMin)/16;
	var ColumnRefWidth=(XMax-XMin)/16;
	
	var TheGoogleMapsProjector=new CMProjectorGoogleMaps();
	TheGoogleMapsProjector.SetZoomLevel(18);
	
	for (var Row=0;Row<=16;Row++)
	{
		var TheRow=[];
		TheArray[Row]=TheRow;
		
		for (var Column=0;Column<=16;Column++)
		{
			TheRow[Column]=[];
			
			// convert tile coordinates to geographic
			var Northing=YMin+(Row*RowRefHeight);
			var Easting=XMin+(Column*ColumnRefWidth);
			
			var Result=TheGoogleMapsProjector.ProjectToGeographic(Easting,Northing);
			
			var Lon=Result[0];
			var Lat=Result[1];
			
			// convert geographic coordinates to spherical
			var Elevation=0;//TheElevationGroup.Offset;
			
			// project the coordinate
			var Result=TheGeo.ProjectFromGeographic(Lon, Lat, Elevation);
			
			// setup the grid to contain the projected coordinate
			TheRow[Column]=Result;
			//TheRow[Column][1]=Vector1.y;
			//TheRow[Column][2]=Vector1.z;
		}
	}
	
	// setup the THREEE geomery which will become part of the mesh
	var NumColumns=TheArray[0].length;
	var NumRows=TheArray.length;
	
	var TheGeometry=new CM3SurfaceGeometry(NumColumns-1,NumRows-1);
	
	// set the positions within the grid
	TheGeometry.SetPositions(TheArray);
				
	return(TheGeometry);
}
//***********************************************************************
// Protected CMTileOpenFormat functions
//***********************************************************************

/**
* Paint a single tile.  Adds appropriate content to the Geo based on the content type.
* @protected
*/
CMTileOpenFormat.prototype.PaintTile=function(TheLayer,TheView,SelectedOnly) 
{
	var NumPainted=0; // 1 if this tile painted it's contents, 0 otherwise
		
	var Extent=this.GetTileExtent();
	
	var TileRefWidth=Extent.XMax-Extent.XMin;
	
	var RefX=Extent.XMin; 
	var RefY=Extent.YMax;
	
	var PaintTile=true; // true to paint this tile, false to paint children
	
	//***********************************************************
	// 3D painting
	if ((typeof(CM3View)!="undefined")&&(TheView instanceof CM3View)) // have a 3d view
	{
		if ((TheLayer.GetVisible()))
		{
			// jjg - cheating a bit here and finding the disance from one corner of the tile
			// to the other corner (should do this in the projector where arc disances can be computed)
			
			var Result1=CMTileOpenFormat.GetGlobalCoordinate(TheLayer,Extent.XMin,Extent.YMin,0);  // south west to 
			var Result2=CMTileOpenFormat.GetGlobalCoordinate(TheLayer,Extent.XMax,Extent.YMax,0); // north east
			
			var DX=Result1[0]-Result2[0];
			var DY=Result1[1]-Result2[1];
			var DZ=Result1[2]-Result2[2];
			
			var TileRefWidth2=Math.sqrt(DX*DX+DY*DY+DZ*DZ)*2/2;
	
			// check if we are going to paint this tile or go after child tiles
			PaintTile=this.CheckPaintTile(TheView,TileRefWidth2,TheLayer);
			
			if (PaintTile) // painting this tile
			{
				if (this.ChildTiles!=null) // if we have child tiles, hide item
				{
					for (var RowIndex=0;RowIndex<2;RowIndex++)
					{
						if (this.ChildTiles[RowIndex]!=null)
						{
							for (var ColumnIndex=0;ColumnIndex<2;ColumnIndex++)
							{
								if (this.ChildTiles[RowIndex][ColumnIndex]!=null)
								{
									var TheChildTile=this.ChildTiles[RowIndex][ColumnIndex];
									
									TheChildTile.HideMesh(TheLayer);
								}
							}
						}
					}
				}
				//*************************************************************************
				// paint a raster from a raster file (e.g. PNG or JPG)
				if (this.MeshIsInGeo==false) // the mesh is not in the Geo, we need to add it
				{
					if (this.TheMesh==null) // the mesh is not available, we need to create it
					{
						// try to get the DEM info
						var TheGeometry=this.SetupGeometry(TheLayer);
						
						// setup the material
						var TheFileName=this.GetTileImageFilePath();
						
						// setup the material
						var SolidColorMaterial=CMTile.GetNewMaterial(TheFileName);
					
						// setup the mesh
						this.TheMesh=new THREE.Mesh(TheGeometry,SolidColorMaterial);
					}
					// add the mesh to the geo
					
					if (this.TheMesh!=null)
					{ 
						var TheGeo=TheLayer.GetParent(CMGeo);
						TheGeo.AddOGLObjectToGeo(this.TheMesh);
						
//						var helper = new THREE.VertexNormalsHelper( this.TheMesh, 20, 0x00ff00, 10 );
//						TheGeo.AddOGLObjectToGeo(helper);
						
						this.MeshIsInGeo=true;
					}
				}
			}
			if (PaintTile==false) // remove this tile from the Geo, if needed, and paint the children
			{
				this.HideMesh(TheLayer);
			}
		}
	}
	//**************************************************************
	// 2D painting
	else if(CMUtilities.BoundsOverlap(Extent,TheView.GetBounds()))
	{
		//if ((this.TheData!=null)) // data will be null until recieved
		{
			// paint the child
			
			PaintTile=this.CheckPaintTile(TheView,TileRefWidth);
			
			var NumChildrenPainted=0;
			if (PaintTile) // paint this tile instead of the children
			{
				if (this.TheRasterRequest!=null)
				{
					if (this.TheRasterRequest.LoadStatus==CMDataset.LOAD_STATUS_LOADED)
					{
						var ImageRefWidth=TileRefWidth;
						var ImageRefHeight=-TileRefWidth;
						
						var ImageRefX=RefX;
						var ImageRefY=RefY;
						
						TheView.PaintRefImageScaled(this.TheRaster,ImageRefX,ImageRefY,ImageRefWidth,ImageRefHeight);
					}
					else if (this.TheRasterRequest.LoadStatus==CMDataset.LOAD_STATUS_CANCELED)
					{
						this.LoadTileRaster();
					}
				}
				else
				{
					this.LoadTileRaster();
				}
			}
		}
		//*************************************************************************
		// paint the debugging information (2d)
		if (PaintTile) // 
		{
			if ((this.PaintTileInfo))
			{
				var TheStyle2={ 
					"font":"20px Arial",
					"fillStyle":"red",
					"lineWidth":1,
					"strokeColor": "red",
					"strokeStyle":"#000",
				};
				if (TheStyle2!=null) TheView.SetStyle(TheStyle2);
					
				var CenterX=RefX+TileRefWidth/2;
				var CenterY=RefY-TileRefWidth/2;
				
				var TheText=this.ZoomLevel+"_"+this.GlobalColumn+"_"+this.GlobalRow;
				
				TheView.PaintRefText(TheText,CenterX,CenterY,20);
				
				TheView.PaintRefLine(RefX,RefY,RefX,RefY-TileRefWidth); // left
				TheView.PaintRefLine(RefX+TileRefWidth,RefY,RefX+TileRefWidth,RefY-TileRefWidth); // right
				
				TheView.PaintRefLine(RefX,RefY,RefX+TileRefWidth,RefY); // top 
				TheView.PaintRefLine(RefX,RefY-TileRefWidth,RefX+TileRefWidth,RefY-TileRefWidth); // bottom
				
				if (TheStyle2!=null) TheView.RestoreStyle();
			}
		}
	}
	//**************************************************************************
	// if nothing was painted above and we have data, try to paint the children
	if (PaintTile==false) // paint the child tiles in this tile
	{
		//if (this.TheData.NumChildTiles>0)
		{
			for (var RowIndex=0;RowIndex<2;RowIndex++)
			{
				for (var ColumnIndex=0;ColumnIndex<2;ColumnIndex++)
				{
					// negative for one feature ID, >0 for mulitple and must load a child tile
					
					if (this.ChildTiles[RowIndex][ColumnIndex]!=null)
					{
						var TheChildTile=this.ChildTiles[RowIndex][ColumnIndex];
						
						// the tiles in the next step are 1/2 the reference width of the tiles in this step
						
						var NextStepTileRefWidth=TileRefWidth/2;
						
						//if (TheChildTile.GlobalRow==0)
						{
							NumChildrenPainted+=TheChildTile.PaintTile(TheLayer,TheView,SelectedOnly,NextStepTileRefWidth);
						}
					}
				}
			}
		}
	}
	NumPainted++;
	
	// if the children of this tile painted, we do not need to paint
	
	if (NumChildrenPainted<4)
	{
		var Test=12;
	}
	
	return(NumPainted);
}
//***********************************************************************
// Public CMTileOpenFormat functions
//***********************************************************************
/**
* Sets whether the tile will paint it's debugging information.
* The parent dataset has a setting for this.
* @public
*/
CMTileOpenFormat.prototype.SetPaintTileInfo=function(NewPaintTileInfo) 
{
	this.PaintTileInfo=NewPaintTileInfo;
}
CMTileOpenFormat.prototype.GetData=function() 
{
	throw("Sorry, CMTileOpenFormat does not support GetData()");
}
