/**
* CMView Class
*
* This class manages the canvas and allows the map to be zoomed and panned.
* It also provides commont painting functions such as rectangles, circles,
* text, and rasters.  The functions are available with locations in pixel
* or reference (map) coordinate values.
*
* The view is also an item that can appear in other views.  This is important
* for authors working with 3D scenes to be able to see where their views are.
*
* @module CMView
* @Copyright HSU, Jim Graham, 2019
*/

/**
* Tool definitions
* @enum
*/
CMView.TOOL_NONE=0; // no tool (disables moving the background)
CMView.TOOL_HAND=1; // pan the map (the hand)
CMView.TOOL_INFO=2; // click to get information on features ("I" tool)
CMView.TOOL_EDIT=3; // user is editing existing spatial data  (not used)
CMView.TOOL_ADD=4; // user is adding data (not used)
CMView.TOOL_SELECT=5; // combined with INFO on the web (arrow tool)

/**
* Message definitions
*/
CMView.MESSAGE_MOUSE_MOVE=CMBase.GetUniqueNumber(); // AdditionalInfo=TheEvent
CMView.MESSAGE_MOUSE_DOWN=CMBase.GetUniqueNumber(); // AdditionalInfo=TheEvent
CMView.MESSAGE_MOUSE_UP=CMBase.GetUniqueNumber(); // AdditionalInfo=TheEvent
CMView.MESSAGE_MOUSE_WHEEL=CMBase.GetUniqueNumber(); // AdditionalInfo=TheEvent
CMView.MESSAGE_KEY_DOWN=CMBase.GetUniqueNumber(); // AdditionalInfo=TheEvent
CMView.MESSAGE_KEY_UP=CMBase.GetUniqueNumber(); // AdditionalInfo=TheEvent

//******************************************************************
// Constructors
//******************************************************************
CMView.UniqueNumber=0;
/**
* @public, @constructs
*/
function CMView() 
{
	CMBase.call(this);
	this.UniqueNumber=CMView.UniqueNumber;
	CMView.UniqueNumber++;

	// CMBase does not have settings so this is the main property for settings
	this.Settings={};
	
	// additional settings (cached)
	
	this.CurrentTool=CMView.TOOL_SELECT;

	this.CollisionChecking=true;
	
	// other properties
	this.MapElements=[]; // should either be in the view or in a layer
	
	this.TheCanvasElement=null; // the canvas that contains this view
	
	this.ToolHandler=null;
	
	// Get the drawing context from our <canvas> 
	
	this.TheContext=null; // just used to access context quicker for drawing
	
	// objects that have been drawn into the view thus far for collision detection
	
	this.CollisionArray=null;
}
CMView.prototype=Object.create(CMBase.prototype); // inherit prototype functions from CMItem

CMView.prototype.contructor=CMView; // override the constructor to go to ours

//******************************************************************
// Definitions
//******************************************************************
/**
* Below are the settings definitions.
* @public, @settings
*/
CMView.SettingDefintions=
{
	View:
	{
		CollisionChecking: { Name:"Collision Checking",Type:CMBase.DATA_TYPE_BOOLEAN, Default:true }
		
		// these are similar to those in CMItem (jjg - in the future, CMView will inherit from CMItem)
	},
	Item:
	{
		Name: { Name:"Name",Type:CMBase.DATA_TYPE_STRING, Default:"" }, // Name that appears in the item and layer lists
		
		Status: { Name:"Status",Type:CMBase.DATA_TYPE_ENUMERATED, Options:[CMItem.STATUS_HIDDEN,CMItem.STATUS_VISIBLE,CMItem.STATUS_SELECTABLE,CMItem.STATUS_EDITABLE],Default:CMItem.STATUS_SELECTABLE },// How the item appears to the user and how the user can interact with it

		PositionOffset: { Name:"Position Offset Vector",Type:CMBase.DATA_TYPE_VECTOR, Default:null }, // x,y, and z values to offset the item
		Rotation: { Name:"Rotation Vector",Type:CMBase.DATA_TYPE_VECTOR, Default:null }, // pitch, roll, and yaw values (x,y,z) values to roatate the item in degrees
	},
};

//******************************************************************
// CMBase Functions
//******************************************************************

CMView.prototype.GetSettingsDefinitions=function() 
{
	return(CMView.SettingDefintions); 
}

CMView.prototype.CMBase_SetSettings=CMBase.prototype.SetSettings;

CMView.prototype.SetSettings=function(NewSettings,TimeSlice) 
{
	if (NewSettings["View"]!=undefined)
	{
		if (NewSettings["View"]["CollisionChecking"]!=undefined)
		{
			this.CollisionChecking=NewSettings["View"]["CollisionChecking"];
		}
	}
	this.CMBase_SetSettings(NewSettings,TimeSlice);
}
//******************************************************************
// Protected CMView Functions
//******************************************************************
/**
* Sets a style for painting.  The existing style is saved.
* New style is an array of key value pairs (i.e.  objects).  This
* may be called on each item so it must run as fast as possible.
* @protected
* @param NewStyle the style to use for painting
* @param SaveCurrentStyle - true to have the current style saved
*/
CMView.SetStyleToContext=function(NewStyle,SaveCurrentStyle,TheContext,TheView) 
{
	// this has to be saved each time so the restores are synchronized
	if (SaveCurrentStyle!==false) TheContext.save();
	
	if ((NewStyle!=null)&&(NewStyle!=undefined))
	{
		TheContext.shadowBlur=0;
		
		for (var key in NewStyle)
		{
			if (key=="PatternImage")
			{
				var ThePattern = TheContext.createPattern(NewStyle[key], "repeat");
				TheContext["fillStyle"]=ThePattern;
			}
			else if (key=="GradientType")
			{
				var TheColors=NewStyle["GradientColors"];
				var TheCoordinates=NewStyle["GradientCoordinates"];
				var GradientRadius1=NewStyle["GradientRadius1"];
				var GradientRadius2=NewStyle["GradientRadius2"];
				
				if (GradientRadius1==undefined) GradientRadius1=0;
				if (GradientRadius2==undefined) GradientRadius2=100;
				
				var TheGradient=null;
				
				var X0=0;
				var Y0=0;
				var X1=500;
				var Y1=500;
				
				if (TheCoordinates!=undefined)
				{
					X0=TheCoordinates.Xs[0];
					Y0=TheCoordinates.Ys[0];
					X1=TheCoordinates.Xs[1];
					Y1=TheCoordinates.Ys[1];
				}
				
				if (NewStyle[key]=="Linear")
				{
					TheGradient=TheContext.createLinearGradient(X0,Y0,X1,Y1);
				}
				else
				{
					TheGradient=TheContext.createRadialGradient(X0,Y0,GradientRadius1,X1,Y1,GradientRadius2);
				}
				
				if (TheColors!=undefined)
				{
					for (var i=0;i<TheColors.length;i++)
					{
						TheGradient.addColorStop(i,TheColors[i]);
					}
				}
				TheContext["fillStyle"]=TheGradient;
			}
			else  if ((key=="GradientColors")||(key=="GradientCoordinates")||(key=="GradientRadius")) // ignore these as they are handled above
			{
			}
			else  if ((key=="font")) // have to get the font information and conver tthe size to pixels if needed
			{
				var Font=NewStyle[key];
				
				var Tokens=Font.split(" ");
				
				var FontSize="12px";
				
				// find the font size as it is the only one that starts with a digit
				var SizeIndex=-1;
				for (var i=0;i<Tokens.length;i++)
				{
					var Token=Tokens[i];
					
					// if the first charater is a number, must be the font size entry
					if (Token[i] >= '0' && Token[i]<='9' ) SizeIndex=i;
				}
				// convert the font size to pixels if needed
				if (SizeIndex!=-1)
				{
					FontSize=Tokens[SizeIndex];
					
					var Index=FontSize.indexOf("px");
					
					if (Index==-1) // "px" not specified, convert to reference units
					{
						FontSize=parseFloat(FontSize); //.substring(0,Index); // get just the size
						FontSize=-TheView.GetPixelHeightFromRefHeight(parseFloat(FontSize)); // convert the size to pixel units
						
						FontSize=FontSize+"px"; // convert the FontSize back to a string
					}
				}
				Font="";
				for (var i=0;i<Tokens.length;i++)
				{
					if (i!=0) Font+=" ";
					
					if (i==SizeIndex) Font+=FontSize;
					else Font+=Tokens[i];
				}
				TheContext[key]=Font;
			}
			else
			{
				TheContext[key]=NewStyle[key];
			}
		}
	}
}

//******************************************************************
// Public CMView Functions
//******************************************************************
CMView.prototype.Repaint=function() 
{
	var Result=this.GetParent(CMScene);
	if (Result!=null) Result.Repaint();
}

/**
* Returns the context for the CanvasMap's canvas.  This is not used
* often but is required do things like great gradients.
* @public
*/
CMView.prototype.GetContext=function()
{
	return(this.TheContext);
}


//*******************************************************************
// CMView functions to handle events.
// These functions should be overriden by subclasses
//*******************************************************************

/**
* Handles a mouse down event in the view.  Can be overriden to place items in the view
* and have the user interact with them.
* @protected
* @override
* @returns Flag - returns True if the function "used" the event so other stuff does not use it.
*/
CMView.prototype.MouseDown=function(TheEvent)
{
	var Used=false;
		
	CMMainContainer.HidePopupWindows();
	
	return(Used);
}
/**
* Handles a mouse move event in the view.  Can be overriden to place items in the view
* and have the user interact with them.
* @protected
* @override
* @returns Flag - returns True if the function "used" the event so other stuff does not use it.
*/
CMView.prototype.MouseMove=function(TheEvent)
{
	var Used=false;
	
	if (!TheEvent) { TheEvent=window.event; }

	this.SendMessageToListeners(CMView.MESSAGE_MOUSE_MOVE,TheEvent);
	
	return(Used);
}
/**
* Handles a mouse move up in the view.  Can be overriden to place items in the view
* and have the user interact with them.
* @protected
* @override
* @returns Flag - returns True if the function "used" the event so other stuff does not use it.
*/
CMView.prototype.MouseUp=function(TheEvent)
{
	var Used=false;
		
	if (!TheEvent) { TheEvent=window.event; }
	
	var Coordinate=CMUtilities.GetElementCoordinate(TheEvent.clientX,TheEvent.clientY,this.TheCanvasElement);
	var PixelX=Coordinate.x;
	var PixelY=Coordinate.y;
	
	return(Used);

}
/**
* Handles a mouse wheel event.  Can be overriden to change the action taken when the 
* user moves the wheel.
* @protected
* @override
* @returns Flag - returns True if the function "used" the event so other stuff does not use it.
*/
CMView.prototype.MouseWheel=function(TheEvent)
{
	var Used=false;
	
	CMMainContainer.HidePopupWindows();
	
	if (!TheEvent) { TheEvent=window.event; }
	
	var Delta=TheEvent.detail? TheEvent.detail*(-120) : TheEvent.wheelDelta //delta returns +120 when wheel is scrolled up, -120 when scrolled down
	
	// prevent the wheele from scrolling the page
	
	if (TheEvent.preventDefault)  TheEvent.preventDefault()
	
	return(Used);
}
CMView.prototype.KeyDown=function(TheEvent)
{
	var Used=false;
	
//	CMMainContainer.HidePopupWindows();
	
	return(Used);
}
CMView.prototype.KeyUp=function(TheEvent)
{
	var Used=false;
	
//	CMMainContainer.HidePopupWindows();
	
	return(Used);
}


//**********************************************************
// Additional functions to be overriden by subclasses
//**********************************************************
/**
* @protected, @override
*/
CMView.prototype.GetCoordinateStringFromEvent=function(TheEvent,CoordinateUnits)
{
	
	return(Text);
}
//******************************************************************
// Map Elements
//******************************************************************
/**
* Add a new map element to the scene (e.g. a scale bar)
* @public
* @param TheElement - 
*/
CMView.prototype.AddMapElement=function(TheElement) 
{
	this.MapElements.push(TheElement);
	TheElement.SetParent(this);
}


//******************************************************************
// Protected  Functions
//******************************************************************
/**
* Used by the canvas map to give this view the ability to call its canvas map
* @protected
*/
CMView.prototype.Setup=function(TheCanvasElement)
{
	//this.TheCanvasContainer=TheCanvasContainer;
	this.TheCanvasElement=TheCanvasElement;
	
	// Get the drawing context from our <canvas> 
	
	this.TheContext=this.TheCanvasElement.getContext('2d');
	
	this.TheCanvasElement.style.cursor="crosshair";
	this.SetTool(CMView.TOOL_SELECT);
}
/*
* 
* @protected
*/
CMView.prototype.Resize=function()
{
	var TheElement=this.TheCanvasElement;

	// the canvas aspect ratio is not correct unless we set the "width" and "height" of the element
	// rather than the style.  This must be done or it will display maps distorted
	
	var TheParent=TheElement.parentNode;
	var ParentWidth=jQuery(TheParent).width();
	var ParentHeight=jQuery(TheParent).height();

	TheElement.width=ParentWidth;
	TheElement.height=ParentHeight;
	
	// resize the map elements
	for (var i=0;i<this.MapElements.length;i++)
	{
		var TheMapElement=this.MapElements[i];
	}
}

CMView.prototype.SetToolHandler=function(NewToolHandler)
{
	this.ToolHandler=NewToolHandler;
}
CMView.prototype.GetToolHandler=function()
{
	return(this.ToolHandler);
}
CMView.prototype.GetCanvasElement=function()
{
	return(this.TheCanvasElement);
}

//*******************************************************************
// Mouse events
//*******************************************************************
/**
* Adds the standard event handlers to the view
*/
CMView.prototype.AddMouseEventHandlers=function()
{
	var TheCanvasElement=this.GetCanvasElement();
	
	TheCanvasElement.TheView=this; // required by the mouse functions
	
	TheCanvasElement.addEventListener("mousedown",function(TheEvent) 
	{
		if (this.TheView!=null)
		{
			this.TheView.MouseDown(TheEvent);	
			TheEvent.stopPropagation(); // stop the document from hidding a popup window
		}
	});
	TheCanvasElement.addEventListener("mousemove",function(TheEvent) 
	{
		if (this.TheView!=null)
		{
			this.TheView.MouseMove(TheEvent);
		}
	});
	TheCanvasElement.addEventListener("mouseup",function(TheEvent) 
	{
		if (this.TheView!=null)
		{
			this.TheView.MouseUp(TheEvent);	
		}
	});
	//***************************************************************************************
	// jQuery does not yet support the mouse wheel so we have to do it the old way
	
	var mousewheelevt=(/Firefox/i.test(navigator.userAgent))? "DOMMouseScroll" : "mousewheel" //FF doesn't recognize mousewheel as of FF3.x
	 
	if (TheCanvasElement.attachEvent) //if IE (and Opera depending on user setting)
	{
		TheCanvasElement.attachEvent("on"+mousewheelevt, function(TheEvent)
		{
			var Result; // return result is undefined typically
			
			CMMainContainer.HidePopupWindows(); // static function
			
			var TheEvent=window.event || TheEvent // grab the event passed in or globally (IE compatibility)
			
			var Result=this.TheView.MouseWheel(TheEvent);
				
			return(Result);
		});
	}
	else if (TheCanvasElement.addEventListener) //WC3 browsers
	{
		TheCanvasElement.addEventListener(mousewheelevt, function(TheEvent)
		{
			var Result; // return result is undefined typically
			
			CMMainContainer.HidePopupWindows(); // static function
			
			var TheEvent=window.event || TheEvent // grab the event passed in or globally (IE compatibility)
			
			var Result=this.TheView.MouseWheel(TheEvent);
				
			return(Result);
		},
		false);
	}
}
CMView.prototype.AddKeyEventHandlers=function()
{
	var TheCanvasElement=this.GetCanvasElement();
	
	window.TheView=this; // jjg - there must be  a better way!
	
	window.addEventListener("keydown",function(TheEvent) 
	{
		if (this.TheView!=null)
		{
			this.TheView.KeyDown(TheEvent);	
		}
	}, false );
	
	window.addEventListener("keyup",function(TheEvent) 
	{
		if (this.TheView!=null)
		{
			this.TheView.KeyUp(TheEvent);	
		}
	}, false );
}

/**
* Adds the mobile event handlers based on the Hammer library
*/
CMView.prototype.AddMobileEvents=function(CanvasContainer)
{
	// create a simple instance
	// by default, it only adds horizontal recognizers
/*	var mc = new Hammer(CanvasContainer);
	
	// let the pan gesture support all directions.
	// this will block the vertical scrolling on a touch-device while on the element
	mc.get('pan').set({ direction: Hammer.DIRECTION_ALL });
	var pinch = new Hammer.Pinch();
	
	mc.add([pinch]);
	
	mc.TheView=this;
	
	// listen to events...
	mc.on("pinch panleft panright panup pandown", function(ev)  //  tap press
	{
		var TheView=mc.TheView;
		
		// needs to call function that is overriden by subclasses
		var RefDistance=TheView.GetRefWidthFromPixelWidth(CMMainContainer.GESTURE_PAN);
		var RefCenter=TheView.GetRefCenter();
		var RefX=RefCenter.RefX;
		var RefY=RefCenter.RefY;

		MapHeader.innerHTML=ev.type;
		
		if (ev.type=="panup")
		{
			RefY-=RefDistance;
			TheView.SetRefCenter(RefX,RefY); // set the center of the map
		}
		else if (ev.type=="pandown")
		{
			RefY+=RefDistance;
			TheView.SetRefCenter(RefX,RefY); // set the center of the map
		}
		else if (ev.type=="panleft")
		{
			RefX+=RefDistance;
			TheView.SetRefCenter(RefX,RefY); // set the center of the map
		}
		else if (ev.type=="panright")
		{
			RefX-=RefDistance;
			TheView.SetRefCenter(RefX,RefY); // set the center of the map
		}
		else if (ev.type=="pinch")
		{
			var ZoomLevel=TheView.GetZoomLevel();
			
			MapHeader.textContent = ev.additionalEvent ;
			if (ev.additionalEvent=="pinchin")
			{
				TheView.ZoomTo(ZoomLevel-CMMainContainer.GESTURE_ZOOM);
			}
			else
			{
				TheView.ZoomTo(ZoomLevel+CMMainContainer.GESTURE_ZOOM);
			}
		}
	});	 
*/}
//******************************************************************
// Functions to set the current tool
//******************************************************************

/**
* Change the current tool that will be used when the user clicks with 
* the mouse 
* @public
* @param NewTool from the enums: CMView.TOOL_HAND,...
*/
CMView.prototype.SetTool=function(NewTool)
{
	 switch(NewTool)
	 {
	case CMView.TOOL_HAND:
		this.TheCanvasElement.style.cursor="move";
		break;
	case CMView.TOOL_INFO:
	case CMView.TOOL_SELECT:
		this.TheCanvasElement.style.cursor="crosshair";
		break;
	case CMView.TOOL_EDIT:
		this.TheCanvasElement.style.cursor="crosshair";
		break;
	 }

	this.CurrentTool=NewTool;
}

/**
* Returns the current tool
* @public
* @returns NewTool from the enums: CMView.TOOL_HAND,...
*/
CMView.prototype.GetTool=function() { return(this.CurrentTool); }

//*******************************************************************
// CMView functions to manage painting.  Specifically styles,
// get information on fonts, and manage collisions.
//*******************************************************************
CMView.prototype.SetStyle=function(NewStyle,SaveCurrentStyle) 
{
	console.log("CMView.SetStyle() needs to be overriden");
}
/**
* Saves the current style to a stack
* @public
*/
CMView.prototype.SaveStyle=function() 
{
	this.TheContext.save();
}
/**
* Restores the style from the stack
* @public
*/
CMView.prototype.RestoreStyle=function() 
{
	this.TheContext.restore();
}
/**
* Starts painting.  Only called by the Scene
* @protected
*/
CMView.prototype.PaintStart=function() 
{
	if (this.GetSetting("View","CollisionChecking")) // jjg - might want to use a local variable for speed later
	{
		this.CollisionArray=[];
	}
}
/**
* Paints the contents of the view.  Called by the Scene which is called by the
* CanvasMap object.
* @public
*/
CMView.prototype.Paint=function() 
{
	this.TheContext.clearRect(0,0,this.TheCanvasElement.width,this.TheCanvasElement.height);
	this.GetParent(CMScene).Paint(this);
};
/**
* Ends painting.  Only called by the Scene
* @protected
*/
CMView.prototype.PaintEnd=function() 
{
	// paint the additional map elemetns (north arrows, scale bars, etc).
	
	for (var i=0;i<this.MapElements.length;i++)
	{
		this.MapElements[i].Paint(this);
	}

	this.ResetCollisions();
}
/**
* Checks if there was a collision between the specific bounds and a point feature
* that has already been painted.  Typically only called by a layer.
* @protected
*/
CMView.prototype.CheckCollision=function(Bounds) 
{
	var Result=false;
	
	if (this.CollisionArray!==null) // CollisionChecking==true
	{
		for (var i=0;(i<this.CollisionArray.length)&&(Result==false);i++)
		{
			if (CMUtilities.BoundsOverlap(Bounds,this.CollisionArray[i]))
			{
				Result=true;
			}											   
		}
	}
	return(Result);
}
/**
* Checks if there was a collision between the specific bounds and a point feature
* that has already been painted.  Typically only called by a layer.
* @protected
*/
CMView.prototype.AddToCollisions=function(Bounds) 
{
	if (this.CollisionArray!==null) // CollisionChecking==true
	{
		this.CollisionArray.push(Bounds);
	}

}
/**
* Collisions are reset at the end of painting.  This function will reset the collision
* array during painting if desired
*/
CMView.prototype.ResetCollisions=function() 
{
	if (this.GetSetting("View","CollisionChecking")) // jjg - might want to use a local variable for speed later
	{
		this.CollisionArray=[];
	}
}
/**
* Gets the width of the specified text in pixels based on the current font settings
*/
CMView.prototype.GetTextWidthInPixels=function(Text)
{
	var TextWidth=this.TheContext.measureText(Text).width;
	return(TextWidth);
}

//*******************************************************************
// CMView functions to paint simple graphic elements with pixel coordinates
//*******************************************************************

/**
* Paints an image at the specified pixel location.
* CanvasMap object.
* @public
* @param TheImage
* @param PixelX
* @param PixelY
*/
CMView.prototype.PaintImage=function(TheImage,PixelX,PixelY)
{
	this.TheContext.drawImage(TheImage,PixelX,PixelY);
};
/**
* Paints the background of the canvas element.
* @public
*/
CMView.prototype.PaintBackground=function()
{
//	if (Color!=null) 
	{
		this.TheContext.fillRect(0,0,this.TheCanvasElement.width,this.TheCanvasElement.height);
	}
};
/**
* Paints a circle using pixel coordinate values
* @public
* @param X - horizontal pixel center of the circle
* @param Y - vertical pixel center of the circle
* @param RadiusInPixels - 
*/
CMView.prototype.PaintCircle=function(X,Y,RadiusInPixels)
{
	this.TheContext.beginPath();
	
	this.TheContext.arc(X,Y,RadiusInPixels,0,2*Math.PI);
	
	if (this.TheContext.strokeStyle!=null) this.TheContext.stroke();
	if (this.TheContext.fillStyle!=null) this.TheContext.fill();
};
/*
* Paints a rectangle based on the specified bounds
* @public
* @param PixelXMin - Left side of the rectangle
* @param PixelXMax - Right side of the rectangle
* @param PixelYMin - Top of the rectangle
* @param PixelYMax - Bottom of the rectangle
*/
CMView.prototype.PaintRect=function(PixelXMin,PixelXMax,PixelYMin,PixelYMax)
{
	// Fill the path we just finished drawing with color
	if (this.TheContext.fillStyle!=null) 
	{
		this.TheContext.fillRect(PixelXMin,PixelYMin,PixelXMax-PixelXMin,PixelYMax-PixelYMin);
	}
	if (this.TheContext.strokeStyle!=null) 
	{
		if ((this.TheContext.fillStyle==null)||(this.TheContext.shadowColor==undefined)) // no fill or no shadow
		{
			this.TheContext.stroke();
		}
		else // filled and has a shadow that we must disable
		{
			var TheShadowColor=this.TheContext.shadowColor;
			this.TheContext.shadowColor="rgba(0,0,0,0)";
			
			this.TheContext.strokeRect(PixelXMin,PixelYMin,PixelXMax-PixelXMin,PixelYMax-PixelYMin);
			
			this.TheContext.shadowColor=TheShadowColor;
		}
	}
}
/*
* Paints an arc based on the specified bounds.  The arc will be drawn as if it was
* drawing an oval within the bounds but only the area of the oval defined by the
* start and end angles will be painted.  This can be used to paint circles (or ovals)
* by making the start angle 0 and end angle 2*Math.PI.
*
* @public
* @param PixelXMin - Left side of the rectangle
* @param PixelXMax - Right side of the rectangle
* @param PixelYMin - Top of the rectangle
* @param PixelYMax - Bottom of the rectangle
* @param StartAngle - Start angle for the start of the arc, in radians
* @param EndAngle - End angle for the start of the arc, in radians
*/

CMView.prototype.PaintArc=function(PixelXMin,PixelXMax,PixelYMin,PixelYMax,StartAngle,EndAngle)
{
	var PixelCenterX=(PixelXMin+PixelXMax)/2;
	var PixelCenterY=(PixelYMin+PixelYMax)/2;
	
	var Radius=(PixelXMax-PixelXMin)/2;
	
	this.TheContext.save(); // save the trnasformation matrix
	
	this.TheContext.translate(PixelCenterX,PixelCenterY); // transate to center of arc's oval
	
	// scale the arc vertically with the width as the radius and the height as a multiple of the radius
	var HeightFactor=(PixelYMax-PixelYMin)/(PixelXMax-PixelXMin);
	
	this.TheContext.scale(1, HeightFactor);
	
	// draw the arc
	this.TheContext.beginPath();
	this.TheContext.arc(0, 0, Radius,StartAngle,EndAngle, false);
	
	this.TheContext.restore(); // put the original translation matrix back
	
	if (this.TheContext.strokeStyle!=null) 
	{
		this.TheContext.stroke();
	}
	if (this.TheContext.fillStyle!=null) 
	{
		this.TheContext.fill();
	}
}

/*
* Function to paint a rounded rectangle
* Adapted from: http://stackoverflow.com/questions/1255512/how-to-draw-a-rounded-rectangle-on-html-canvas
* @public
* @param PixelXMin - Left side of the rectangle
* @param PixelXMax - Right side of the rectangle
* @param PixelYMin - Top of the rectangle
* @param PixelYMax - Bottom of the rectangle
* @param PixelXRadius - Horizontal radius or both if PixelYRadius is not specified
* @param PixelYRadius - Vertical radius of the corners
*/
CMView.prototype.PaintRoundedRect=function(PixelXMin,PixelXMax,PixelYMin,PixelYMax,PixelXRadius,PixelYRadius) 
{
	if (typeof PixelYRadius === 'undefined') { PixelYRadius=PixelXRadius; }
	
	this.TheContext.beginPath();
	this.TheContext.moveTo(PixelXMin + PixelXRadius, PixelYMin); // left side of top
	this.TheContext.lineTo(PixelXMax - PixelXRadius, PixelYMin); // right side of top
	this.TheContext.quadraticCurveTo(PixelXMax, PixelYMin, PixelXMax, PixelYMin+PixelYRadius); // top of right side
	this.TheContext.lineTo(PixelXMax, PixelYMax-PixelYRadius); // bottom of right side
	this.TheContext.quadraticCurveTo(PixelXMax, PixelYMax, PixelXMax - PixelXRadius,PixelYMax); // right of bottom
	this.TheContext.lineTo(PixelXMin+PixelXRadius, PixelYMax); // left of bottom
	this.TheContext.quadraticCurveTo(PixelXMin, PixelYMax, PixelXMin, PixelYMax - PixelYRadius); // bottom of left side
	this.TheContext.lineTo(PixelXMin, PixelYMin+PixelYRadius); // top of left side
	this.TheContext.quadraticCurveTo(PixelXMin, PixelYMin, PixelXMin+PixelXRadius,PixelYMin); // back to left side of top
	this.TheContext.closePath();
	
	if (this.TheContext.strokeStyle!=null) this.TheContext.stroke();
	if (this.TheContext.fillStyle!=null) this.TheContext.fill();
}
/**
* Function to paint text with the current font metrics in the context.
* @public
* @param XInPixels - left edge of the text
* @param YInPixels - base line for the text
* @param Text - the text to paint
* @param RadAngle - optional angle to rotate the text
*/
CMView.prototype.PaintText=function(XInPixels,YInPixels,Text,RadAngle)
{
	if ((RadAngle!=undefined))
	{
		this.TheContext.save();
		try {
			// translate to the deisred location and then rotate about that location
			this.TheContext.translate(XInPixels,YInPixels);
			this.TheContext.rotate(RadAngle);
		
			// render the text
			//this.TheContext.fillText(Text,0,0);
			if (this.TheContext.strokeStyle!=null) this.TheContext.strokeText(Text,0,0);
			if (this.TheContext.fillStyle!=null) this.TheContext.fillText(Text,0,0);
		}
		catch(err) {
			console.log(err.message);
		}
		// take out the rotation and the translation
		this.TheContext.restore();
	}
	else
	{
		if (this.TheContext.strokeStyle!=null) this.TheContext.strokeText(Text,0,0);
		if (this.TheContext.fillStyle!=null) this.TheContext.fillText(Text,0,0);
//		this.TheContext.fillText(Text,XInPixels,YInPixels);
	}
}


